open Syntax;;
open Multisetorder;;
(*** open Multisetorder;; ***)
type rectyId = int
type tvid = int
type ttype = UnitT
           | BoolT
           | IntT
           | ChanT of ttype * level
           | PairT of ttype * ttype
           | FunT of ttype * ttype
           | RecT of rectyId * ttype list 
type tenv = (Syntax.vid, ttype) Utilities.env


exception TODO;;
exception Error of string;;
exception Fatal of string;;
exception VerificationFailure of int list

type mi = (level*vid*vid) list
type mo = MO_atom of level * Syntax.vid * ttype Syntax.valexp
         | MO_assume of ttype Syntax.valexp * mo
         | MO_or of mo * mo
         | MO_par of mo * mo
         | MO_empty

let mi_empty = []
let mi_union mi1 mi2 = mi1@mi2
let mi_atom x = [x]
let mo_empty = MO_empty

type constr = EqL of level * level 
            | SubM of mi * mo;;

type constraints = constr list

let merge_constraints = List.fold_left (@) []

let emptyC = []

let current_tvid = ref 0;;
let new_tvid() = 
    let x = !current_tvid in
    let _ = (current_tvid := !current_tvid+1) in
       x;;
let current_lvid = ref 0;;
let new_lvid() = 
    let x = !current_lvid in
    let _ = (current_lvid := !current_lvid+1) in
       x;;
let dummy_level = LVar(-1);;
let level_table = ref [];;
let register_level (LVar id) st =
    level_table := (id,st)::!level_table;;

let reset_levels() =
    (current_lvid := 0; level_table := []);;

let new_level() = LVar (new_lvid());;

let dummy_pc = -1;;

let rec new_ttype t =
  match t with
    Stype.TVar n -> UnitT
  | Stype.UnitT -> UnitT
  | Stype.BoolT -> BoolT
  | Stype.IntT -> IntT
  | Stype.ChanT t1 -> let l1 = new_level() in
                      (register_level l1 t1; ChanT(new_ttype t1, l1))
  | Stype.PairT(t1,t2) -> PairT(new_ttype t1, new_ttype t2)
  | Stype.FunT(t1,t2) -> FunT(new_ttype t1, new_ttype t2)
  | Stype.RecT(id,tl) -> RecT(id, List.map new_ttype tl)

let const_table = [("+", fun _ -> FunT(PairT(IntT, IntT),IntT));
                   ("-", fun _ -> FunT(PairT(IntT, IntT),IntT));
                   ("*", fun _ -> FunT(PairT(IntT, IntT),IntT));
                   ("/", fun _ -> FunT(PairT(IntT, IntT),IntT));
                   ("=", fun _ -> FunT(PairT(IntT, IntT),BoolT));
                   (">", fun _ -> FunT(PairT(IntT, IntT),BoolT));
                   (">=", fun _ -> FunT(PairT(IntT, IntT),BoolT));
                   ("<", fun _ -> FunT(PairT(IntT, IntT),BoolT));
                   ("<=", fun _ -> FunT(PairT(IntT, IntT),BoolT));
                   ("fst", fun t -> 
                            match t with
                               Stype.FunT(Stype.PairT(t1,t2),t3) -> 
                                 let t1' = new_ttype t1 in
                                 let t2' = new_ttype t2 in
                                     FunT(PairT(t1',t2'),t1')
                             | _ -> raise (Fatal "const_table"));
                   ("snd", fun t -> 
                            match t with
                               Stype.FunT(Stype.PairT(t1,t2),t3) -> 
                                 let t1' = new_ttype t1 in
                                 let t2' = new_ttype t2 in
                                     FunT(PairT(t1',t2'),t2')
                             | _ -> raise (Fatal "const_table"));
                   ("nil", fun t -> new_ttype t);
                   ("cons", fun t ->
                             match t with
                               Stype.FunT(Stype.PairT(t1,Stype.RecT(_, [t2;_])),_) ->
                                    let t1' = new_ttype t1 in
                                    let t2' = new_ttype t2 in
                                        (FunT(PairT(t1', RecT(0, [t2';t2'])), RecT(0, [t1';t2'])))
                             | _ -> raise (Fatal "const_table:cons"));
                   ("hd", fun t -> 
                           match t with
                               Stype.FunT(Stype.RecT(_,[t1;t2]), _) ->
                                    let t1' = new_ttype t1 in
                                    let t2' = new_ttype t2 in
                                       FunT(RecT(0,[t1';t2']), t1')
                             | _ -> raise (Fatal "const_table:hd"));
                   ("tl", fun t -> 
                           match t with
                               Stype.FunT(Stype.RecT(_,[t1;t2]), _) ->
                                    let t1' = new_ttype t1 in
                                    let t2' = new_ttype t2 in
                                       FunT(RecT(0,[t1';t2']), RecT(0,[t2';t2']))
                             | _ -> raise (Fatal "const_table:hd"));
                   ("null", fun t -> 
                           match t with
                               Stype.FunT(Stype.RecT(_,[t1;t2]),_) ->
                                  let t1' = new_ttype t1 in
                                  let t2' = new_ttype t2 in
                                      FunT(RecT(0, [t1';t2']), BoolT)
                            | _ -> raise (Fatal "const_table:null"))];;

let lookup_prim prname t = 
  try
    (List.assoc prname const_table) t
  with
    Not_found -> raise (Fatal ("unknown primitive "^prname));;

let rectype_print_table =
  ref [(0, fun f -> fun tl -> ((f 3 (List.hd(tl)))^" list"))]

let rec print_level_order x =
  match x with
    [] -> (print_string "none\n")
  | [(l1,l2)] -> print_string ((string_of_int l1)^" >= "^(string_of_int l2))
  | (l1,l2)::x' ->
           (print_string ((string_of_int l1)^" >= "^(string_of_int l2)^", ");
            print_level_order x')

let string_of_type t =
  let rec t2string pr t = 
    match t with
      UnitT -> "unit"
    | BoolT -> "bool"
    | IntT -> "nat"
    | ChanT(t1,LVar l) -> "chan("^(t2string 0 t1)^", "^(string_of_int l)^")"
    | PairT(t1,t2) -> 
           (if pr>1 then "(" else "")
           ^(t2string 2 t1)^"*"^(t2string 2 t2)
          ^(if pr>2 then ")" else "")
    | FunT(t1,t2) -> 
           (if pr>0 then "(" else "")
          ^(t2string 1 t1)
          ^"->"
          ^(t2string pr t2)
          ^(if pr>0 then ")" else "")
    | RecT(id, tl) ->
        let f = List.assoc id !rectype_print_table in
          f t2string tl
  in t2string 0 t;;

let print_env env = Utilities.print_env (fun x->x) string_of_type env;;

let tenv_extend x tx env =
    Utilities.env_extend x tx env;;

let rec makeEqT t1 t2 =
  match (t1,t2) with
    (UnitT,UnitT) -> []
  | (BoolT,BoolT) -> []
  | (IntT,IntT) -> []
  | (ChanT(t1',l1), ChanT(t2',l2)) ->
           (EqL (l1, l2))::(makeEqT t1' t2')
  | (PairT(t11,t12), PairT(t21,t22)) ->
         (makeEqT t11 t21)@(makeEqT t12 t22)
  | (FunT(t11,t12), FunT(t21,t22)) ->
         (makeEqT t11 t21)@(makeEqT t12 t22)
  | (RecT(id1, tl1), RecT(id2,tl2)) ->
      if id1=id2 then
        let tpl = List.combine tl1 tl2 in
          List.fold_left (fun c -> fun (t1,t2) -> (makeEqT t1 t2)@c) [] tpl
      else raise (Fatal "makeEqT")
  | _ -> raise (Fatal "makeEqT")

let mkTrue v = v
let mkFalse v = Syntax.App(Syntax.Prim("not", None), v, dummy_pc)


let dummy_mo = MO_atom(dummy_level, "_dummy", Unit)
let c_mleq mi mo newc =
  let mo1 = if newc then MO_par(mo, dummy_mo)
            else mo in
    [SubM(mi, mo1)]

let mo_union mo1 mo2 = MO_par(mo1,mo2)
let mo_add_dummy mo newc =
  if newc then MO_par(mo,dummy_mo) else mo
let mo_cond v mo1 mo2 newc1 newc2 = 
   let mo1' = mo_add_dummy mo1 newc1 in
   let mo2' = mo_add_dummy mo2 newc2 in
       MO_or (MO_assume(mkTrue(v), mo1'), MO_assume(mkFalse(v), mo2'))
let mo_or mo1 mo2 = 
      MO_or(mo1, mo2)
let rec app_subst_for_mo s mo =
  match mo with
    MO_atom(l,x,v) -> MO_atom(l,x,Syntax.apply_vsubst_v s v)
  | MO_or(mo1,mo2) -> MO_or(app_subst_for_mo s mo1, app_subst_for_mo s mo2)
  | MO_assume(v, mo1) -> MO_assume(Syntax.apply_vsubst_v s v, app_subst_for_mo s mo1)
  | MO_par(mo1,mo2) -> MO_par(app_subst_for_mo s mo1, app_subst_for_mo s mo2)
  | MO_empty -> MO_empty

(*** tinf: stype procexp -> tenv -> ttype procexp * constraints * min * mout * newc ***)
let rec tinf (p: Stype.stype procexp) env  =
  match p with
    Zero -> (Zero, emptyC, mi_empty, mo_empty, false)
  | Out(x,v,p1,pc) ->
      let (v', t1, c1) = tinfV v env in
      let (p1',c2,_, mo1, newc1) = tinf p1 env in
      let t3 = tinfVar x env in
      let (c3, mo2) = match t3 with
                 ChanT(t4,lx) -> 
                     (makeEqT t4 t1, MO_atom(lx, x, v'))
               | _ -> raise (Fatal "tinf") in
      let c = merge_constraints [c1;c2;c3] in
        (Out(x,v',p1',pc), c, mi_empty, mo_union mo1 mo2, newc1)
  | In(x,y,t1,p1,pc) ->
      let tx = tinfVar x env in
      let (ty,lx) = match tx with 
                       ChanT(ty,lx) -> (ty,lx)
                     | _ -> raise (Fatal "tinf") in
      let env1 = tenv_extend y ty env in
      let (p1',c, mi1, mo1,newc1) = tinf p1 env1 in
      let mi = mi_union (mi_atom(lx, x, y)) mi1 in
        (In(x,y,ty,p1',pc),c, mi, mo1, newc1)
  | Par(p1,p2) ->
      let (p1',c1, mi1, mo1, newc1) = tinf p1 env in
      let (p2',c2, mi2, mo2, newc2) = tinf p2 env in
      let c = merge_constraints [c1;c2] in
       (Par(p1',p2'), c, mi_empty, mo_union mo1 mo2, newc1||newc2)
  | Rep(p1) ->
      let (p1', c1, mi1, mo1, newc1) = tinf p1 env in
         if mi1=mi_empty then
            raise (Error "General form of replication is not supported.")
         else 
           (Rep(p1'), merge_constraints [(c_mleq mi1 mo1 newc1); c1], mi_empty, mo_empty, false)
  | Nu(x,stx,p1,pc) ->
      let tx = new_ttype stx in
      let env1 = tenv_extend x tx env in
      let (p1',c1, _, mo1, _) = tinf p1 env1 in
        (Nu(x,tx,p1',pc), c1, mi_empty, mo1, true)
  | If(v, p1, p2, pc) ->
      let (v',tv,c0) = tinfV v env in
      let (p1',c1, _, mo1, newc1) = tinf p1 env in
      let (p2',c2, _, mo2, newc2) = tinf p2 env in
      let c = merge_constraints [c0;c1;c2] in
       (If(v',p1',p2',pc), c, mi_empty, mo_cond v' mo1 mo2 newc1 newc2, false)
  | Let(x,t1,v,p1,pc) ->
      let (v',tv, c1) = tinfV v env in
      let env1 = Utilities.env_extend x tv env in
      let (p1',c2, mi1, mo1, newc1) = tinf p1 env1 in 
      let c = merge_constraints [c1;c2] in
      let mo1' = app_subst_for_mo (Syntax.make_vsubst x v') mo1 in
        (Let(x,tv,v', p1',pc), c, mi1, mo1', newc1)
  | MatchL(x,matchlist,pc) -> 
      let tx = tinfVar x env in
      let (matchlist',c, mo1) = tinfMatchL tx matchlist env in
        (MatchL(x,matchlist',pc), c, mi_empty, mo1, false)
  | Special(p1) ->
      let (p1',c1, mi1, mo1, newc1) = tinf p1 env in
       (Special(p1'), c1, mi1, mo1, newc1)
and tinfMatchL t matchlist env  =
   match matchlist with
      [] -> ([], emptyC, mo_empty)
    | [(cid, p1)] -> 
         let (p1',c1, _,mo1,newc1) = tinf p1 env in
            ([(cid,p1')], c1, mo_add_dummy mo1 newc1)
    | (cid, p1)::matchlist1 ->
         let (matchlist1',c1, mo1) = tinfMatchL t matchlist1 env  in
         let (p1',c2, _, mo2, newc2) = tinf p1 env in
         let mo2' = mo_add_dummy mo2 newc2 in
         let c = merge_constraints [c1;c2] in
            ((cid,p1')::matchlist1', c, mo_or mo1 mo2')
and tinfVar x env =
   try 
     Utilities.env_lookup x env
   with
     Not_found -> raise (Fatal ("unknown variables "^x))
and tinfV v env = 
   match v with
     Var(x) -> (Var(x), tinfVar x env, emptyC)
   | Unit -> (Unit, UnitT, emptyC)
   | SBool(b,sl) -> (SBool(b,sl), BoolT, emptyC)
   | SInt(b,sl) -> (SInt(b,sl), IntT, emptyC)
   | Bool(b) -> (Bool(b), BoolT, emptyC)
   | Int(n) -> (Int n, IntT, emptyC)
   | Prim(s,ot) -> 
        (match ot with
           Some t1 -> let t = lookup_prim s t1 in (Prim(s,Some t), t, emptyC)
         | None -> raise (Fatal "tinfV")
         )
   | App(v1,v2, pc) ->
        let (v1',t1, c1) = tinfV v1 env in
        let (v2',t2, c2) = tinfV v2 env in
        let (t11,t12) = match t1 with
                         FunT(t11,t12) -> (t11,t12)
                         | _ -> raise (Fatal "tinfV") in
        let c3 = makeEqT t2 t11 in
        let c = merge_constraints [c1;c2;c3] in
          (App(v1',v2',pc), t12, c)
   | Pair(v1,v2) -> 
        let (v1',t1, c1) = tinfV v1 env in
        let (v2',t2, c2) = tinfV v2 env in
        let c = merge_constraints [c1;c2] in
          (Pair(v1',v2'),PairT(t1,t2), c)

(***
let app_lsubst s x = Utilities.subst_var s x x
let comp_lsubst = Utilities.comp_subst app_lsubst
***)

let rec app_lsubst_for_t s t =
  match t with
    UnitT -> UnitT
  | BoolT -> BoolT
  | IntT -> IntT
  | PairT(t1,t2) -> PairT(app_lsubst_for_t s t1, app_lsubst_for_t s t2)
  | FunT(t1,t2) -> FunT(app_lsubst_for_t s t1, app_lsubst_for_t s t2)
  | RecT(id,tl) -> RecT(id, List.map (app_lsubst_for_t s) tl)
  | ChanT(t1,LVar lev) -> ChanT(app_lsubst_for_t s t1, LVar(app_lsubst s lev))

let rec app_lsubst_for_mo s mo =
  match mo with
    MO_atom(LVar n, x, v) -> MO_atom(LVar(app_lsubst s n), x, v)  
  | MO_or(mo1,mo2) -> MO_or(app_lsubst_for_mo s mo1, app_lsubst_for_mo s mo2)
  | MO_assume(v,mo1) -> MO_assume (v, app_lsubst_for_mo s mo1)
  | MO_par(mo1,mo2) -> MO_par(app_lsubst_for_mo s mo1, app_lsubst_for_mo s mo2)
  | MO_empty -> MO_empty

(*** reduce_eqL:
 ***  constr list ->
 ***   (int, int) Utilities.substitution *
 ***   (int * int * Syntax.vid * ttype Syntax.valexp *
 ***   (Syntax.vid * ttype Syntax.valexp) list * ttype Syntax.valexp list) list
 ***)
let reduce_eqL c =
  let app_subst_for_subm s (mi, mo) =
    let mi' = List.map (fun (LVar n, x, v) -> (LVar(app_lsubst s n), x, v)) mi in
    let mo' = app_lsubst_for_mo s mo in
       (mi', mo') 
  in
  let app_subst_for_ac s ac =
    match ac with
      EqL(LVar n1, LVar n2) -> EqL(LVar(app_lsubst s n1), LVar(app_lsubst s n2))
    | SubM(mi, mo) -> 
        let (mi',mo') = app_subst_for_subm s (mi, mo) in
          SubM(mi', mo') in
  let app_subst_for_c s c =
    List.map (app_subst_for_ac s) c in
  let rec reduce_it c s subm =
    match c with
      [] -> (s, subm)
    | EqL(LVar n1,LVar n2)::c' -> 
         if n1=n2 then reduce_it c' s subm
         else 
           let s1 = Utilities.make_subst n1 n2 in
           let s2 = comp_lsubst s1 s in
           let c'' = app_subst_for_c s1 c' in
           let subm' = List.map (app_subst_for_subm s1) subm in
             reduce_it c'' s2 subm'
   | SubM(mi, mo)::c' -> 
             reduce_it c' s ((mi, mo)::subm) in
  reduce_it c (Utilities.subst_empty) [];;

let rec mk_admissible_type st f =
  match st with
     Stype.UnitT -> (UnitT,f)
   | Stype.BoolT -> (BoolT,f)
   | Stype.IntT -> (IntT,f)
   | Stype.ChanT(st1) ->
       (try 
           (List.assoc st f, f)
        with Not_found ->
           let (t1,f') = mk_admissible_type st1 f in
           let l1 = new_level() in
           let _ = register_level l1 st1 in
           let t = ChanT(t1, l1) in
            (t, (st,t)::f))
   | Stype.PairT(st1,st2) ->
       let (t1,f') = mk_admissible_type st1 f in
       let (t2,f'') = mk_admissible_type st2 f' in
         (PairT(t1,t2), f'')
   | Stype.FunT(st1,st2) ->
       let (t1,f') = mk_admissible_type st1 f in
       let (t2,f'') = mk_admissible_type st2 f' in
         (FunT(t1,t2), f'')
   | Stype.RecT(id,stl) ->
       let (tl,f') = List.fold_left 
                      (fun (tl1,f1) -> fun st -> 
                        let (t1, f1') = mk_admissible_type st f1 in
                           (t1::tl1, f1'))
                      ([], f) stl in
         (RecT(id, tl), f')
   | Stype.TVar _ -> (UnitT, f);;
   
let rec mk_admissible_TE stenv f =
  match stenv with
    [] -> ([], f)
  | (x,st)::stenv' ->
       let (t,f') = mk_admissible_type st f in
       let (tenv, f'') = mk_admissible_TE stenv' f' in
         ((x,t)::tenv, f'');;

(***
let rec normalize_subl c =
  let add_subl ordc (l1,l2,x,v,eql,cond) =
      let s = Utilities.list2subst eql in
      let v' = Syntax.apply_vsubst_v s v in
      let cond' = List.map (Syntax.apply_vsubst_v s) cond in
      try 
        let (labl, ordc') = Utilities.list_assoc2 (l1,l2) ordc in
           ((l1,l2), (x,v',cond')::labl)::ordc'
      with Not_found ->
         ((l1,l2), [(x,v',cond')])::ordc
  in
    List.fold_left add_subl [] c
***)

type svid = string

let current_sid = ref 0;;
let new_sid() = 
    let x = !current_sid in
    let _ = (current_sid := !current_sid+1) in
       x;;
let new_size() = ([(1, new_sid())], 0)

let rec mk_size_template st arg pos =
  match st with
    Stype.TVar _ -> S_NONE
  | Stype.UnitT -> S_NONE
  | Stype.BoolT -> S_NONE
  | Stype.IntT -> S_EQ(arg, pos)
  | Stype.ChanT(_) -> S_NONE
  | Stype.PairT(st1,st2) -> 
      S_PAIR(mk_size_template st1 arg ("fst_"^pos), mk_size_template st2 arg ("snd_"^pos))
  | Stype.FunT(st1,st2) -> S_NONE
  | Stype.RecT(id,stl) -> S_EQ(arg, pos)  (*** todo: This should be changed according to id ***)

(*** type size = (int * svid) list * int  (*** a1x1+...+anxn + b ***) ***)
(*** type size_struct = Snone | Satom of size | Spair of size_struct * size_struct ***)
(***
 *** let rec mk_size_template st =
 ***   match st with
 ***     Stype.IntT -> Satom(new_size())
 ***   | Stype.BoolT -> Satom(new_size())
 ***   | Stype.ChanT(_) -> Snone
 ***   | Stype.PairT(st1,st2) -> Spair(mk_size_template st1, mk_size_template st2)
 ***   | Stype.FunT(st1,st2) -> Snone
 ***   | Stype.RecT(id,stl) -> Satom(new_size())
 ***   | _ -> raise (Fatal "mk_size_template")
 *** let size_int n = Satom ([], n)
 *** let rec size_minus e1 e2 =
 ***   match (e1,e2) with
 ***     ([], []) -> []
 ***   | (e1, []) -> e1
 ***   | ([], e2) -> List.map (fun (a,x)->(-a,x)) e2
 ***   | ((a1,x1)::e1', (a2,x2)::e2') ->
 ***         if x1<x2 then (a1,x1)::(size_minus e1' e2)
 ***         else if x1>x2 then (-a2,x2)::(size_minus e1 e2')
 ***         else (a1-a2,x1)::(size_minus e1' e2')
 ***)

let unsupported x nonzeros = S_NONE;;
let size_decrement x nonzeros =
  match x with
    S_EQ(arg,pos) -> S_LT(arg,pos)
  | S_LEQ(arg,pos) -> S_LT(arg,pos)
  | S_LT(arg,pos) -> S_LT(arg,pos)
  | _ -> S_NONE;;

let size_fun_table =
  [("+", unsupported);
   ("-", fun s -> fun nonzeros ->
          (match s with
             S_PAIR(s1,s2) ->
              (let dec =  
                (match s2 with 
                   S_EQ(arg,pos) -> List.mem (arg,pos) nonzeros
                  | S_CONST(n) -> n>0
                  | _ -> false) in
                match s1 with
                   S_EQ(arg,pos) -> if dec then S_LT(arg,pos) else S_LEQ(arg,pos)
                 | S_LEQ(arg,pos) -> if dec then S_LT(arg,pos) else S_LEQ(arg,pos)
                 | S_LT(arg,pos) -> S_LT(arg,pos)
                 | _ -> S_NONE)
           | _ -> S_NONE));
   ("fst", fun s -> fun nonzeros -> match s with (S_PAIR(s1,s2)) -> s1 | _ -> S_NONE);
   ("snd", fun s -> fun nonzeros -> match s with (S_PAIR(s1,s2)) -> s2 | _ -> S_NONE);
   ("hd", fun s -> fun nonzeros -> S_NONE);
   ("tl", size_decrement)]

let lookup_size_fun f =
  try List.assoc f size_fun_table
  with Not_found -> unsupported;;

(***
 *** let rec eval_size v senv =
 ***   match v with
 ***     Var x -> (try List.assoc x senv with Not_found -> Snone)
 ***   | Unit -> Snone
 ***   | SBool(b,_) -> eval_size (Bool(b)) senv
 ***   | SInt(n,_) -> eval_size (Int(n)) senv
 ***   | Bool(b) -> if b then size_int 1 else size_int 0
 ***   | Int(n) -> size_int n
 ***   | Prim(f,_) -> Snone
 ***   | App(Prim (f,_), v1,_) -> lookup_size_fun f (eval_size v1 senv)
 ***   | Pair(v1,v2) -> Spair(eval_size v1 senv, eval_size v2 senv)
 ***   | _ -> Snone
 ***)

let rec eval_size v senv nonzeros =
  match v with 
    Var x -> (try (List.assoc x senv) with Not_found -> S_NONE)
  | Unit -> S_NONE
  | SBool(b,_) -> S_NONE
  | SInt(n,_) -> S_CONST n
  | Bool(b) -> S_NONE
  | Int(n) -> S_CONST n
  | Prim(f,_) -> S_NONE
  | App(Prim (f,_), v1,_) -> lookup_size_fun f (eval_size v1 senv nonzeros) nonzeros
  | Pair(v1,v2) -> S_PAIR(eval_size v1 senv nonzeros, eval_size v2 senv nonzeros)
  | _ -> S_NONE



(*** find_nonzero finds variables whose values are not zero
 *** The current implementaiton is extremely naive, and should be revised later
 ***)
let rec find_nonzero b senv nonzeros =
     let union x y = Utilities.merge_and_unify compare x y in
     match b with
       App(Prim ("not",_), b',_) -> union (find_may_zero b' senv nonzeros) nonzeros
     | App(Prim (">",_), Pair(v1, _),_) ->
          let sv = eval_size v1 senv nonzeros in
            (match sv with
               S_LT(arg,pos) -> union [(arg,pos)] nonzeros
             | S_LEQ(arg,pos) -> union [(arg,pos)] nonzeros
             | S_EQ(arg,pos) -> union [(arg,pos)] nonzeros
             | _ -> nonzeros)
     | App(Prim (">=",_), Pair(v1, _),_) ->
          let sv = eval_size v1 senv nonzeros in
            (match sv with
               S_LT(x) -> union [x] nonzeros
             | _ -> nonzeros)
     | _ -> nonzeros
and find_may_zero b senv nonzeros=
     match b with
       App(Prim ("=",_), Pair(v1,Int(0)),_) ->
          let sv = eval_size v1 senv nonzeros in
            (match sv with
               S_EQ(arg,pos) -> [(arg,pos)]
             | _ -> [])
     | App(Prim ("<=",_), Pair(v1,Int(n)),_) ->
          let sv = eval_size v1 senv nonzeros in
            (match sv with
               S_LT(arg,pos) -> [(arg,pos)]
             | S_LEQ(arg,pos) -> [(arg,pos)]
             | S_EQ(arg,pos) -> [(arg,pos)]
             | _ -> [])
     | _ -> []

let rec mo2exlevel senv mo nonzeros =
  match mo with
    MO_atom(l, subj, v) -> let size = eval_size v senv nonzeros in [[(l, subj, size)]]
  | MO_assume(b, mo1) -> 
       let nonzeros' = find_nonzero b senv nonzeros in
             mo2exlevel senv mo1 nonzeros'
  | MO_or(mo1, mo2) -> 
       let exlevel_mm1 = mo2exlevel senv mo1 nonzeros in
       let exlevel_mm2 = mo2exlevel senv mo2 nonzeros in
            exlevel_mm1@exlevel_mm2
  | MO_par(mo1, mo2) ->
       let exlevel_mm1 = mo2exlevel senv mo1 nonzeros in
       let exlevel_mm2 = mo2exlevel senv mo2 nonzeros in
       let x = Utilities.make_pairs exlevel_mm1 exlevel_mm2 in
        List.map (fun (l1,l2)->l1@l2) x
  | MO_empty -> [[]]

let subm2msetr (mi, mo) =
  let senv = List.map (fun (LVar(l), subj, arg) -> 
                 (arg, mk_size_template (List.assoc l !level_table) arg "x")) mi in
  let exlev_list = mo2exlevel senv mo [] in
     List.map (fun exlev -> (mi, exlev)) exlev_list

let rec constraints2msetr c =
  match c with
    [] -> []
  | (mi,mo)::c' -> (subm2msetr(mi,mo))@(constraints2msetr c')

let rec delete n ms =
  match ms with
   [] -> raise Not_found
 | m::ms' -> if n=m then ms' else m::(delete n ms')

let rec delete_same_names (ms1,ms2) =
 match ms1 with
   [] -> (ms1,ms2)
 | n::ms1' -> try
               let ms2' = delete n ms2 in delete_same_names(ms1',ms2')
              with
               Not_found -> 
                 let (ms1'',ms2'') = delete_same_names(ms1', ms2) in (n::ms1'',ms2'')
      
let check_ordc (ms1, ms2) ord =
  let (ms1', ms2') = delete_same_names (ms1, ms2) in
  if ms1'=[] || not(List.sort compare (List.map snd ms1')=List.sort compare (List.map snd ms2'))
  then false
  else
    let ord' = List.filter 
               (fun (x,y) -> 
                  try
                    List.assoc x ms1' = List.assoc y ms2'
                  with
                    Not_found -> false)
               ord 
    in
    let subsumed_elements = List.map snd ord' in
      List.for_all (fun (x,_)->List.mem x subsumed_elements) ms2' 
 
(***    in
      Utilities.perfect_matching (List.map fst ms1') (List.map fst ms2') ord'
 ***)

let rec check_ordcs ordcs ord =
 match ordcs with
   [] -> None
 | c::ordcs' -> 
     if (check_ordc c ord) 
     then (check_ordcs ordcs' ord)
     else Some(c)

let rec print_order ord =
  match ord with
    [] -> ()
  | [(x,y)] -> print_string (" "^x^">>"^y)
  | (x,y)::ord' -> (print_string (" "^x^">>"^y^", "); print_order ord')

let rec string_of_names ns =
  match ns with
     [] -> ""
   | n::ns' -> n^" "^(string_of_names ns')

let string_of_ordc (ms1, ms2) =
  let ms1' = List.map fst ms1 in
  let ms2' = List.map fst ms2 in
    (string_of_names ms1')^">>"^(string_of_names ms2')

let termination_check (p, stenv) =
  let _ = reset_levels() in
  let (tenv,_) = mk_admissible_TE (Utilities.env2list stenv) [] in
  (*** TODO: alpha conversion is necessary before tinf is called ***)
  let (p1',c, mi1, mo1, newc)= tinf p (Utilities.list2env tenv)  in
  let (s1, c') = reduce_eqL c in
  let msetr = constraints2msetr c' in
  let (edges, s2, success, ordopt) = 
       try 
         let (edges, s, success) = Multisetorder.solve_constraints msetr in
           (edges, s, success, None)
       with OrdConstr(ordcs, edges, s) ->
         let (p2, ord) = Otype.elaborate (p, stenv) in
         let result = check_ordcs ordcs ord in
         let success =
             match result with
                None -> None
              | Some(ms1,ms2) ->
                 Some ("The following type 4 constraint could not be solved\n  "
                       ^string_of_ordc(ms1,ms2))
         in
           (edges, s, success, Some ord) 
  in
  let s = Multisetorder.comp_lsubst s2 s1 in
  let p'' = Syntax.map_for_t (app_lsubst_for_t s) p1' in
  let p3 = Syntax.undo_alphaconv p'' in
  let tenv' = Utilities.env_map (app_lsubst_for_t s) (Utilities.list2env tenv) in
    (Syntax.pp_proc (Some string_of_type) p3;
     print_string "\n where its type environment is:\n"; 
     print_env tenv';
     print_string "\nThe order between levels is:\n  ";
     print_level_order edges;
     print_string "\n\n";
     (match ordopt with
        None -> ()
      | Some(ord) -> 
          (print_string "The following name ordering was used\n";
           print_order(ord);
           print_string "\n\n"));
     match success with
       None -> 
         (print_string ("Program is robustly terminating!\n"); true)
     | Some(s) ->
         (print_string s; print_newline();
          (print_string "Verification failed\n"); false))

let elaborate p =
  let p' = Syntax.alphaconv p in
  let (p1, stenv) = Stype.elaborate p' in
  let _ = termination_check (p1, stenv) in
    ()


