(**** Copyright 2004-2006 (see README and COPYING for details in the top directory) ***)

open Syntax;;
open Utilities;;

type slvid = int;;
type tvid = int;;
type lvid = int;;
type levexp = LevRec of int 
            | LevConst of int
            | LevInfty
            | LevVar of lvid
            | LevIfInfty of slevel (*** InvIfInfty(s): if s=LOW then Infty else 0 ***)
            | LevPlus of levexp * int
            | LevMax of levexp list
            | LevMin of levexp list
and slevel = HI
           | LOW
           | SVar of slvid
           | SIfInf of levexp * slevel (*** SIfInf(l,s) = if l=Infty then s else H ***)
           | Sglb of slevel * slevel
          ;;
(*** type level = LConst of int | LInfty | LVar of lvid | PlusL of level * int;; ***)

type uvid = int;;

type usage = UVar of uvid
           | ZeroU 
           | InU of levexp * levexp * usage 
              (*** The first level is obligation and the second one is capability ***)
           | OutU of levexp * levexp * usage
           | ParU of usage * usage
           | OrU of usage * usage
           | RepU of usage
           | IncU of usage
           | UpU of usage * levexp * levexp
               (*** The first level describes an input obligation and 
                    the second one describes an output obligation ***)
           | URec of uvid * usage;;

type rectyId = int;; (*** Identifier of recursive types: e.g. 0: list ***)

type vtype = TVar of tvid 
           | UnitT 
           | BoolT of slevel 
           | IntT of slevel 
           | ChanT of vtype * slevel * usage 
           | PairT of vtype*vtype 
           | FunT of vtype*vtype 
	       (*** The security level is not needed for a function type:
                *** Its security level is the same as that of the return type
                ***)
           | RecT of rectyId * (vtype list) * slevel
               (*** Recursive (variant) type, a la ML datatypes. 'a list = RecT(0, ['a, 'a], sl) 
                *** sl is the security level of information about the outermost label of the variant ***)
           | ParT of vtype * vtype
           | IncT of vtype
           | RepT of vtype
           | UpT of vtype * levexp;;

type atomic_constraint =
     EqT of vtype * vtype
   | TopT of vtype
   | SubT of vtype * vtype
   | WfT of slevel * vtype 
      (** vtype is well-formed, and its security level is more than or equal to slevel **)
   | LowT of vtype 
      (** LowT(t) <--> all the security levels in t are low and 
                       capability levels are infinity **)
   | SubU of uvid * usage
   | RelU of usage
   | SubL of levexp * levexp
   | SubS of slevel * slevel
   | SubSIf of levexp * slevel * slevel (*** l=LevInfty -> s1 < s2 ***)
   | EqS of slevel * slevel
   | NocapT of vtype;;

type constr = atomic_constraint list;;

exception FailWlockAnalysis;;
exception TypeError of string;;
exception UnExpectedConstraint of atomic_constraint;;
exception UnExpectedUsage of usage;;
exception UnExpectedSL of slevel;;
exception UnExpectedType of vtype;;
exception SomethingIsWrongIn of string;;

(**** 2007/9: BEGIN ****)
let te_table = ref [];;
let reg_TEtable pc te = (te_table := (pc,te)::!te_table);;
let rem_TEtable pc = 
  (te_table := List.filter (fun (pc1,_)->not(pc=pc1)) !te_table);;
let get_TEtable () = List.map snd !te_table;;
let map_TEtable f = 
  (te_table := List.map (fun (pc,g) -> (pc, f g)) !te_table);;
let reset_TEtable () =
  (te_table := []);;
(**** 2007/9: END ****)


(*** type_fold_usage: (usage ->'a) -> ('a->'a->'a) -> 'a -> vtype -> 'a ***)
let rec type_fold_usage f combine default t =
  match t with
    TVar _ -> default
  | UnitT -> default
  | BoolT _ -> default
  | IntT _ -> default
  | ChanT(t1,_,u) -> combine(type_fold_usage f combine default t1) (f u)
  | ParT(t1,t2) -> combine (type_fold_usage f combine default t1)
                            (type_fold_usage f combine default t2)
  | PairT(t1,t2) -> combine (type_fold_usage f combine default t1)
                            (type_fold_usage f combine default t2)
  | FunT(t1,t2) ->  combine (type_fold_usage f combine default t1)
                            (type_fold_usage f combine default t2)
  | RecT(id, tl, sl) -> let l = List.map (type_fold_usage f combine default) tl in
                       List.fold_left combine default l
  | IncT(t1) -> (type_fold_usage f combine default t1)
  | RepT(t1) -> (type_fold_usage f combine default t1)
  | UpT(t1,_) -> (type_fold_usage f combine default t1);;

let emptyC = [];;
let makeC_EqT(t1,t2):constr = [EqT(t1,t2)];;
let makeC_TopT(t):constr = [TopT(t)];;
let makeC_LowT(t):constr = [LowT(t)];;
let makeC_SubT(t1,t2):constr = [SubT(t1,t2)];;
let makeC_SubS(s1,s2):constr = [SubS(s1,s2)];;
let makeC_SubL(l1,l2):constr = [SubL(l1,l2)];;
let makeC_SubSIf(lv, s1,s2):constr = [SubSIf(lv,s1,s2)];;
let makeC_SubU(uvid,u):constr = [SubU(uvid,u)];;
let makeC_RelU(u):constr = [RelU(u)];;
let makeC_WfT(sl, t):constr = [WfT(sl, t)];;
let unionC c1 c2 = List.append c1 c2;;
let unionClist = List.fold_left unionC emptyC;;


let emptyTE = [];;

type tenv = (vid * vtype) list;;
let emptyTE:tenv = [];;
let makeTE(vid,t): tenv = [(vid,t)];;

exception Undef;;
exception NonUnifiable of vtype*vtype;;

let emptyC = [];;
let current_tvid = ref 0;;
let new_tvid() = 
    let x = !current_tvid in
    let _ = (current_tvid := !current_tvid+1) in
       x;;
let new_tvar() = 
    TVar(new_tvid());;

let current_svid = ref 0;;
let new_svid() = 
    let x = !current_svid in
    let _ = (current_svid := !current_svid+1) in
       x;;
let new_slev() = 
    SVar(new_svid());;

let current_uvid = ref 0;;
let new_uvid() = 
    let x = !current_uvid in
    let _ = (current_uvid := !current_uvid+1) in
       x;;
let new_uvar() = 
    UVar(new_uvid());;

let current_lvid = ref 0;;
let new_lvid() = 
    let x = !current_lvid in
    let _ = (current_lvid := !current_lvid+1) in
       x;;
let new_level() = 
    LevVar(new_lvid());;

let reset_names() =
  (current_tvid := 0; current_svid := 0; current_uvid := 0; current_lvid := 0);;

let const_table = [("+", (FunT(PairT(IntT(SVar(0)), IntT(SVar(1))),IntT(SVar(2))), 
                      unionC (makeC_SubS(SVar(0),SVar(2))) 
                             (makeC_SubS(SVar(1),SVar(2)))));
                   ("-", (FunT(PairT(IntT(SVar(0)), IntT(SVar(1))),IntT(SVar(2))), 
                      unionC (makeC_SubS(SVar(0),SVar(2))) 
                             (makeC_SubS(SVar(1),SVar(2)))));
                   ("*", (FunT(PairT(IntT(SVar(0)), IntT(SVar(1))),IntT(SVar(2))), 
                      unionC (makeC_SubS(SVar(0),SVar(2))) 
                             (makeC_SubS(SVar(1),SVar(2)))));
                   ("/", (FunT(PairT(IntT(SVar(0)), IntT(SVar(1))),IntT(SVar(2))), 
                      unionC (makeC_SubS(SVar(0),SVar(2))) 
                             (makeC_SubS(SVar(1),SVar(2)))));
                   ("=", (FunT(PairT(IntT(SVar(0)),IntT(SVar(1))),BoolT(SVar(2))),
                      unionC (makeC_SubS(SVar(0),SVar(2))) 
                             (makeC_SubS(SVar(1),SVar(2)))));
                   (">", (FunT(PairT(IntT(SVar(0)),IntT(SVar(1))),BoolT(SVar(2))),
                      unionC (makeC_SubS(SVar(0),SVar(2))) 
                             (makeC_SubS(SVar(1),SVar(2)))));
                   (">=", (FunT(PairT(IntT(SVar(0)),IntT(SVar(1))),BoolT(SVar(2))),
                      unionC (makeC_SubS(SVar(0),SVar(2))) 
                             (makeC_SubS(SVar(1),SVar(2)))));
                   ("<", (FunT(PairT(IntT(SVar(0)),IntT(SVar(1))),BoolT(SVar(2))),
                      unionC (makeC_SubS(SVar(0),SVar(2))) 
                             (makeC_SubS(SVar(1),SVar(2)))));
                   ("<=", (FunT(PairT(IntT(SVar(0)),IntT(SVar(1))),BoolT(SVar(2))),
                      unionC (makeC_SubS(SVar(0),SVar(2))) 
                             (makeC_SubS(SVar(1),SVar(2)))));
                   ("fst",(FunT(PairT(TVar(0),TVar(1)),TVar(0)), makeC_TopT(TVar(1))));
                   ("snd",(FunT(PairT(TVar(0),TVar(1)),TVar(1)), makeC_TopT(TVar(0))));
                   ("nil", (RecT(0, [TVar(0); TVar(1)], SVar(0)), emptyC));
                   ("cons", (FunT(PairT(TVar(0), RecT(0, [TVar(1);TVar(1)], SVar(0))), 
                                  RecT(0, [TVar(0);TVar(1)], SVar(0))),
                             emptyC));
                   ("hd", (FunT(RecT(0, [TVar(0);TVar(1)], SVar(0)), TVar(0)), makeC_TopT(TVar(1))));
                   ("tl", (FunT(RecT(0, [TVar(0);TVar(1)], SVar(0)), RecT(0, [TVar(1);TVar(1)],SVar(0))),
                            makeC_TopT(TVar(0))));
                   ("null", (FunT(RecT(0, [TVar(0);TVar(1)], SVar(0)), BoolT(SVar(0))), 
                             unionC (makeC_TopT(TVar(0))) (makeC_TopT(TVar(1))))) ];;
let lookup_const prname = List.assoc prname const_table;;

(***** For recurisve datatypes *****)
(*** Information about the position in which type parameters appear in a recursive type ***)
type polarity_info = Multiplicative | Strictly_Positive | Positive | Negative | Invariant

let rectype_table: (rectyId * polarity_info list) list ref = 
   ref [(0, [Multiplicative; Multiplicative])
           (*** 0: list ('a, 'b). 'a is the type of the head element, 
                   and 'b is the type of elements of the tail ***)
       ] 

let id_of_RecT t =
   match t with 
     RecT(id,_,_) -> id
   | _ -> raise (SomethingIsWrongIn "id_of_RecT");;

let attributes_of_RecT id =
   List.assoc id (!rectype_table)

(*** returns a new instance of recursive type whose identifier is id ***)
let new_instance_of_RecT id =
   let al = attributes_of_RecT id in
   let tl = List.map (function _ -> new_tvar()) al in
   let sl = new_slev() in
    (tl, sl);;

let is_multiplicative a =
  match a with 
    Multiplicative -> true
  | _ -> false

let is_strictly_positive a =
  match a with
    Multiplicative -> true
  | Strictly_Positive -> true
  | _ -> false

let is_covariant a = 
  match a with
    Multiplicative -> true
  | Strictly_Positive -> true
  | Positive -> true
  | _ -> false

let is_contravariant a = 
  match a with
    Negative -> true
  | _ -> false;;

type constructor_tab = (cid * (vtype * constr)) list
let constructor_table: constructor_tab ref = 
     ref [("nil", (RecT(0, [TVar(0); TVar(1)], SVar(0)), emptyC));
      ("cons", (RecT(0, [TVar(0); TVar(1)], SVar(0)), 
                 unionC (makeC_TopT(TVar(0))) (makeC_TopT(TVar(1)))))];;

let lookup_constructor con = List.assoc con !constructor_table

let rec tyvars_in_t t =
  match t with
    TVar tv -> [tv]
  | UnitT -> []
  | BoolT _ -> []
  | IntT _ -> []
  | ChanT(t1,_,u) -> (tyvars_in_t t1) 
  | ParT(t1,t2) -> merge_and_unify compare (tyvars_in_t t1)
                            (tyvars_in_t t2)
  | PairT(t1,t2) -> merge_and_unify compare (tyvars_in_t t1)
                            (tyvars_in_t t2)
  | FunT(t1,t2) ->  merge_and_unify compare (tyvars_in_t t1)
                            (tyvars_in_t t2)
  | RecT(id,tl, _) -> List.fold_right (fun t -> fun vl -> 
                                        merge_and_unify compare vl (tyvars_in_t t)) tl []
  | IncT(t1) -> (tyvars_in_t t1)
  | RepT(t1) -> (tyvars_in_t t1)
  | UpT(t1,_) -> (tyvars_in_t t1);;

let tyvars_in_te te = 
   List.fold_right (function (x,t)->merge_and_unify compare (tyvars_in_t t)) te [];;
let tyvars_in_tytab = tyvars_in_te;;

(*** Varord keeps the order of variables appearing in a process expression ***)
(*** Intuitively, the order represents which channels have been last created ***)
(*** For example, x?[y].new z in ... introduces the order z < y ***)
type varord = VarOrd of (vid * varord) list
let emptyVord = VarOrd [];;
let rec remove_var_from_vord vid (VarOrd l) =
  match l with
    [] -> VarOrd []
   | (x,VarOrd l1)::l2 ->
       let VarOrd(l2') = remove_var_from_vord vid (VarOrd l2) in
       if x=vid then VarOrd(l1@l2')
       else VarOrd((x,remove_var_from_vord vid (VarOrd l1))::l2')

let extend_vord_with_bv vord vid =  
     let VarOrd(l) = remove_var_from_vord vid vord in
        VarOrd((vid,VarOrd [])::l)
let extend_vord_with_nv vord vid = 
     let vord' = remove_var_from_vord vid vord in VarOrd [(vid,vord')]

let rec vord_member x (VarOrd l) =
  match l with
     [] -> false
   | (y, vord)::l' -> if x=y then true else (vord_member x vord)||(vord_member x (VarOrd l'))

let name_order = ref [];;

let vord_less x y vord =
 let rec vord_less_aux x y (VarOrd l) =
   (match l with
     [] -> false
    | (z, vord)::l' -> if z=x then vord_member y vord 
                      else (vord_less_aux x y (VarOrd l')|| vord_less_aux x y vord))
 in
   (List.mem (y,x) !name_order) || (vord_less_aux x y vord);;

let rec lookup_type x g: vtype * tenv =
  match g with
    [] -> raise Not_found
  | ((y,t)::g') ->
         if x<y then raise Not_found
         else if x=y then (t,g')
         else let (t',g'') = lookup_type x g' in (t', (y,t)::g'');;

let rec par_TElist l: tenv * constr = 
  match l with
    [] -> (emptyTE,emptyC)
  | [g1] -> (g1,emptyC)
  | g1::gl -> 
      let (g2, c1) = par_TElist(gl) in
      let (g, c2) = par_TE(g1,g2) in
        (g, unionC c1 c2)
and par_TE(g1,g2) =
  match (g1,g2) with
    ([],_) -> (g2, emptyC)
  | (_,[]) -> (g1, emptyC)
  | ((x1,t1)::g1', (x2,t2)::g2') ->
         if x1=x2 then
             let (g', c') = par_TE(g1',g2') in
             let t = new_tvar() in
               ((x1,t)::g', unionC (makeC_SubT(t, ParT(t1,t2))) c')
         else if x1<x2 then
             let (g', c) = par_TE(g1',g2) in
               ((x1,t1)::g', c)
         else let (g', c) = par_TE(g1,g2') in
               ((x2,t2)::g', c);;

let rec or_TElist l: tenv * constr = 
  match l with
    [] -> (emptyTE,emptyC)
  | [g1] -> (g1,emptyC)
  | g1::gl -> 
      let (g2, c1) = or_TElist(gl) in
      let (g, c2) = or_TE(g1,g2) in
        (g, unionC c1 c2)
and or_TE(g1,g2) =
  match (g1,g2) with
    ([], []) -> ([], emptyC)
  | ([],(x2,t2)::g2') -> 
          let (g', c) = or_TE(g1,g2') in
	  let t = new_tvar() in
             ((x2,t)::g', unionClist([makeC_TopT(t);makeC_SubT(t,t2);c]))
  | ((x1,t1)::g1', []) -> 
          let (g', c) = or_TE(g1',g2) in
	  let t = new_tvar() in
             ((x1,t)::g', unionClist([makeC_TopT(t);makeC_SubT(t,t1);c]))
  | ((x1,t1)::g1', (x2,t2)::g2') ->
         if x1=x2 then
             let (g', c') = or_TE(g1',g2') in
             let t = new_tvar() in
               ((x1,t)::g', 
                 unionClist([makeC_SubT(t, t1);makeC_SubT(t,t2);c']))
         else if x1<x2 then
             let (g', c) = or_TE(g1',g2) in
	     let t = new_tvar() in
               ((x1,t)::g', unionClist([makeC_TopT(t);makeC_SubT(t,t1);c]))
         else let (g', c) = or_TE(g1,g2') in
	     let t = new_tvar() in
               ((x2,t)::g', unionClist([makeC_TopT(t);makeC_SubT(t,t2);c]));;

let rec rep_type t = 
  match t with 
     ChanT(t,sl,u) -> ChanT(t,sl,RepU(u))
   | TVar(vid) -> RepT(t)
   | RepT(t) -> RepT(t)
   | PairT(t1,t2) -> PairT(rep_type(t1),rep_type(t2))
   | RecT(id,tl, sl) ->
       let al = attributes_of_RecT id in
       let l = List.combine al tl in
       let tl' = List.map (fun (a,t) ->
                            if is_multiplicative(a) then rep_type t else t) l in
           RecT(id, tl',sl)
   | IncT(_) -> RepT(t)
   | UpT(_) -> RepT(t)
   | _ -> t;;
  
let rec rep_TE g = 
  match g with
    [] -> ([], emptyC)
   | (x,t)::g1 ->
        let (g1', c) = rep_TE(g1) in
        let t1 = new_tvar() in
           ((x,t1)::g1', unionC (makeC_SubT(t1,rep_type(t))) c);;

let up_type lev t =
  match t with 
     UnitT -> (UnitT, emptyC)
   | BoolT(_) -> (t, emptyC)
   | IntT(_) -> (t, emptyC)
   | _ -> let t' = new_tvar() in (t', makeC_SubT(t', UpT(t, lev)));;

let up_TE lev vid vord g =
  List.fold_right (fun (x,t) -> fun (g',c') -> 
   	         let (t',c1) = up_type (if (not(!Flag.livelock))&&(vord_less vid x vord)
						     then lev else LevPlus(lev, 1)) t in
                              ((x,t')::g', unionC c1 c')) g (emptyTE, emptyC);;

let rec makeC_reliable_ty t =
  match t with
    ChanT(_,_,u) -> makeC_RelU(u)
  | PairT(t1,t2) -> unionC (makeC_reliable_ty t1) (makeC_reliable_ty t2)
  | RecT(id, tl, sl) -> 
       let al = attributes_of_RecT id in
       let l = List.map is_multiplicative al in
       let tl' = List.map snd (List.filter (fun (b,t)->b) (List.combine l tl)) in
         List.fold_right (fun t -> fun c -> unionC (makeC_reliable_ty t) c) tl' emptyC
  | _ -> emptyC;;
let makeC_reliable_te te =
  unionClist(List.map (function (x,t)->makeC_reliable_ty t) te);;

let emptySubst = [];;
let emptyTSubst = [];;
let makeSubst x t = [(x,t)];;
let mkTypeSubst = makeSubst;;


(*** Rename type variables and secrecy level variables.  ***)
let rec merge_subst s1 s2 =
   match (s1,s2) with
     ([], _) -> s2
   | (_, []) -> s1
   | ((x1,t1)::s1', (x2,t2)::s2') ->
        if x1<x2 then (x1,t1)::(merge_subst s1' s2)
        else if x1=x2 then (x2,t2)::(merge_subst s1' s2')
        else (x2,t2)::(merge_subst s1 s2');;
let rec rename_slevel(sl, ss) =
   match sl with
     SVar(x) ->
         (try (SVar(List.assoc x ss), ss) with
            Not_found ->
	      let y = new_svid() in
                 (SVar(y), merge_subst (makeSubst x y) ss))
   | Sglb(sl1,sl2) ->
         let (sl1',ss') = rename_slevel(sl1,ss) in
         let (sl2',ss'') = rename_slevel(sl2,ss') in
            (Sglb(sl1',sl2'), ss'')
   | _ -> (sl, ss);; (*** It is assumed that sl is not of the form SIfInf(l,sl1) ***)

let rec rename_type(t, s, ss) =
   match t with
     TVar(x) -> (try (TVar(List.assoc x s), s, ss) with
                  Not_found ->
                    let y = new_tvid() in
                      (TVar(y), merge_subst (makeSubst x y) s, ss))
   | BoolT(sl) -> 
         let (sl', ss') = rename_slevel(sl,ss) in
               (BoolT(sl'), s, ss')
   | IntT(sl) -> 
         let (sl', ss') = rename_slevel(sl,ss) in
               (IntT(sl'), s, ss')
   | ChanT(t1, sl, u) -> 
         let (t1', s',ss') = rename_type(t1,s,ss) in
         let (sl', ss'') = rename_slevel(sl, ss') in
             (ChanT(t1', sl', u), s', ss'')
   | PairT(t1,t2) ->
         let (t1', s', ss') = rename_type(t1,s,ss) in
         let (t2', s'',ss'') = rename_type(t2,s',ss') in
             (PairT(t1',t2'), s'',ss'')
   | FunT(t1,t2) ->
         let (t1', s', ss') = rename_type(t1,s,ss) in
         let (t2', s'',ss'') = rename_type(t2,s',ss') in
             (FunT(t1',t2'), s'',ss'')
   | RecT(id, tl, sl) ->
         let (tl', s', ss') = rename_tylist(tl,s,ss) in
         let (sl', ss'') = rename_slevel(sl, ss') in
             (RecT(id, tl', sl'), s',ss'')
   | ParT(t1,t2) ->
         let (t1', s', ss') = rename_type(t1,s,ss) in
         let (t2', s'',ss'') = rename_type(t2,s',ss') in
             (ParT(t1',t2'), s'',ss'')
   | IncT(t1) -> 
         let (t1', s',ss') = rename_type(t1,s,ss) in
             (IncT(t1'), s', ss')
   | UpT(t1,l) -> 
         let (t1', s',ss') = rename_type(t1,s,ss) in
             (UpT(t1',l), s', ss')
   | _ -> (t, s, ss)
and rename_tylist (tl, s, ss) =
  match tl with
     [] -> ([], s, ss)
   | t::tl1 ->
        let (t', s', ss') = rename_type(t, s, ss) in
        let (tl1', s'', ss'') = rename_tylist(tl1, s', ss') in
           (t'::tl1', s'', ss'');;
let rename_atomic_constr(ac,s,ss) = 
  match ac with
    EqT(t1,t2) -> 
         let (t1', s', ss') = rename_type(t1,s,ss) in
         let (t2', s'',ss'') = rename_type(t2,s',ss') in
             (EqT(t1',t2'), s'',ss'')
  | TopT(t1) ->
         let (t1', s', ss') = rename_type(t1,s,ss) in
          (TopT(t1'), s', ss')
  | SubT(t1,t2) ->    
         let (t1', s', ss') = rename_type(t1,s,ss) in
         let (t2', s'',ss'') = rename_type(t2,s',ss') in
             (SubT(t1',t2'), s'',ss'')
  | WfT(sl1,t1) ->
         let (t1', s', ss') = rename_type(t1,s,ss) in
         let (sl1', ss'') = rename_slevel(sl1,ss') in
          (WfT(sl1',t1'), s', ss'')
  | LowT(t1) ->
         let (t1', s', ss') = rename_type(t1,s,ss) in
          (LowT(t1'), s', ss')
  | SubS(sl1,sl2) ->
         let (sl1', ss') = rename_slevel(sl1,ss) in
         let (sl2', ss'') = rename_slevel(sl2,ss') in
             (SubS(sl1',sl2'), s, ss'')
  | SubSIf(l,sl1,sl2) ->
         let (sl1', ss') = rename_slevel(sl1,ss) in
         let (sl2', ss'') = rename_slevel(sl2,ss') in
             (SubSIf(l,sl1',sl2'), s, ss'')
  | EqS(sl1,sl2) ->
         let (sl1', ss') = rename_slevel(sl1,ss) in
         let (sl2', ss'') = rename_slevel(sl2,ss') in
             (EqS(sl1',sl2'), s, ss'')
  | _ -> (*** We assume that SubU, RelU, SubL do not appear in constant tables ***)
        (ac,s,ss);; 

let rec rename_constr(c, s, ss) =
   match c with
      [] -> c
    | ac::c' ->
         let (ac', s',ss') = rename_atomic_constr(ac,s,ss) in
             ac'::(rename_constr(c', s',ss'));;

(*** rename the type of a constant ***)   
let rename_const_type (t, c) =
     let (t', s, ss) = rename_type(t,emptyTSubst,emptySubst) in
     let c' = rename_constr(c, s, ss) in
        (t', c');;

type capmap = (pc * levexp) list;;
  (*** the capability of each input/output primitive ***)
let empty_capmap:capmap = [];;
let ext_capmap pc c cm =
     (pc,c)::cm;;
let merge_capmap cm1 cm2 = cm1@cm2;;
let global_capmap = ref empty_capmap;;

(*** type_table is the type of a table associating variables 
 ***  in the source programs with their types 
 ***)
type type_table = (pc * vtype) list;;
let empty_tytab:type_table = [];;
let ext_tytab pc t tytab: type_table =
     (pc,t)::tytab;;
let merge_tytab tytab1 tytab2: type_table = tytab1@tytab2;;

(*** incT is the same as the constructor IncT when the flag livelock is true ***)
let incT t = if !Flag.livelock then IncT(t) else t;;

(*** "nocap" predicates (used for weak lock-freedom analysis) ***)
let nocapT t = [NocapT(t)]
let nocapTE g =
  unionClist(List.map (fun (x,t) -> nocapT(t)) g);;

(*** ifloPT takes a process and outputs a principal typing. ***)
(*** iflowPT: procexp -> varord -> tenv * slevel * constr * capmap * type_table ***)
(*** slevel is the secrecy level of information about whether the process has been executed ***)
(*** capmap records the capability of each input/output and is used to report 
     the result of deadlock/livelock analysis ***)
(*** type_table records the type of certain subexpressions, 
     which is used to perform useless-code elimination ***)
let rec iflowPT p vord: tenv * slevel * constr * capmap * type_table = 
   match p with
     Zero -> 
       let sl = new_slev() in
         (emptyTE, sl, emptyC, empty_capmap,empty_tytab)
   | Out(x,v1,p1, pc) ->
       let (g1, t1, c1, tytab1) = iflowPTv(v1) in
       let (g2, sl2, c2, cm2,tytab2) = iflowPT p1 vord in
       let (g3, c3) = par_TE(g1,g2) in
       let t2 = new_tvar() in
       let uv = new_uvid() in
       let u = UVar(uv) in
       let sl3 = new_slev() in
       let (g4, c4) = 
          (try 
             let (t3, g) = lookup_type x g3 in
             let c = makeC_EqT(ChanT(t2,sl3,u),t3) in
                 (g, c)
           with 
             Not_found -> 
              (g3, makeC_SubU(uv,ZeroU))) in
       let l1 = LevConst(0) in
       let l2 = new_level() in
       let (g5, c5) =
           up_TE l2 x vord g4 in
       let sl = new_slev() in
       let uv4 = new_uvid() in
       let t4 = ChanT(t2,sl3,UVar(uv4)) in
       let (g,c6) = par_TE(makeTE(x,t4), g5) in
       let c7 = makeC_SubS(sl,sl2) in
       let c8 = makeC_SubS(sl,sl3) in
       let c9 = makeC_SubSIf(l2,sl3,sl2) in 
       let c10 = makeC_SubL(LevIfInfty(sl3),l2) in
       let c11 = makeC_SubU(uv4,OutU(l1,l2,u)) in
       let c = unionClist([makeC_SubT(t1,incT(t2));c1;c2;c3;c4;c5;c6;c7;c8;c9;c10;c11]) in
          (g, sl, c, ext_capmap pc l2 cm2, 
           merge_tytab tytab1 (ext_tytab pc t4 tytab2))
   | In(x, y,_,p1, pc) ->
       let (g1, sl1, c1, cm1, tytab1) = iflowPT p1 (extend_vord_with_bv vord y) in
       let ((t1,g2), c1') = 
           try (lookup_type y g1, emptyC) with
                      Not_found -> 
                         let t = new_tvar() in
                             ((t,g1), makeC_TopT(t)) in
       let t2 = new_tvar() in
       let c2 = makeC_SubT(t2, t1) in
       let uv = new_uvid() in
       let u = UVar(uv) in
       let sl3 = new_slev() in
       let (g4, c4) = 
          (try 
             let (t3, g) = lookup_type x g2 in
             let c = makeC_EqT(ChanT(t2,sl3,u),t3) in
                 (g, c)
           with 
             Not_found -> 
              (g2, makeC_SubU(uv,ZeroU))) in
       let l1 = LevConst(0) in
       let l2 = new_level() in
       let (g5, c5) =
           up_TE l2 x vord g4 in
       let sl = new_slev() in
       let uv4 = new_uvid() in
       let t4 = ChanT(t2,sl3,UVar(uv4)) in
       let (g,c6) = par_TE(makeTE(x,t4), g5) in
       let c7 = makeC_SubS(sl,sl1) in
       let c8 = makeC_SubS(sl,sl3) in
       let c9 = makeC_SubSIf(l2,sl3,sl1) in 
       let c10 = makeC_SubL(LevIfInfty(sl3),l2) in
       let c11 = makeC_SubU(uv4,InU(l1,l2,u)) in
       let c = unionClist([c1;c1';c2;c4;c5;c6;c7;c8;c9;c10;c11]) in
          (g, sl, c, ext_capmap pc l2 cm1, ext_tytab pc t4 tytab1)
   | Rep(In(x, y,_,p1, pc)) -> (*** for replicated inputs ***)
       let (g1, sl1, c1, cm1, tytab1) = iflowPT p1 (extend_vord_with_bv vord y) in
       let ((t1,g2), c1') = 
           try (lookup_type y g1, emptyC) with
                      Not_found -> 
                         let t = new_tvar() in
                             ((t,g1), makeC_TopT(t)) in
       let t2 = new_tvar() in
       let c2 = makeC_SubT(t2, t1) in
       let uv = new_uvid() in
       let u = UVar(uv) in
       let sl3 = new_slev() in
       let (g4, c4) = 
          (try 
             let (t3, g) = lookup_type x g2 in
             let c = makeC_EqT(ChanT(t2,sl3,u),t3) in
                 (g, c)
           with 
             Not_found -> 
              (g2, makeC_SubU(uv,ZeroU))) in
       let l1 = LevConst(0) in
       let l2 = new_level() in
       let (g5, c5) =
           up_TE l2 x vord g4 in
       let l3 = new_level() in
       let l4 = new_level() in
       let (g6, c6) =
           up_TE l4 x vord g4 in
       let (g7, c7) = rep_TE(g6) in
       let (g8, c8) = par_TE(g5, g7) in
       let sl = new_slev() in
       let uv4 = new_uvid() in
       let t4 = ChanT(t2,sl3,UVar(uv4)) in
       let (g,c9) = par_TE(makeTE(x,t4), g8) in
       let c10 = makeC_SubS(sl,sl1) in
       let c11 = makeC_SubS(sl,sl3) in
       let c12 = makeC_SubSIf(l4,sl3,sl1) in 
       let c13 = makeC_SubL(LevIfInfty(sl3),l2) in
       let c14 = makeC_SubU(uv4,InU(l1,l2,ParU(u,RepU(InU(l1,l3,u))))) in
       let c15 = makeC_SubL(l2, l4) in
       let c16 = makeC_SubL(l3, l4) in
       let c = unionClist([c1;c1';c2;c4;c5;c6;c7;c8;c9;c10;c11;c12;c13;c14;c15;c16]) in
          (g, sl, c, ext_capmap pc l4 cm1, ext_tytab pc t4 tytab1)
   | Par(p1,p2) ->
       let (g1, sl1, c1, cm1,tytab1) = iflowPT p1 vord in
       let (g2, sl2, c2, cm2,tytab2) = iflowPT p2 vord in
       let (g, c3) = par_TE(g1, g2) in
       let sl = new_slev() in
       let c = unionClist([makeC_SubS(sl,sl1);makeC_SubS(sl,sl2);c1;c2;c3]) in
         (g, sl, c, merge_capmap cm1 cm2, merge_tytab tytab1 tytab2)
   | Rep(p1) ->
       let (g1, sl1, c1, cm1,tytab1) = iflowPT p1 vord in
       let (g, c2) = rep_TE(g1) in
          (g, sl1, unionClist([c1;c2]), cm1, tytab1)
   | Nu(x, _,p1, pc) ->
       let (g1, sl1, c1, cm1, tytab1) = iflowPT p1 (extend_vord_with_nv vord x) in
       let (t1,g) = try lookup_type x g1 with
                      Not_found -> (new_tvar(),g1) in
       let t2 = new_tvar() in
       let uv = new_uvid() in
       let u = UVar(uv) in
       let sl2 = new_slev() in
       let c2 = makeC_EqT(ChanT(t2,sl2,u), t1) in
       let c3 = makeC_RelU(u) in
       let c4 = makeC_WfT(sl2, t2) in
       (*** 2007/9 ***)
       let _ = if !Flag.wlauto then reg_TEtable pc g else () in
          (g, sl1, unionClist([c1;c2;c3;c4]), cm1, ext_tytab pc t1 tytab1)
   | If(v, p1, p2, pc) ->
       let (g1, sl1, c1, cm1, tytab1) = iflowPT p1 vord in
       let (g2, sl2, c2, cm2, tytab2) = iflowPT p2 vord in
       let (g3, t3, c3, tytab3) = iflowPTv v in
       let (g4, c4) = or_TE(g1, g2) in
       let (g, c5) = par_TE(g3,g4) in
       let sl3 = new_slev() in
       let sl = new_slev() in
       let c = unionClist([makeC_EqT(BoolT(sl3), t3);
			   makeC_SubS(sl3,sl1);makeC_SubS(sl3,sl2);
			   makeC_SubS(sl,sl1);makeC_SubS(sl,sl2);c1;c2;c3;c4;c5]) in
         (g, sl, c, merge_capmap cm1 cm2, 
          ext_tytab pc t3 (merge_tytab tytab1 (merge_tytab tytab2 tytab3)))
   | MatchL(x, ml, pc) ->
       let (g, id, sll, c1, cm, tytab) = iflowPTmatch x ml vord in
       let (t,_) = lookup_type x g in
       let (tl,sl1) = new_instance_of_RecT(id) in
       let c2 = List.map (fun sl -> SubS(sl1,sl)) sll in
       let sl2 = new_slev() in  (*** the security level of the entire process ***)
       let c3 = List.map (fun sl -> SubS(sl2, sl)) sll in
       
         (g, sl2, unionClist([makeC_EqT(RecT(id,tl,sl1),t);c1;c2;c3]), cm, ext_tytab pc t tytab)
   | Let(x, _,v, p1, pc) ->
       let (g1, sl1, c1, cm1, tytab1) = iflowPT p1 (extend_vord_with_bv vord x) in
       let ((t1,g2),c2) = 
                  try (lookup_type x g1, emptyC) with
                      Not_found -> 
                         let t = new_tvar() in
                             ((t,g1), makeC_TopT(t)) in
       let (g3, t3, c3, tytab3) = iflowPTv v in
       let (g, c4) = par_TE(g3, g2) in
       let c = unionClist([makeC_SubT(t3,t1);c1;c2;c3;c4]) in
           (g, sl1, c, cm1, merge_tytab tytab3 (ext_tytab pc t1 tytab1))
   | Special(p1) ->  
       if !Flag.wlock && !Flag.livelock
       then 
         let _ =  (Flag.livelock := false) in
         let (g1, sl1, c1, cm1, tytab1) = iflowPT p1 vord in
         let c2 = nocapTE(g1) in
         (***  TODO: add nocap constraint ***)
         let _ = (Flag.livelock := true) in
           (g1, sl1, unionClist[c1;c2], cm1, tytab1)
       else
         iflowPT p1 vord
and iflowPTmatch x ml vord =
(*** a function for handling a list of matches (ml). ml should not be null ***)
(*** x is the variable being matched ***)
(*** should return (g, sll, c, cm, tytab) ***)
(***  g is the type environment (containing the binding on x) ***)
(***  sll is the list of security levels of matches ***)
  match ml with
    [(lab, p)] -> 
       let (g1, sl1, c1, cm1, tytab1) = iflowPT p vord in
       let (t, c) = lookup_constructor(lab) in
       let (t2, c2) = rename_const_type(t,c) in
       let id = id_of_RecT(t) in
       let (g3, c3) = par_TE(g1, makeTE(x,t2)) in
         (g3, id, [sl1], unionClist([c1;c2;c3]), cm1, tytab1)
  | (lab,p)::ml' ->
       let (g1, sl1, c1, cm1, tytab1) = iflowPT p vord in
       let (t2, c2) = rename_const_type(lookup_constructor(lab)) in
       let (g3, c3) = par_TE(g1, makeTE(x,t2)) in
       let (g4, id4, sll4, c4, cm4, tytab4) = iflowPTmatch x ml' vord in
       let (g5, c5) = or_TE(g3, g4) in
         (g5, id4, sl1::sll4, unionClist([c1;c2;c3;c4;c5]), 
          merge_capmap cm1 cm4, merge_tytab tytab1 tytab4)
   | _ -> raise (SomethingIsWrongIn "iflowPTmatch")
and iflowPTv(v) =
   match v with
     Unit -> (emptyTE, UnitT, emptyC, empty_tytab)
   | Bool(b) -> 
        let sl = new_slev() in
           (emptyTE,BoolT(sl),emptyC, empty_tytab)
   | SBool(b, slva) -> 
        let sl = match slva with
                     Hi -> HI
                   | Low -> LOW in
           (emptyTE,BoolT(sl),emptyC, empty_tytab)
   | Int(n) -> 
        let sl = new_slev() in
           (emptyTE,IntT(sl),emptyC, empty_tytab)
   | SInt(b, slva) -> 
        let sl = match slva with
                     Hi -> HI
                   | Low -> LOW in
           (emptyTE,IntT(sl),emptyC, empty_tytab)
   | Var(x) -> 
        let t = new_tvar() in (makeTE(x,t), t, emptyC, empty_tytab)
   | Prim(name,_) ->
         let (t,c) = rename_const_type(lookup_const(name)) in 
            (emptyTE, t, c, empty_tytab)
   | App(v1,v2, pc) ->
        let (g1,t1,c1,tytab1)=iflowPTv(v1) in
        let (g2,t2,c2,tytab2)=iflowPTv(v2) in
        let (g, c3) = par_TE(g1,g2) in
        let t3 = new_tvar() in
        let c4 = makeC_EqT(t1, FunT(t2,t3)) in
           (g, t3, unionClist([c1;c2;c3;c4]),
            merge_tytab (ext_tytab pc t2 tytab1) tytab2)
   | Pair(v1,v2) ->
        let (g1,t1,c1,tytab1)=iflowPTv(v1) in
        let (g2,t2,c2,tytab2)=iflowPTv(v2) in
        let (g, c3) = par_TE(g1,g2) in
           (g, PairT(t1,t2), unionClist([c1;c2;c3]), merge_tytab tytab1 tytab2);;

(*** take a list of constraints and separate it into 
 *** subtype constraints and the other constraints ***)
(*** let rec take_TyC (c:constr) = 
 ***  match c with
 ***     [] -> ([], [])
 ***  | EqT(t1,t2)::c' -> 
 ***        let (tyC, non_tyC) = take_TyC c' in
 ***            ((t1,t2)::tyC, non_tyC)
 ***   | SubT(t1,t2)::c' -> 
 ***        let (tyC, non_tyC) = take_TyC c' in
 ***            ((t1,t2)::tyC, non_tyC)
 ***   | ac::c' ->
 ***        let (tyC, non_tyC) = take_TyC c' in
 ***            (tyC, ac::non_tyC);;
 ***)


(*** Functions to reduce constraints on types
 *** TODO: recursive types, replacement with a more efficient algorithm
 ***)
let empty_weak_unif_constraints = [];;
let rec mk_atype_pairs (tl:vtype list) =
   match tl with
      [] -> []
    | [t1] -> []
    | [t1;t2] -> [(t1,t2)]
    | t1::t2::tl' -> (t1,t2)::mk_atype_pairs(t2::tl');;
let rec atomic_types t =
   match t with
      ParT(t1,t2) -> (atomic_types t1)@(atomic_types t2)
    | IncT(t') -> atomic_types t'
    | RepT(t') -> atomic_types t'
    | UpT(t',_) ->  atomic_types t'
    | _ -> [t];;

let mk_weak_unif_constraints (t1,t2) =
   mk_atype_pairs ((atomic_types t1)@(atomic_types t2));;
let rec take_weak_unif_constraints (c:constr) = 
   match c with
     [] -> empty_weak_unif_constraints
   | EqT(t1,t2)::c' -> 
        (mk_weak_unif_constraints (t1,t2))@(take_weak_unif_constraints c')
   | SubT(t1,t2)::c' -> 
        (mk_weak_unif_constraints (t1,t2))@(take_weak_unif_constraints c')
   | ac::c' -> 
       (*** Types may appear in WfT(t), LowT(t), and TopT(t), but they are always single
	*** type variables by the definition of iflowPT
        ***)
         take_weak_unif_constraints c';;

let make_tvid_pair m n = 
  if m<=n then (m,n) else (n,m);;

exception Failure_of_Weak_Unification of vtype*vtype;;

type ty_substitution = (tvid * vtype) list;;
let empty_tsubst:ty_substitution = [];;
let rec apply_tsubst_type (s:ty_substitution) t =
  match t with
     TVar(m) -> (try List.assoc m s with Not_found -> TVar(m))
   | ChanT(t1,sl,u) -> ChanT(apply_tsubst_type s t1, sl, u)
   | PairT(t1,t2) -> PairT(apply_tsubst_type s t1,apply_tsubst_type s t2)
   | FunT(t1,t2) -> FunT(apply_tsubst_type s t1,apply_tsubst_type s t2)
   | RecT(id, tl, sl) -> RecT(id, List.map (apply_tsubst_type s) tl, sl)
   | ParT(t1,t2) -> ParT(apply_tsubst_type s t1,apply_tsubst_type s t2)
   | IncT(t1) -> IncT(apply_tsubst_type s t1)
   | RepT(t1) -> RepT(apply_tsubst_type s t1)
   | UpT(t1,l) -> UpT(apply_tsubst_type s t1,l)
   | _ -> t
let apply_tsubst_tpl s = 
  List.map (function (t1,t2)->(apply_tsubst_type s t1, apply_tsubst_type s t2));;
let comp_tsubst s1 s2 =
  let s2' = List.map (function (tvid,t)->(tvid, apply_tsubst_type s1 t)) s2 in
  let s1' = List.filter (function (tvid,t) -> 
                           List.for_all (function (tvid',t')->tvid!=tvid') s2) s1 in
   s1'@s2';;
let rec apply_tsubst_tvpl s tvpl =
  match tvpl with
     [] -> ([], [])
   | (m,n)::tvpl' ->
       let (tpl1,tvpl1) = apply_tsubst_tvpl s tvpl' in

       let t1 = apply_tsubst_type s (TVar m) in 
       let t2 = apply_tsubst_type s (TVar n) in 
         (match (t1,t2) with
           (TVar(m'),TVar(n')) -> (tpl1,(m',n')::tvpl1)
          | _ -> ((t1,t2)::tpl1, tvpl1));;

let apply_tsubst_te s = List.map(function (vid,t)->(vid,apply_tsubst_type s t));;
let apply_tsubst_tytab s = List.map(function (pc,t)->(pc,apply_tsubst_type s t));;
let rec apply_tsubst_aconstr (s:ty_substitution) ac =
  match ac with
    EqT(t1,t2) -> EqT(apply_tsubst_type s t1, apply_tsubst_type s t2)
  | TopT(t) ->  TopT(apply_tsubst_type s t)
  | SubT(t1,t2) -> SubT(apply_tsubst_type s t1, apply_tsubst_type s t2)
  | WfT(sl,t) ->  WfT(sl, apply_tsubst_type s t)
  | LowT(t) ->  LowT(apply_tsubst_type s t)
  | NocapT(t) -> NocapT(apply_tsubst_type s t)
  | _ -> (*** NOTE: if a future extension allows other constraints to contain types,
          *** this case should be revised
          ***)
       ac;;

let apply_tsubst_constr s =
   List.map (apply_tsubst_aconstr s);; 

type sl_substitution = (slvid * slevel) list;;
type u_substitution = (uvid * usage) list;;
type l_substitution = (lvid * levexp) list;;
type substitution = ty_substitution * sl_substitution * u_substitution * l_substitution;;
let empty_subst:substitution = ([], [], [], []);;
let tsubst2subst t_s = (t_s, [], [], []);;
let slsubst2subst sl_s = ([], sl_s, [], []);;
let usubst2subst u_s = ([], [], u_s, []);;
let lsubst2subst l_s = ([], [], [], l_s);;
let subst2tsubst (t_s,_,_,_)=t_s;;
let subst2slsubst (_,sl_s,_,_)=sl_s;;
let subst2usubst (_,_,u_s,_)=u_s;;
let subst2lsubst (_,_,_,l_s)=l_s;;
let make_slsubst sv sl = slsubst2subst [(sv,sl)];;

exception Test of atomic_constraint list;;

let sglb (sl1,sl2) =
  match (sl1,sl2) with
      (LOW, _) -> LOW
    | (_, LOW) -> LOW
    | (HI, _) -> sl2
    | (_, HI) -> sl1
    | _ -> Sglb(sl1,sl2);;
let sIfInf(l, sl) =
  match l with
    LevInfty -> sl
  | LevConst _ -> HI
  | _ -> SIfInf(l,sl);;

exception UnExpectedLevExp of levexp;;


(*** This will also be used in reductions to Petri nets, so that
 *** we must ensure that lexp_max on a finite set of levels do not
 *** generate an infinite number of lexp ***)
let lexp_max lexp1 lexp2 =
  if lexp1=lexp2 then lexp1 
  else
    match (lexp1, lexp2) with
     (LevInfty, _) -> LevInfty
   | (_, LevInfty) -> LevInfty
   | (LevConst(n), LevConst(m)) -> LevConst(max m n)
   | (LevConst(0), _) -> lexp2
   | (_, LevConst(0)) -> lexp1
   | (LevMax(lexpl1), LevMax(lexpl2)) -> LevMax(merge_and_unify compare lexpl1 lexpl2)
   | (LevMax(lexpl1), _) -> LevMax(merge_and_unify compare lexpl1 [lexp2])
   | (_, LevMax(lexpl2)) -> LevMax(merge_and_unify compare [lexp1] lexpl2)
   | _ -> LevMax(merge_and_unify compare [lexp1] [lexp2]);;

let lexp_min lexp1 lexp2 =
  if lexp1=lexp2 then lexp1 
  else
    match (lexp1, lexp2) with
     (LevInfty, _) -> lexp2
   | (_, LevInfty) -> lexp1
   | (LevConst(n), LevConst(m)) -> LevConst(min m n)
   | (LevConst(0), _) -> LevConst(0)
   | (_, LevConst(0)) -> LevConst(0)
   | _ -> LevMin([lexp1;lexp2]);;

let rec levmax ll =
  match ll with
     [] -> LevConst(0)
   | [l] -> l
   | l::ll' -> lexp_max l (levmax ll')

let rec levmin ll =
  match ll with
     [] -> LevInfty
   | [l] -> l
   | l::ll' -> lexp_min l (levmin ll')

let rec apply_subst_lev (s:substitution) l:levexp =
  match l with
    LevVar(x) -> (try List.assoc x (subst2lsubst s) with Not_found -> l)
  | LevIfInfty(sl) -> LevIfInfty(apply_subst_sl s sl)
  | LevPlus(l',n) -> LevPlus(apply_subst_lev s l', n)
  | LevMax(ll) -> LevMax(List.map (apply_subst_lev s) ll)
  | LevMin(ll) -> LevMin(List.map (apply_subst_lev s) ll)
  | _ -> l
and apply_subst_sl (s:substitution) sl =
  match sl with
     SVar(x) -> (try List.assoc x (subst2slsubst s) with Not_found -> sl)
   | Sglb(sl1,sl2) -> Sglb(apply_subst_sl s sl1, apply_subst_sl s sl2)
   | SIfInf(l, sl') -> 
            SIfInf(apply_subst_lev s l, apply_subst_sl s sl')
   | _ -> sl;;

exception Failure_subst_uvar of usage;;
let apply_subst_uvar (s:substitution) uv =
  try 
     let u = List.assoc uv (subst2usubst s) in
       (match u with
          UVar(x) -> x
        | _ -> raise (Failure_subst_uvar u))
  with Not_found -> uv;;
  

let rec apply_subst_usage (s:substitution) u: usage =
  match u with
    UVar(x) -> (try List.assoc x (subst2usubst s) with Not_found -> UVar(x))
  | ZeroU -> ZeroU
  | InU(l1,l2,u') -> InU(apply_subst_lev s l1,apply_subst_lev s l2,apply_subst_usage s u')
  | OutU(l1,l2,u') -> OutU(apply_subst_lev s l1,apply_subst_lev s l2,apply_subst_usage s u')
  | ParU(u1,u2) -> ParU(apply_subst_usage s u1,apply_subst_usage s u2)
  | OrU(u1,u2) -> OrU(apply_subst_usage s u1,apply_subst_usage s u2)
  | RepU(u') -> RepU(apply_subst_usage s u')
  | IncU(u') -> IncU(apply_subst_usage s u')
  | UpU(u',l1,l2) -> UpU(apply_subst_usage s u',l1,l2)
  | URec(x,u') -> URec(x,apply_subst_usage 
                         (usubst2subst (List.filter (function (y,u1)->x!=y) (subst2usubst s)))
                          u');;

let apply_subst_uc (s:substitution) uc =
  List.map (fun (uv,u) -> (uv, apply_subst_usage s u)) uc;;

let rec apply_subst_type (s:substitution) t =
  match t with
     TVar(m) -> (try List.assoc m (subst2tsubst s) with Not_found -> TVar(m))
   | UnitT -> UnitT
   | BoolT(sl) -> BoolT(apply_subst_sl s sl)
   | IntT(sl) -> IntT(apply_subst_sl s sl)
   | ChanT(t1,sl,u) -> ChanT(apply_subst_type s t1, apply_subst_sl s sl, apply_subst_usage s u)
   | PairT(t1,t2) -> PairT(apply_subst_type s t1,apply_subst_type s t2)
   | RecT(id,tl, sl) -> RecT(id, List.map (apply_subst_type s) tl, apply_subst_sl s sl)
   | FunT(t1,t2) -> FunT(apply_subst_type s t1,apply_subst_type s t2)
   | ParT(t1,t2) -> ParT(apply_subst_type s t1,apply_subst_type s t2)
   | IncT(t1) -> IncT(apply_subst_type s t1)
   | RepT(t1) -> RepT(apply_subst_type s t1)
   | UpT(t1,l) -> UpT(apply_subst_type s t1,l)

let apply_subst_ac (s:substitution) ac =
  match ac with
    EqT(t1,t2) -> EqT(apply_subst_type s t1, apply_subst_type s t2)
  | TopT(t) -> TopT(apply_subst_type s t)
  | WfT(sl,t) -> WfT(apply_subst_sl s sl, apply_subst_type s t)
  | LowT(t) -> LowT(apply_subst_type s t)
  | NocapT(t) -> NocapT(apply_subst_type s t)
  | SubT(t1,t2)-> SubT(apply_subst_type s t1, apply_subst_type s t2)
  | SubU(uv,u)-> SubU(apply_subst_uvar s uv, apply_subst_usage s u)
  | RelU(u) -> RelU(apply_subst_usage s u)
  | SubL(l1,l2)-> SubL(apply_subst_lev s l1,apply_subst_lev s l2)
  | SubS(sl1,sl2) -> SubS(apply_subst_sl s sl1,apply_subst_sl s sl2)
  | SubSIf(l,sl1,sl2)-> SubSIf(apply_subst_lev s l, apply_subst_sl s sl1,apply_subst_sl s sl2)
  | EqS(sl1,sl2) -> EqS(apply_subst_sl s sl1,apply_subst_sl s sl2);;

let apply_subst_c s = List.map (apply_subst_ac s);;
let apply_subst_te s = List.map (function (x,t)-> (x,apply_subst_type s t));;
let apply_subst_tytab s = List.map (function (pc,t)-> (pc,apply_subst_type s t));;
let comp_subst s1 s2 =
  let t_s2 = List.map (function (tvid,t) -> (tvid, apply_subst_type s1 t)) (subst2tsubst s2) in
  let t_s1 = List.filter (function (tvid,t)->
                (List.for_all (function (tvid',t')->tvid!=tvid') t_s2)) (subst2tsubst s1) in
  let sl_s2 = List.map (function (slid,sl)->(slid, apply_subst_sl s1 sl)) (subst2slsubst s2) in
  let sl_s1 = List.filter (function (slid,_)->
                (List.for_all (function (slid',_)->slid!=slid') sl_s2)) (subst2slsubst s1) in
  let u_s2 = List.map (function (uvid,u)->(uvid, apply_subst_usage s1 u)) (subst2usubst s2) in
  let u_s1 = List.filter (function (uvid,_)->
                (List.for_all (function (uvid',_)->uvid!=uvid') u_s2)) (subst2usubst s1) in
  let l_s2 = List.map (function (lvid,l)->(lvid, apply_subst_lev s1 l)) (subst2lsubst s2) in
  let l_s1 = List.filter (function (lvid,_)->
                (List.for_all (function (lvid',_)->lvid!=lvid') l_s2)) (subst2lsubst s1) in
    (t_s1@t_s2, sl_s1@sl_s2, u_s1@u_s2, l_s1@l_s2);;

let rec weak_unif_list (tpl, eqrels) =
  match tpl with
    [] -> eqrels
  | (TVar(m1),TVar(m2))::tpl' -> (
         if m1=m2 then weak_unif_list(tpl', eqrels)
         else
           let (class1, eqrels1) = Eqclass.find_and_add m1 eqrels in
           let (class2, eqrels2) = Eqclass.find_and_add m2 eqrels1 in
             if Eqclass.eqclass class1 class2 then weak_unif_list(tpl', eqrels2)
             else 
               let r1 = Eqclass.representative(class1) in
               let r2 = Eqclass.representative(class2) in
               let eqrels3  = Eqclass.merge class1 class2 eqrels2 in
                  match (r1,r2) with
                    (Some t1, Some t2) -> weak_unif_list ((t1,t2)::tpl', eqrels3)
                   | _ -> weak_unif_list (tpl', eqrels3)
         )
  | (TVar(m1), t2)::tpl' -> (
       let (class1, eqrels1) = Eqclass.find_and_add m1 eqrels in
       let r1 = Eqclass.representative(class1) in
         match r1 with
             None -> let eqrels2 = Eqclass.add_representative t2 class1 eqrels1 in
                         weak_unif_list (tpl', eqrels2)
           | Some(t1) ->
                weak_unif_list ((t1,t2)::tpl', eqrels1)
       )
  | (t1, TVar(m2))::tpl' ->
        weak_unif_list((TVar(m2),t1)::tpl',eqrels)
  | (UnitT, UnitT)::tpl' -> weak_unif_list (tpl', eqrels)
  | (BoolT _, BoolT _)::tpl' -> weak_unif_list (tpl', eqrels)
  | (IntT _, IntT _)::tpl' -> weak_unif_list (tpl', eqrels)
  | (ChanT(t1,_,_), ChanT(t2,_,_))::tpl' -> weak_unif_list ((t1,t2)::tpl', eqrels)
  | (PairT(t11,t12), PairT(t21,t22))::tpl' ->
         weak_unif_list ((t11,t21)::(t12,t22)::tpl', eqrels)
  | (FunT(t11,t12), FunT(t21,t22))::tpl' ->
         weak_unif_list ((t11,t21)::(t12,t22)::tpl', eqrels)
  | (RecT(id1, tl1, _) as t1, (RecT(id2, tl2, _) as t2))::tpl' ->
         if id1=id2 then 
            weak_unif_list((List.combine tl1 tl2)@tpl', eqrels)
         else
            raise (Failure_of_Weak_Unification (t1,t2))
  | (t1,t2)::_ -> raise (Failure_of_Weak_Unification (t1,t2))

        
exception SLevel_violation;;
let eq_sl(sl1,sl2) = 
     if sl1=sl2 then slsubst2subst []
     else 
       (match (sl1,sl2) with
            (SVar x, _) -> slsubst2subst [(x,sl2)]
          | (_, SVar x) -> slsubst2subst [(x,sl1)]
          | _ -> raise SLevel_violation);;
exception EqU_failure of usage*usage;;
let eq_u(u1,u2) =
  if u1=u2 then usubst2subst []
  else
    match (u1,u2) with
      (UVar x, _) -> usubst2subst [(x,u2)]
    | (_, UVar x) -> usubst2subst [(x,u1)]
    | _ -> raise (EqU_failure (u1, u2));;
exception Failure_EqT of vtype*vtype;;
exception Failure_SubT of vtype*vtype;;
let rec elim_ty_constraints_aux (c, tc, others, tveq, uveq)
      : constr * constr * (tvid * tvid) list * (uvid * uvid) list =
  match c with
     [] -> (tc, others, tveq, uveq)
   | EqT(t1,t2)::c' ->
      (match (t1,t2) with
          (TVar m, TVar n) ->
          if m=n then elim_ty_constraints_aux (c', tc, others, tveq, uveq)
          else elim_ty_constraints_aux(c', tc, others, (m,n)::tveq, uveq)
      | (UnitT,UnitT) ->
          elim_ty_constraints_aux (c', tc, others, tveq, uveq)
      | (BoolT(sl1),BoolT(sl2))-> 
          if sl1=sl2 then elim_ty_constraints_aux (c', tc, others, tveq, uveq)
          else elim_ty_constraints_aux(c', tc, EqS(sl1,sl2)::others, tveq, uveq)
      | (IntT(sl1),IntT(sl2)) -> 
          if sl1=sl2 then elim_ty_constraints_aux (c', tc, others, tveq, uveq)
          else elim_ty_constraints_aux(c', tc, EqS(sl1,sl2)::others, tveq, uveq)
      | (ChanT(t1',sl1,UVar(uv1)), ChanT(t2',sl2,UVar(uv2))) ->
          elim_ty_constraints_aux(EqT(t1',t2')::c', 
                                  tc, EqS(sl1,sl2)::others, tveq, (uv1,uv2)::uveq)
      | (PairT(t11,t12), PairT(t21,t22)) ->
          elim_ty_constraints_aux (EqT(t11,t21)::EqT(t12,t22)::c', tc, others, tveq, uveq)
      | (FunT(t11,t12), FunT(t21,t22)) ->
          elim_ty_constraints_aux (EqT(t11,t21)::EqT(t12,t22)::c', tc, others, tveq, uveq)
      | (RecT(id1, tl1, sl1), RecT(id2, tl2, sl2)) ->
          if id1=id2 then
             let c1 = List.map (function (t1,t2)->EqT(t1,t2)) (List.combine tl1 tl2) in
                if sl1=sl2 then 
                     elim_ty_constraints_aux (c1@c', tc, others, tveq, uveq)
                else
                      elim_ty_constraints_aux (c1@c', tc, EqS(sl1,sl2)::others, tveq, uveq)
          else
             raise (Failure_EqT (t1,t2))
      | _ -> raise (Failure_EqT (t1,t2)))
  | TopT(t)::c' ->
     ( match t with
         UnitT -> elim_ty_constraints_aux (c', tc, others, tveq, uveq)
       | (BoolT _) -> elim_ty_constraints_aux (c', tc, others, tveq, uveq)
       | (IntT _) -> elim_ty_constraints_aux (c', tc, others, tveq, uveq)
       | ChanT(t1,sl1,UVar(uid)) -> 
                   elim_ty_constraints_aux (c', tc, SubU(uid,ZeroU)::others, tveq, uveq)
       | PairT(t1,t2) -> 
              elim_ty_constraints_aux (TopT(t1)::TopT(t2)::c', tc, others, tveq, uveq)
       | FunT(t1,t2) -> elim_ty_constraints_aux (c', tc, others, tveq, uveq)
       | RecT(id,tl,_) -> 
            let al = attributes_of_RecT id in
            let l = List.map is_multiplicative al in
            let tl' = List.map snd (List.filter (fun (b,t)->b) (List.combine l tl)) in 
	               (*** tl' <-- a subset of types in tl that are multiplicative ***)
            let c1 = List.map (function t->TopT(t)) tl' in
              elim_ty_constraints_aux (c1@c', tc, others, tveq, uveq)
       | _ -> elim_ty_constraints_aux (c', TopT(t)::tc, others, tveq, uveq)
      )
  | WfT(sl, t)::c' ->
     (match t with 
         UnitT -> elim_ty_constraints_aux (c', tc, others, tveq, uveq)
       | (BoolT sl1) -> 
             if sl=LOW then elim_ty_constraints_aux (c', tc, others, tveq, uveq)
             else elim_ty_constraints_aux (c', tc, SubS(sl,sl1)::others, tveq, uveq)
       | (IntT sl1) -> 
             if sl=LOW then elim_ty_constraints_aux (c', tc, others, tveq, uveq)
             else elim_ty_constraints_aux (c', tc, SubS(sl,sl1)::others, tveq, uveq)
       | ChanT(t1,sl1,u) -> 
             let c1 = if sl=LOW then [] else makeC_SubS(sl,sl1) in
               elim_ty_constraints_aux(WfT(sl1,t1)::c', tc, c1@others, tveq, uveq)
       | PairT(t1,t2) -> 
               elim_ty_constraints_aux (WfT(sl,t1)::WfT(sl,t2)::c', tc, others, tveq, uveq)
       | FunT(t1,t2) -> 
               elim_ty_constraints_aux (WfT(sl,t2)::c', tc, others, tveq, uveq)
       | RecT(id,tl, sl1) -> 
            let al = attributes_of_RecT id in
            let l = List.map is_strictly_positive al in
            let tl' = List.map snd (List.filter (fun (b,t)->b) (List.combine l tl)) in 
              (*** tl' <-- a subset of tl that occurs in strictly positive positions ***)
            let c1 = List.map (function t->WfT(sl,t)) tl' in
               if sl=LOW then 
                  elim_ty_constraints_aux (c1@c', tc, others, tveq, uveq)
               else elim_ty_constraints_aux (c1@c', tc, SubS(sl,sl1)::others, tveq, uveq)
       | _ -> elim_ty_constraints_aux (c', WfT(sl,t)::tc, others, tveq, uveq)
      )
  | LowT(t)::c' ->
     (match t with 
         UnitT -> elim_ty_constraints_aux (c', tc, others, tveq, uveq)
       | TVar tv -> elim_ty_constraints_aux (c', LowT(t)::tc, others, tveq, uveq)
       | (BoolT sl1) -> 
             elim_ty_constraints_aux (c', tc, SubS(sl1,LOW)::others, tveq, uveq)
       | (IntT sl1) -> 
             elim_ty_constraints_aux (c', tc, SubS(sl1,LOW)::others, tveq, uveq)
       | ChanT(t1,sl1,u) -> 
             elim_ty_constraints_aux (LowT(t1)::c', tc, SubS(sl1,LOW)::others, tveq, uveq)
       | PairT(t1,t2) -> 
            elim_ty_constraints_aux (LowT(t1)::LowT(t2)::c', tc, others, tveq, uveq)
       | FunT(t1,t2) -> 
            elim_ty_constraints_aux (LowT(t1)::LowT(t2)::c', tc, others, tveq, uveq)
       | RecT(id,tl, sl1) -> 
            let c1 = List.map (fun t -> LowT(t)) tl in
                elim_ty_constraints_aux (c1@c', tc, SubS(sl1,LOW)::others, tveq, uveq)
       | ParT(t1,t2) -> 
            elim_ty_constraints_aux (LowT(t1)::LowT(t2)::c', tc, others, tveq, uveq)
       | IncT(t1) -> 
            elim_ty_constraints_aux (LowT(t1)::c', tc, others, tveq, uveq)
       | RepT(t1) -> 
            elim_ty_constraints_aux (LowT(t1)::c', tc, others, tveq, uveq)
       | UpT(t1,l) -> 
            elim_ty_constraints_aux (LowT(t1)::c', tc, others, tveq, uveq)
      )
  | SubT(t1,t2)::c'->
     ( match (t1,t2) with
        (TVar(_), t) -> elim_ty_constraints_aux (c', SubT(t1,t2)::tc, others, tveq, uveq)
      | (UnitT, UnitT) -> elim_ty_constraints_aux (c', tc, others, tveq, uveq)
      | (BoolT(sl1),BoolT(sl2)) -> 
                       elim_ty_constraints_aux (c', tc, SubS(sl1,sl2)::others, tveq, uveq)
      | (IntT(sl1),IntT(sl2)) -> 
                       elim_ty_constraints_aux (c', tc, SubS(sl1,sl2)::others, tveq, uveq)
      | (ChanT(t1',sl1,UVar(uid1)),ChanT(t2',sl2,u2)) -> 
                 elim_ty_constraints_aux (EqT(t1',t2')::c', tc, 
                                          EqS(sl1,sl2)::SubU(uid1,u2)::others, tveq, uveq)
      | (PairT(t11,t12),PairT(t21,t22)) ->
            elim_ty_constraints_aux (SubT(t11,t21)::SubT(t12,t22)::c', tc, others, tveq, uveq)
      | (FunT(t11,t12),FunT(t21,t22)) ->
            elim_ty_constraints_aux (SubT(t21,t11)::SubT(t12,t22)::c', tc, others, tveq, uveq)
      | (RecT(id1,tl1,sl1) as t1, (RecT(id2, tl2,sl2) as t2)) ->
            if id1=id2 then
               let al = attributes_of_RecT id1 in
               let l = List.combine al (List.combine tl1 tl2) in
               let c1 = List.map
                          (fun (a, (t1', t2')) ->
                             if is_covariant(a) then SubT(t1',t2')
                             else if is_contravariant(a) then SubT(t2',t1')
                             else EqT(t1',t2'))
                         l in
                  elim_ty_constraints_aux (c1@c', tc, SubS(sl1,sl2)::others, tveq, uveq)
            else
               raise (Failure_SubT(t1,t2))
      | _ -> raise (Failure_SubT (t1,t2))
     )
  | ac::c' -> elim_ty_constraints_aux (c', tc, ac::others, tveq, uveq)

let eqclass2subst vl =
  match vl with
    [] -> []
  | v::vl' -> List.map (function v' -> (v',v)) vl';;

let eqs2subst eqs =
  let eqrel = Scc.compute_scc (eqs@(List.map (function (x,y)->(y,x)) eqs)) in
   List.fold_right (function vl -> function s -> (eqclass2subst vl)@s) eqrel [];;

let elim_ty_constraints c =
  let (tc1, c1, tveq, uveq) = elim_ty_constraints_aux (c, [], [], [], []) in
  let ts1 = List.map (function (x,y) -> (x,TVar y)) (eqs2subst tveq) in
  let us1 = List.map (function (x,y) -> (x, UVar y)) (eqs2subst uveq) in
  let s = comp_subst (tsubst2subst ts1) (usubst2subst us1) in
  let tc2 = apply_subst_c s tc1 in
  let c2 = apply_subst_c s c1 in
    (tc2, c2, s);;

let non_chan_type t =
  match t with
     UnitT -> true
   | BoolT _ -> true
   | IntT _ -> true
   | FunT _ -> true
   | _ -> false;;

let rec normalize_type t =
   match t with
     TVar(x) -> ([], t)
   | UnitT -> ([], UnitT)
   | BoolT(_)->([], t)
   | IntT(_)->([], t)
   | ChanT(t1,sl,u) ->
         let (c,t1')=normalize_type t1 in
         (c, ChanT(t1', sl, u))
   | PairT(t1,t2) ->
         let (c1,t1')=normalize_type t1 in 
         let (c2,t2')=normalize_type t2 in
         (c1@c2, PairT(t1', t2'))
   | FunT(t1,t2) ->
         let (c1,t1')=normalize_type t1 in
         let (c2,t2')=normalize_type t2 in
         (c1@c2, FunT(t1', t2'))
   | RecT(id, tl,sl) ->
         let (c1,tl')=normalize_tylist tl in
         (c1, RecT(id, tl',sl))
   | ParT(t1,t2) ->
      let (c1,t1') = normalize_type t1 in
      let (c2,t2') = normalize_type t2 in
        (match (t1',t2') with
           (ChanT(t11,sl1,u1),ChanT(t12,sl2,u2)) ->
               (EqT(t11,t12)::EqS(sl1,sl2)::(c1@c2), ChanT(t11,sl1,ParU(u1,u2)))
         | (UnitT, UnitT) -> ([], UnitT)
         | (BoolT(sl1),BoolT(sl2))-> ([EqS(sl1,sl2)], BoolT(sl1))
         | (IntT(sl1),IntT(sl2))-> ([EqS(sl1,sl2)], IntT(sl1))
         | (PairT(t11,t12),PairT(t21,t22)) ->
             let (c3,t1') = normalize_type (ParT(t11,t21)) in
             let (c4,t2') = normalize_type (ParT(t12,t22)) in
                (c1@c2@c3@c4, PairT(t1',t2'))
         | (FunT(t11,t12),FunT(t21,t22)) ->
            ([EqT(t11,t21);EqT(t12,t22)], FunT(t11,t12))
         | (RecT(id1,tl1, sl1),RecT(id2, tl2, sl2)) ->
             if id1=id2 then
               let al = attributes_of_RecT id1 in
               let l = List.combine al (List.combine tl1 tl2) in
               let ctl = List.map
                          (function (a, (t1', t2')) ->
                             if is_multiplicative(a) then 
                                normalize_type (ParT(t1', t2'))
                             else ([EqT(t1',t2')], t1'))
                         l in
               let c3 = List.fold_right (@) (List.map fst ctl) [] in
               let tl = List.map snd ctl in
                 (unionClist([[EqS(sl1,sl2)];c1;c2;c3]), RecT(id1, tl, sl1))
             else (c1@c2, ParT(t1', t2'))
         | _ -> (c1@c2, ParT(t1',t2')))
   | IncT(t1) ->  
      let (c1,t1') = normalize_type t1 in
        if non_chan_type(t1') then (c1, t1')
        else 
         (match t1' with
            ChanT(t11,sl1,u1) -> (c1, ChanT(t11,sl1,IncU(u1)))
          | PairT(t11,t12) -> 
              let (c2, t11') = normalize_type (IncT(t11)) in
              let (c3, t12') = normalize_type (IncT(t12)) in
                (c1@c2@c3, PairT(t11',t12'))
          | RecT(id1, tl, sl) ->
               let al = attributes_of_RecT id1 in
               let l = List.combine al tl in
               let ctl = List.map
                          (function (a, t) ->
                             if is_multiplicative(a) then 
                                normalize_type (IncT(t))
                             else (emptyC, t))
                         l in
               let c2 = List.fold_right (@) (List.map fst ctl) [] in
               let tl' = List.map snd ctl in
                  (c1@c2, RecT(id1,tl', sl))
          | _ -> (c1, IncT(t1')))
   | RepT(t1) -> 
      let (c1,t1') = normalize_type t1 in
        if non_chan_type(t1') then (c1, t1')
        else 
         (match t1' with
            ChanT(t11,sl1,u1) -> (c1, ChanT(t11,sl1,RepU(u1)))
          | PairT(t11,t12) -> 
              let (c2, t11') = normalize_type (RepT(t11)) in
              let (c3, t12') = normalize_type (RepT(t12)) in
                (c1@c2@c3, PairT(t11',t12'))
          | RecT(id1, tl, sl) ->
               let al = attributes_of_RecT id1 in
               let l = List.combine al tl in
               let ctl = List.map
                          (function (a, t) ->
                             if is_multiplicative(a) then 
                                normalize_type (RepT(t))
                             else (emptyC, t))
                         l in
               let c2 = List.fold_right (@) (List.map fst ctl) [] in
               let tl' = List.map snd ctl in
                  (c1@c2, RecT(id1, tl', sl))
          | _ -> (c1, RepT(t1')))
   | UpT(t1,l) -> 
      let (c1,t1') = normalize_type t1 in
        if non_chan_type(t1') then (c1, t1')
        else 
         (match t1' with
            ChanT(t11,sl1,u1) -> (c1, ChanT(t11,sl1,UpU(u1,l, l)))
          | PairT(t11,t12) -> 
              let (c2, t11') = normalize_type (UpT(t11,l)) in
              let (c3, t12') = normalize_type (UpT(t12,l)) in
                (c1@c2@c3, PairT(t11',t12'))
          | RecT(id1, tl, sl) ->
               let al = attributes_of_RecT id1 in
               let atl = List.combine al tl in
               let ctl = List.map
                          (function (a, t) ->
                             if is_multiplicative(a) then 
                                normalize_type (UpT(t, l))
                             else (emptyC, t))
                         atl in
               let c2 = List.fold_right (@) (List.map fst ctl) [] in
               let tl' = List.map snd ctl in
                  (c1@c2, RecT(id1, tl', sl))
          | _ -> (c1, UpT(t1',l)))
and normalize_tylist tl =
  match tl with
    [] -> (emptyC, [])
  | t::tl' ->
      let (c1,t') = normalize_type t in
      let (c2, tl'') = normalize_tylist tl' in
        (c1@c2, t'::tl'');;
   
let normalize_ty_constraint ac =
  match ac with
    EqT(t1,t2) ->
      let (c1,t1')=normalize_type(t1) in
      let (c2,t2')=normalize_type(t2) in
        EqT(t1',t2')::(c1@c2)
  | TopT(t) ->
      let (c1,t1)=normalize_type(t) in 
         TopT(t1)::c1
  | SubT(t1,t2)->
      let (c1,t1')=normalize_type(t1) in
      let (c2,t2')=normalize_type(t2) in
        SubT(t1',t2')::(c1@c2)
  | WfT(sl, t) ->
      let (c1,t1)=normalize_type(t) in 
         WfT(sl, t1)::c1
  | LowT(t) ->
      let (c1,t1)=normalize_type(t) in 
         LowT(t1)::c1
  | NocapT(t) ->
      let (c1,t1)=normalize_type(t) in 
         NocapT(t1)::c1
  | _ -> [ac];;
let normalize_ty_constraints c=
  List.fold_right (function ac ->function c -> (normalize_ty_constraint ac)@c) c [];;

let rec normalize_te te =
  match te with
    [] -> ([], emptyC)
  | (x,t)::te1 ->
        let (c1, te1') = normalize_te te1 in
        let (c2, t') = normalize_type t in
          (c2@c1, (x,t')::te1');;
let normalize_tytab = normalize_te;;   

let rec order_of_eqclasses elements classes =
  if Util.IntSet.is_empty elements then []
  else
    let n1 = Util.IntSet.choose elements in
    let cl1 = Eqclass.find n1 classes in
    let r1 = Eqclass.representative cl1 in
    let elements1 = Eqclass.elements_of_class cl1 in
      match r1 with
         None -> order_of_eqclasses (Util.IntSet.diff elements elements1) classes
       | Some(t) ->
           let tvars = tyvars_in_t t in
           let tv1 = Eqclass.choose_from_class cl1 in
           let rel = List.fold_right
                         (function tv -> function rel ->
                            try
			       merge_and_unify compare 
			       [(tv1, Eqclass.choose_from_class (Eqclass.find tv classes))]
                               rel
                            with Not_found -> rel) 
                        tvars [] in
             merge_and_unify compare rel 
                      (order_of_eqclasses (Util.IntSet.diff elements elements1) classes);; 

let check_acyclicity classes =
  let elements = Eqclass.sets_of_elements classes in
  let ord_eqclasses = order_of_eqclasses elements classes in
    Scc.check_acyclicity(ord_eqclasses);;

let rec make_type_from_eqrel tv classes =
  try
    let cl1 = Eqclass.find tv classes in
    let r1 = Eqclass.representative cl1 in
      match r1 with
        None -> new_tvar()
      | Some(t) -> make_type_from_eqrel_aux t classes
  with Not_found ->
    new_tvar()
and make_type_from_eqrel_aux t classes =
  match t with
    TVar(tv) -> make_type_from_eqrel tv classes
  | UnitT -> t
  | BoolT _-> BoolT(new_slev())
  | IntT(_) -> IntT(new_slev())
  | ChanT(t1,sl,u) -> ChanT(make_type_from_eqrel_aux t1 classes, new_slev(), new_uvar())
  | PairT(t1,t2) -> PairT(make_type_from_eqrel_aux t1 classes,make_type_from_eqrel_aux t2 classes)
  | FunT(t1,t2) -> FunT(make_type_from_eqrel_aux t1 classes,make_type_from_eqrel_aux t2 classes)
  | RecT(id, tl, sl) -> RecT(id, List.map (fun t -> make_type_from_eqrel_aux t classes) tl, new_slev())
  | _ -> raise (SomethingIsWrongIn "make_type_from_eqrel_aux");;

let make_tsubst_from_eqrel classes =
  let elements = Eqclass.sets_of_elements classes in
    Util.IntSet.fold (function tv -> 
               merge_and_unify compare [(tv, make_type_from_eqrel tv classes)]) elements [];;

let reduce_ty_constraints (te, sl, c, tytab) =
  let _ = if !Flag.debug then Printf.printf "Start reduction of type constraints: %f\n" (Sys.time()) in
  let tpl = take_weak_unif_constraints(c) in
  let _ = if !Flag.debug then 
          (Printf.printf "Reduction of type constraints--phase 2: %f\n" (Sys.time()); flush(stdout)) in
(***  let (tvpl,ts1) = weak_unif_list(tpl,Eqclass.empty) in ***)
  let classes = weak_unif_list(tpl,Eqclass.empty) in
  let _ = if !Flag.debug then 
            (Printf.printf "Reduction of type constraints--phase 3: %f\n" (Sys.time());flush stdout) in
  let _ = check_acyclicity(classes) in
  let ts1 = make_tsubst_from_eqrel classes in
  let te1 = apply_tsubst_te ts1 te in
  let _ = if !Flag.debug 
          then (Printf.printf "Reduction of type constraints--phase 4: %f\n" (Sys.time());flush stdout) in
  let (c11, te2) = normalize_te te1 in
  let _ = if !Flag.debug 
          then (Printf.printf "Reduction of type constraints--phase 5: %f\n" (Sys.time());flush stdout) in
  let tytab1 = apply_tsubst_tytab ts1 tytab in
  let _ = if !Flag.debug then Printf.printf "Reduction of type constraints--phase 6: %f\n" (Sys.time()) in
  let (c12, tytab2) = normalize_tytab tytab1 in
  let _ = if !Flag.debug then Printf.printf "Reduction of type constraints--phase 7: %f\n" (Sys.time()) in
  let c2 = apply_tsubst_constr ts1 c in
  let _ = if !Flag.debug then Printf.printf "Reduction of type constraints--phase 8: %f\n" (Sys.time()) in
  let c3 = normalize_ty_constraints (c11@c12@c2) in
  let _ = if !Flag.debug then Printf.printf "Reduction of type constraints--phase 9: %f\n" (Sys.time()) in
  let (tc, c4, s2) = elim_ty_constraints c3 in
  let _ = if !Flag.debug then Printf.printf "Reduction of type constraints--phase 10: %f\n" (Sys.time()) in
    (apply_subst_te s2 te2,apply_subst_sl s2 sl,c4,tc, 
     apply_subst_tytab s2 tytab2, comp_subst s2 (tsubst2subst ts1));;

let eliminate_type_constraints (te, sl, c, tytab) =
   let _ = if !Flag.debug then 
            Printf.printf "Start elimination of type constraints: %f\n" (Sys.time()) in
   let c1 = unionClist(List.map (function (x,t)->makeC_WfT(LOW,t)) te) in
      (*** for each type in the type environment, add the well-formedness conditions on types ***)
   let c2 = unionC c c1 in
   let _ = if !Flag.debug then 
            Printf.printf "Elimination of type constraints: phase 2: %f\n" (Sys.time()) in
   let (te', sl', c3, tc, tytab, s) = reduce_ty_constraints(te,sl,c2, tytab) in
   let _ = if !Flag.debug then 
            Printf.printf "Elimination of type constraints: phase 3: %f\n" (Sys.time()) in
   let c4 = if !Flag.free_channels then
                makeC_reliable_te te'
            else emptyC in
   let _ = if !Flag.debug then 
            Printf.printf "Elimination of type constraints finished: %f\n" (Sys.time()) in
     (te', sl', c4@c3, tc, tytab, s);;
   
  
let rec uv_occur_in_usage uv u =
  match u with
    UVar(uv') -> uv=uv'
   | ZeroU -> false
   | InU(_,_,u') -> uv_occur_in_usage uv u'
   | OutU(_,_,u') -> uv_occur_in_usage uv u'
   | ParU(u1,u2) -> (uv_occur_in_usage uv u1)||(uv_occur_in_usage uv u2)
   | OrU(u1,u2) -> (uv_occur_in_usage uv u1)||(uv_occur_in_usage uv u2)
   | RepU(u') -> uv_occur_in_usage uv u'
   | IncU(u') -> uv_occur_in_usage uv u'
   | UpU(u',_,_) -> uv_occur_in_usage uv u'
   | URec(uv',u') -> (not(uv=uv')) && (uv_occur_in_usage uv u')

(***  
let rec solve_subusage_constraints c rc s =
  match c with
    [] -> (rc, s)
  | SubU(uv,u)::c' ->
       if uv_occur_in_usage uv u then
          solve_subusage_constraints c' ((SubU(uv,u))::rc) s
       else
          let s1 = usubst2subst ([(uv,u)]) in
          let c'' = apply_subst_c s1 c' in
          let rc' = apply_subst_c s1 rc in
          let s2 = comp_subst s1 s in
            solve_subusage_constraints c'' rc' s2
  | ac::c' -> raise (UnExpectedConstraint ac)
***)
let rec solve_subusage_constraints c rc s =
  match c with
    [] -> (apply_subst_c s rc, s)
  | SubU(uv,u)::c' ->
       let u' = apply_subst_usage s u in
       if uv_occur_in_usage uv u' then
          solve_subusage_constraints c' ((SubU(uv,u'))::rc) s
       else
          let s1 = usubst2subst ([(uv,u')]) in
          let s2 = comp_subst s1 s in
            solve_subusage_constraints c' rc s2
  | ac::c' -> raise (UnExpectedConstraint ac)

let get_subusage_constraint c= 
   List.fold_right 
    (function ac -> function (c1,c2)-> 
        match ac with SubU _ -> (ac::c1, c2) | _ -> (c1,ac::c2))
    c 
    ([], []);;

let comp_subusage_constraint ac1 ac2 =
      match (ac1,ac2) with
        (SubU(uv1,_), SubU(uv2,_)) -> compare uv1 uv2
       | _ -> raise (UnExpectedConstraint ac1);;
let sort_subusage_constraint = 
   List.sort comp_subusage_constraint;;
   

let rec merge_subusage_constraints_on_same_var c =
   match c with
      [] -> []
    | SubU(uv1,u1)::c1 ->   
        let c1' = merge_subusage_constraints_on_same_var c1 in
          (match c1' with
             [] -> c
           | (SubU(uv2,u2)::c2) ->
                if uv1=uv2 then SubU(uv1, OrU(u1,u2))::c2 else SubU(uv1,u1)::c1'
           | ac::_ -> raise (UnExpectedConstraint ac))
   | ac::_ -> raise (UnExpectedConstraint ac);;

let left_vars_in_subusage_constraints =
   List.map (function ac -> match ac with (SubU(uv,_)) -> uv | _ -> raise (UnExpectedConstraint ac));;
let rec uvars_in_usage u =
  match u with
    UVar(uv) -> [uv]
  | ZeroU  -> []
  | InU(_,_,u1) -> uvars_in_usage u1
  | OutU(_,_,u1) -> uvars_in_usage u1
  | ParU(u1,u2) -> List.merge compare (uvars_in_usage u1) (uvars_in_usage u2) 
  | OrU(u1,u2) -> List.merge compare (uvars_in_usage u1) (uvars_in_usage u2) 
  | RepU(u1) -> uvars_in_usage u1
  | IncU(u1) -> uvars_in_usage u1
  | UpU(u1,_,_) -> uvars_in_usage u1
  | _ -> raise (UnExpectedUsage u) (*** For recursive usage, which actually never appear ***);;
let uvars_in_ty =
 type_fold_usage uvars_in_usage (merge_and_unify compare) [];;

let uvars_in_te te =
  List.fold_right (function (x,t)->merge_and_unify compare (uvars_in_ty t)) te [];;
let uvars_in_ac ac =
  match ac with
    EqT(t1,t2) -> merge_and_unify compare (uvars_in_ty t1)  (uvars_in_ty t2) 
  | TopT(t1) -> uvars_in_ty t1
  | SubT(t1,t2) -> merge_and_unify compare (uvars_in_ty t1)  (uvars_in_ty t2) 
  | WfT(sl1,t1) -> uvars_in_ty t1
  | LowT(t1) -> uvars_in_ty t1
  | NocapT(t1) -> uvars_in_ty t1
  | SubU(uv,u1) -> merge_and_unify compare [uv] (uvars_in_usage u1)
  | RelU(u1) -> uvars_in_usage u1
  | SubL(l1,l2) -> []
  | SubS(sl1,sl2) -> []
  | SubSIf _ -> []
  | EqS _ -> [];;
let uvars_in_c c =
  List.fold_right (function ac -> merge_and_unify compare (uvars_in_ac ac)) c [];;

let diff_vars vs1 vs2 = 
    List.filter (function v -> not (List.mem v vs2)) vs1;;
let right_vars_in_subusage_constraints c =
  List.fold_right 
     (function ac -> 
          merge_and_unify compare
           (match ac with (SubU(_,u)) -> uvars_in_usage(u) | _ -> raise (UnExpectedConstraint ac)))
     c 
     [];;

let add_undefined_uvars_in_subusage_constraints c uvs= 
  let def_uvs = left_vars_in_subusage_constraints c in
  let undef_uvs = diff_vars uvs def_uvs in
  let c1 = List.map (function uv -> SubU(uv, UVar(uv))) undef_uvs in
    List.merge comp_subusage_constraint c c1;;

let normalize_subusage_constraint c uvs =
  let c1 = sort_subusage_constraint c in
  let c2 = merge_subusage_constraints_on_same_var c1 in
    add_undefined_uvars_in_subusage_constraints c2 uvs;;

let solve_SubU (te, c) =
   let uvs = merge_and_unify compare (uvars_in_te te) (uvars_in_c c) in
   let (c1,c2) = get_subusage_constraint c in
   let c1' = normalize_subusage_constraint c1 uvs in
   let (uc, s) = solve_subusage_constraints c1' [] empty_subst in
     (apply_subst_te s te, uc, apply_subst_c s c2, s);;
   
(*** Removing IncU(U) ***)
type lev_equation = (int * levexp) list;;


let lev2levexp l = l;;

let rec get_levexp u mode =
  match u with
     ZeroU -> LevInfty
  |  UVar(uv) -> LevRec(uv)
  | InU(l1, l2, u) -> 
        if mode then lev2levexp(l1) else LevInfty
  | OutU(l1, l2, u) -> 
        if mode then LevInfty else lev2levexp(l1)
  | ParU(u1,u2) ->
       lexp_min (get_levexp u1 mode) (get_levexp u2 mode)
  | OrU(u1,u2) ->
       lexp_max (get_levexp u1 mode) (get_levexp u2 mode)
  | RepU(u1) ->
      get_levexp u1 mode
  | IncU(u1)->
      LevPlus(get_levexp u1 mode, 1)
  | UpU(u1, l1, l2) ->
       if mode then 
         LevMax([get_levexp u1 mode;lev2levexp(l1)])
       else 
         LevMax([get_levexp u1 mode;lev2levexp(l2)])
  | URec _ -> raise (UnExpectedUsage u);;
(*** get level equations from subusage constraints ***)
(*** It is assumed that the input subusage constraints are normalized ***)
let rec get_lev_equation (c:constr) (mode:bool):lev_equation  =
  match c with
    [] -> []
  | (SubU(uv, u))::c' -> 
        let lexp = get_levexp u mode in
        let leq = get_lev_equation c' mode in
          (uv, lexp)::leq
  | ac::c' -> raise (UnExpectedConstraint ac)

let rec uv_occur_in_lexp uv lexp =
  match lexp with
    LevRec uv' -> uv=uv'
  | LevConst(n) -> false
  | LevInfty -> false
  | LevIfInfty(_) -> false
  | LevVar(_)->false
  | LevPlus(lexp',_) -> uv_occur_in_lexp uv lexp'
  | LevMax(lexpl) -> List.fold_right (function lexp' -> (||)(uv_occur_in_lexp uv lexp')) lexpl false
  | LevMin(lexpl) -> List.fold_right (function lexp' -> (||)(uv_occur_in_lexp uv lexp')) lexpl false;;

let rec subst_lexp_lexp uv lexp lexp' =
  match lexp' with
    LevRec uv' -> if uv=uv' then lexp else lexp'
  | LevConst(n) -> lexp'
  | LevInfty -> lexp'
  | LevIfInfty(_) -> lexp'
  | LevVar(_)-> lexp'
  | LevPlus(lexp1,n) -> LevPlus(subst_lexp_lexp uv lexp lexp1, n)
  | LevMax(lexpl) -> LevMax (List.map (subst_lexp_lexp uv lexp) lexpl)
  | LevMin(lexpl) -> LevMin (List.map (subst_lexp_lexp uv lexp) lexpl);;
  
let subst_eqs uv lexp =
  List.map (function (uv', lexp') -> (uv', subst_lexp_lexp uv lexp lexp'));;

let add_new_lexp_subst uv lexp subst =
   (uv,lexp)::(subst_eqs uv lexp subst);;



let eliminate_duplication_lexpl (l:levexp list) = delete_duplication_unsorted l;;

(*** levPlus is semantically the same as the constructor LevPlus, except that
 *** levPlus outputs a normal form, where LevPlus is applied only to atomic level expressions
 ***)
let rec levPlus n lexp =
  match lexp with
    LevConst(m) -> LevConst(m+n)
  | LevInfty -> LevInfty
  | LevPlus(l1,m) -> LevPlus(l1, m+n)
  | LevMax(ll) -> levmax(List.map (levPlus n) ll)
  | LevMin(ll) -> levmin(List.map (levPlus n) ll)
  | _ -> LevPlus(lexp,n);;

let rec replace_rec_lv_with_inf uv lexp =
  match lexp with
     LevRec uv' -> if uv=uv' then LevInfty else lexp
   | LevConst(n) -> lexp
   | LevInfty -> lexp
   | LevVar(lv') -> lexp
   | LevIfInfty(sl) -> lexp  (*** We assume that sl does not contain level expressions ***)
   | LevPlus(l1,n) -> levPlus n (replace_rec_lv_with_inf uv l1)
   | LevMax(ll) -> levmax (List.map (replace_rec_lv_with_inf uv) ll)
   | LevMin(ll) -> levmin (List.map (replace_rec_lv_with_inf uv) ll);;

let rec solve_rec_leveq uv lexp = 
  match lexp with
     LevRec uv' -> if uv=uv' then LevConst(0) else lexp
   | LevConst(n) -> lexp
   | LevInfty -> LevInfty
   | LevVar(lv') -> lexp
   | LevIfInfty(sl) -> lexp  (*** We assume that sl does not contain level expressions ***)
   | LevPlus(l1,n) -> if n=0 then solve_rec_leveq uv l1
                      else levPlus n (replace_rec_lv_with_inf uv l1)
   | LevMax(ll) -> levmax (List.map (solve_rec_leveq uv) ll)
   | LevMin(ll) -> levmin (List.map (solve_rec_leveq uv) ll);;

let rec solve_lev_equations eqs subst =
  match eqs with
    [] -> subst
  | (uv, lexp)::eqs' ->
      let lexp' = 
        if uv_occur_in_lexp uv lexp then
           solve_rec_leveq uv lexp
        else lexp in 
      let eqs'' = subst_eqs uv lexp' eqs' in
      let subst' = add_new_lexp_subst uv lexp' subst in
        solve_lev_equations eqs'' subst';;

let rec get_levels u eI eO =
  match u with
    UVar(uv) -> (List.assoc uv eI, List.assoc uv eO)
  | ZeroU -> (LevInfty, LevInfty)
  | InU(l1,l2,u1) -> (l1, LevInfty)
  | OutU(l1,l2,u1) -> (LevInfty, l1)
  | ParU(u1,u2) -> 
      let (l11, l12) = get_levels u1 eI eO in
      let (l21, l22) = get_levels u2 eI eO in
        (lexp_min l11 l21, lexp_min l12 l22)
  | OrU(u1,u2) -> 
      let (l11, l12) = get_levels u1 eI eO in
      let (l21, l22) = get_levels u2 eI eO in
        (lexp_max l11 l21, lexp_max l12 l22)
  | RepU(u1) -> get_levels u1 eI eO
  | IncU(u1) -> 
      let (l1, l2) = get_levels u1 eI eO in
      let l1' = levPlus 1 l1 in
      let l2' = levPlus 1 l2 in
        (l1',l2')
  | UpU(u1, l1,l2) ->
      let (l1', l2') = get_levels u1 eI eO in
         (lexp_max l1 l1', lexp_max l2 l2')
  | _ -> raise (UnExpectedUsage u);;


let rec normalizing_UpU (u,l1,l2) =
  match u with
     UVar(uv) -> UpU(u,l1,l2)
   | ZeroU -> ZeroU
   | InU(l1',l2',u1) ->
         InU(lexp_max l1' l1,l2',u1)
   | OutU(l1',l2',u1) ->
         OutU(lexp_max l1' l2,l2', u1)
   | ParU(u1,u2) ->
         ParU(normalizing_UpU(u1, l1,l2),normalizing_UpU(u2, l1,l2))
   | OrU(u1,u2) ->
         OrU(normalizing_UpU(u1, l1,l2),normalizing_UpU(u2, l1,l2))
   | RepU(u1) ->
         RepU(normalizing_UpU(u1, l1, l2))
   | UpU(u1,l1',l2') ->
         UpU(u1, lexp_max l1 l1', lexp_max l2 l2')
   | _ -> raise (UnExpectedUsage u);;
let rec eliminate_IncU_in_u u eI eO =
  match u with
     UVar(uv) -> u
   | ZeroU -> ZeroU
   | InU(l1,l2,u1) ->
         InU(l1,l2,eliminate_IncU_in_u u1 eI eO)
   | OutU(l1,l2,u1) ->
         OutU(l1,l2,eliminate_IncU_in_u u1 eI eO)
   | ParU(u1,u2) ->
         ParU(eliminate_IncU_in_u u1 eI eO,eliminate_IncU_in_u u2 eI eO)
   | OrU(u1,u2) ->
         OrU(eliminate_IncU_in_u u1 eI eO,eliminate_IncU_in_u u2 eI eO)
   | RepU(u1) ->
         RepU(eliminate_IncU_in_u u1 eI eO)
   | IncU(u1) ->
         let u1' = eliminate_IncU_in_u u1 eI eO in
         let (l1,l2) = get_levels u1' eI eO in
         let l1' = levPlus 1 l1 in
         let l2' = levPlus 1 l2 in
           normalizing_UpU(u1', l1',l2')
   | UpU(u1,l1,l2) ->
         normalizing_UpU(eliminate_IncU_in_u u1 eI eO,l1,l2) 
   | _ -> raise (UnExpectedUsage u);;
let rec eliminate_IncU_in_ty t eI eO =
  match t with
     TVar(_) -> t
   | UnitT -> t
   | BoolT(_) -> t
   | IntT(_) -> t
   | ChanT(t1,sl,u) -> 
       ChanT(eliminate_IncU_in_ty t1 eI eO, sl, eliminate_IncU_in_u u eI eO)
   | PairT(t1,t2) ->
       PairT(eliminate_IncU_in_ty t1 eI eO,eliminate_IncU_in_ty t2 eI eO)
   | FunT(t1,t2) ->
       FunT(eliminate_IncU_in_ty t1 eI eO,eliminate_IncU_in_ty t2 eI eO)
   | RecT(id, tl, sl) ->
       RecT(id, List.map (fun t-> eliminate_IncU_in_ty t eI eO) tl, sl)
   | _ -> (*** In the other cases, t is a type expression consisting of type variables, 
               ParT, IncT, etc. ***)
       t;; 
let eliminate_IncU_in_te te eI eO = 
   List.map (function (x,t)->(x,eliminate_IncU_in_ty t eI eO)) te;;
let eliminate_IncU_in_tytab tytab eI eO = 
   List.map (function (pc,t)->(pc,eliminate_IncU_in_ty t eI eO)) tytab;;

let eliminate_IncU_in_ac ac eI eO = 
  match ac with
    SubU(uv,u) -> SubU(uv,eliminate_IncU_in_u u eI eO)
  | RelU(u) -> RelU(eliminate_IncU_in_u u eI eO)
  | _ -> ac;;

let eliminate_IncU_in_c c eI eO = 
   List.map (function ac -> (eliminate_IncU_in_ac ac eI eO)) c;;

let rec find_subuc uv uc =
  match uc with
    [] -> raise Not_found
  | SubU(uv1,u)::uc' ->
        if uv=uv1 then u else find_subuc uv uc'
  | ac::uc' -> find_subuc uv uc';;
let rec find_upU_in_u u upU_map =
  match u with
    UVar(uv) -> []
  | ZeroU -> []
  | InU(l1,l2,u1) -> find_upU_in_u u1 upU_map
  | OutU(l1,l2,u1) -> find_upU_in_u u1 upU_map
  | ParU(u1,u2) -> merge_and_unify compare (find_upU_in_u u1 upU_map) (find_upU_in_u u2 upU_map) 
  | OrU(u1,u2) -> merge_and_unify compare (find_upU_in_u u1 upU_map) (find_upU_in_u u2 upU_map) 
  | RepU(u1) -> find_upU_in_u u1 upU_map
  | UpU(_) -> (try (List.assoc u upU_map; []) with Not_found -> [u])
  | _ -> raise (UnExpectedUsage u);;
let find_upU_in_ac ac upU_map =
  match ac with
    SubU(uv, u) -> find_upU_in_u u upU_map
  | RelU(u) -> find_upU_in_u u upU_map
  | _ -> [];;
let find_upU c upU_map =
  List.fold_right (function ac -> merge_and_unify compare (find_upU_in_ac ac upU_map)) c [];;

let rec type_map_usage f t  =
  match t with
    TVar _ -> t
  | UnitT -> t
  | BoolT _ -> t
  | IntT _ -> t
  | ChanT(t1,sl,u) -> ChanT(type_map_usage f t1, sl, f u)
  | ParT(t1,t2) -> ParT(type_map_usage f t1, type_map_usage f t2)
  | PairT(t1,t2) -> PairT(type_map_usage f t1, type_map_usage f t2)
  | FunT(t1,t2) ->  FunT(type_map_usage f t1, type_map_usage f t2)
  | RecT(id, tl, sl) -> RecT(id, List.map (type_map_usage f) tl, sl)
  | IncT(t1) -> IncT(type_map_usage f t1)
  | RepT(t1) -> RepT(type_map_usage f t1)
  | UpT(t1,l) -> UpT(type_map_usage f t1, l);;

   
let find_upU_in_t =
  type_fold_usage (function u -> find_upU_in_u u []) (merge_and_unify compare) [];;
let find_upU_in_te te =
  List.fold_right (function (x,t)->merge_and_unify compare (find_upU_in_t t)) te [];;

let rec mk_upU_map uc unprocessed_uc upU_map =
  let upUs = find_upU unprocessed_uc upU_map in
    if upUs = [] then ([], upU_map)
    else
       let new_upU_map = List.map (function x -> (x,new_uvid())) upUs in
       let new_uc = List.map (function (UpU(UVar(uv1),l1,l2),uv2) -> 
                       let u1 = find_subuc uv1 uc in
                           SubU(uv2, normalizing_UpU(u1, l1, l2))
                       | _ -> raise (SomethingIsWrongIn "mk_upU_map")) new_upU_map in
       let (new_uc', new_upU_map') = mk_upU_map uc new_uc (new_upU_map@upU_map) in
           (new_uc'@new_uc, new_upU_map');;

let rec replace_UpU_u upU_map u =
  match u with
    InU(l1,l2,u1) -> InU(l1,l2,replace_UpU_u upU_map u1)
  | OutU(l1,l2,u1) -> OutU(l1,l2,replace_UpU_u upU_map u1)
  | ParU(u1,u2) -> ParU(replace_UpU_u upU_map u1,replace_UpU_u upU_map u2)
  | OrU(u1,u2) -> OrU(replace_UpU_u upU_map u1,replace_UpU_u upU_map u2)
  | RepU(u1) -> RepU(replace_UpU_u upU_map u1)
  | UpU(_) -> (try UVar(List.assoc u upU_map) with Not_found -> u)
  | _ -> u;;
let replace_UpU_ac upU_map ac =
  match ac with
    SubU(uv, u) -> SubU(uv, replace_UpU_u upU_map u)
  | RelU(u) -> RelU(replace_UpU_u upU_map u)
  | _ -> ac;;
let replace_UpU_in_c upU_map =
  List.map (replace_UpU_ac upU_map);;

let replace_UpU_in_t upU_map t =
  type_map_usage (replace_UpU_u upU_map) t;;

let replace_UpU_in_te upU_map te =
  List.map (function (x,t) -> (x, replace_UpU_in_t upU_map t)) te;;

let rec eliminate_UpU_in_uc uc upU_map =
  let (new_uc, upU_map') = mk_upU_map uc uc upU_map in
     (replace_UpU_in_c upU_map' (new_uc@uc), upU_map');;

let eliminate_UpU_in_te te uc =
  let upUs = find_upU_in_te te in
  let upU_map = List.map (function x -> (x,new_uvid())) upUs in
  let new_uc = List.map (function (UpU(UVar(uv1),l1,l2),uv2) -> 
                       let u1 = find_subuc uv1 uc in
                           SubU(uv2, normalizing_UpU(u1, l1, l2))
                          | _ -> raise (SomethingIsWrongIn "eliminate_UpU_in_te")) upU_map in
  let te' = replace_UpU_in_te upU_map te in
    (te', upU_map, new_uc);;

let eliminate_UpU_in_TEtable uc upU_map =
  let _ = map_TEtable (replace_UpU_in_te upU_map) in
  let all_te = (get_TEtable()) in
  let upUs = find_upU_in_te (List.fold_left (@) [] all_te) in
  let new_upU_map = List.map (function x -> (x,new_uvid())) upUs in
  let new_uc = List.map (function (UpU(UVar(uv1),l1,l2),uv2) -> 
                       let u1 = find_subuc uv1 uc in
                           SubU(uv2, normalizing_UpU(u1, l1, l2))
                          | _ -> raise (SomethingIsWrongIn "eliminate_UpU_in_te")) new_upU_map in
  let _ = map_TEtable (replace_UpU_in_te upU_map) in
     (new_upU_map, new_uc)
  

let eliminate_UpU_in_tytab tytab uc upU_map=
  let upUs1 = find_upU_in_te tytab in
  let upUs = list_diff upUs1 (List.map (function (u,uv)->u) upU_map) in
  let upU_map2 = List.map (function x -> (x,new_uvid())) upUs in
  let new_uc = List.map (function (UpU(UVar(uv1),l1,l2),uv2) -> 
                       let u1 = find_subuc uv1 uc in
                           SubU(uv2, normalizing_UpU(u1, l1, l2))
                        | _ -> raise (SomethingIsWrongIn "eliminate_UpU_in_tytab")) upU_map2 in
  let tytab' = replace_UpU_in_te (upU_map@upU_map2) tytab in
    (tytab', upU_map2, new_uc);;

let eliminate_UpU_in_c c uc upU_map =
  let upUs = find_upU c upU_map in
  let upU_map2 = List.map (function x -> (x,new_uvid())) upUs in
  let new_uc = List.map (function (UpU(UVar(uv1),l1,l2),uv2) -> 
                       let u1 = find_subuc uv1 uc in
                           SubU(uv2, normalizing_UpU(u1, l1, l2))
                         | _ -> raise (SomethingIsWrongIn "eliminate_UpU_in_c")) upU_map2 in
  let c' = replace_UpU_in_c (upU_map@upU_map2) c in
    (c', upU_map2, new_uc);;

let mk_obtable eI eO upU_map =
  let eI' = List.sort (function (uv1,_) -> function (uv2,_) -> compare uv1 uv2) eI in 
  let eO' = List.sort (function (uv1,_) -> function (uv2,_) -> compare uv1 uv2) eO in 
  let obtable1 = List.map (function ((uv1,l1),(uv2,l2)) -> 
                                      if uv1=uv2 then (uv1,(l1,l2)) 
                                      else raise (SomethingIsWrongIn "mk_obtable"))
                   (List.combine eI' eO') in
  let obtable2 = List.map (function (u,uv) -> (uv, get_levels u eI' eO')) upU_map in
  let obtable2' = List.sort (function (uv1,_) -> function (uv2,_) -> compare uv1 uv2) obtable2 in
      merge (function (uv1,_) -> function (uv2,_) -> compare uv1 uv2) obtable1 obtable2';;

let eliminate_Up_and_Inc (te, uc, c,tytab) =
  let e1 = get_lev_equation uc true in
  let e2 = get_lev_equation uc false in
  let e1' = solve_lev_equations e1 [] in 
  let e2' = solve_lev_equations e2 [] in 
  let te' = eliminate_IncU_in_te te e1' e2' in
  let _ = if !Flag.wlauto then map_TEtable (fun te -> eliminate_IncU_in_te te e1' e2') 
          else () in
  let tytab' = eliminate_IncU_in_tytab tytab e1' e2' in
  let uc' = eliminate_IncU_in_c uc e1' e2' in
  let c' = eliminate_IncU_in_c c e1' e2' in
  let (te'',upU_map1,new_uc1) = eliminate_UpU_in_te te' uc' in
  let (tytab'',upU_map2,new_uc2) = eliminate_UpU_in_tytab tytab' uc' upU_map1 in
  let upU_map12 = upU_map1@upU_map2 in
  let (upU_map3, new_uc3) = if !Flag.wlauto then
                               eliminate_UpU_in_TEtable uc' upU_map12 
                            else ([], []) in
  let upU_map123 = upU_map12@upU_map3 in
  let (c'', upU_map4, new_uc4) = eliminate_UpU_in_c c' uc' upU_map123 in
  let (uc'', upU_map) = 
      eliminate_UpU_in_uc (new_uc1@new_uc2@new_uc3@new_uc4@uc') (upU_map123@upU_map4) in
  let obtable = mk_obtable e1' e2' upU_map in
    (te'', uc'', c'', obtable, tytab'');;

(*** functions for simplifying SubS and EqS constraints ***)
let rec eliminate_EqS c (s:substitution) rc =
  match c with
    [] -> (s, rc)
  | EqS(sl1,sl2)::c1 ->
       (match (sl1,sl2) with
          (SVar(sv1), _) -> 
              let s1 = slsubst2subst [(sv1,sl2)] in
              let s2 = comp_subst s1 s in
              let c1' = apply_subst_c s1 c1 in
              let rc' = apply_subst_c s1 rc in
                 eliminate_EqS c1' s2 rc'
         | (_, SVar(sv2)) -> 
              let s1 = slsubst2subst [(sv2,sl1)] in
              let s2 = comp_subst s1 s in
              let c1' = apply_subst_c s1 c1 in
              let rc' = apply_subst_c s1 rc in
                 eliminate_EqS c1' s2 rc'
         | (LOW, LOW) -> eliminate_EqS c1 s rc
         | (HI, HI) -> eliminate_EqS c1 s rc
         | _ -> if sl1=sl2 then eliminate_EqS c1 s rc
                else raise (TypeError "eliminate_EqS"))
  | ac::c1 ->
       eliminate_EqS c1 s (ac::rc);;

let mk_slsubst_from_eqs eqs =
  match eqs with
    [] -> []
  | [x] -> []
  | x::l -> List.map (function v -> (v,SVar(x))) l;;

let rec mk_slsubst_from_eqvrel eqvrel =
  match eqvrel with
    [] -> []
  | eqs::eqvrel' ->
       (mk_slsubst_from_eqs eqs)@(mk_slsubst_from_eqvrel eqvrel');;
let simplify_SubS c =
  let r = List.fold_right
            (function SubS(SVar(sv1),SVar(sv2))->
                          (function l -> (sv1,sv2)::l)
                | _ -> id) c [] in
  let eqvrel = Scc.compute_scc r in
  let s = slsubst2subst(mk_slsubst_from_eqvrel eqvrel) in
  let c' = apply_subst_c s c in
    (s, delete_duplication_unsorted c');;

let rec get_svrel (c, svrel, rest_c) =
  match c with
    [] -> (svrel, rest_c)
   | (SubS(SVar(sv1),SVar(sv2)) as ac)::c' -> get_svrel(c', (sv1,sv2)::svrel, ac::rest_c)
   | EqS(SVar(sv1),SVar(sv2))::c' -> get_svrel(c', (sv1,sv2)::(sv2,sv1)::svrel, rest_c)
   | ac::c' -> get_svrel(c', svrel, ac::rest_c)

let rec eliminate_tautology_SubS c rc =
  match c with
    [] -> rc
  | SubS(sl1,sl2)::c1 ->
       (match (sl1,sl2) with
          (LOW, _) ->  eliminate_tautology_SubS c1 rc
         | (_,HI) ->   eliminate_tautology_SubS c1 rc
         | _ -> if sl1=sl2 then eliminate_tautology_SubS c1 rc
                else eliminate_tautology_SubS c1 (SubS(sl1,sl2)::rc))
  | ac::c1 ->
       eliminate_tautology_SubS c1 (ac::rc);;

let simplify_SubS_and_EqS c =
  let _ = if !Flag.debug then Printf.printf "simplify_SubS phase 1: %f\n" (Sys.time()) in
  let (svrel, c') = get_svrel (c, [], []) in
  let _ = if !Flag.debug then Printf.printf "simplify_SubS phase 2: %f\n" (Sys.time()) in
  let eqvrel = Scc.compute_scc svrel in
  let _ = if !Flag.debug then Printf.printf "simplify_SubS phase 3: %f\n" (Sys.time()) in
  let sl_s = List.map (function (sv1,sv2) -> (sv1,SVar(sv2)))
               (List.fold_right (function vl -> (@)(eqclass2subst vl)) eqvrel []) in
  let _ = if !Flag.debug then Printf.printf "simplify_SubS phase 4: %f\n" (Sys.time()) in
  let s = slsubst2subst(sl_s) in
  let _ = if !Flag.debug then Printf.printf "simplify_SubS phase 5: %f\n" (Sys.time()) in
  let c'' = apply_subst_c s c' in
  let _ = if !Flag.debug then Printf.printf "simplify_SubS phase 6: %f\n" (Sys.time()) in
  let c3 = eliminate_tautology_SubS(delete_duplication_unsorted c'') [] in
  let _ = if !Flag.debug then Printf.printf "simplify_SubS phase 7: %f\n" (Sys.time()) in
    (s, c3);;


let rec eliminate_obvious_SubSIf c =
  match c with
    [] ->[]
  | SubSIf(_,LOW,_)::c1 -> eliminate_obvious_SubSIf c1
  | SubSIf(_,_,HI)::c1 -> eliminate_obvious_SubSIf c1
  | SubSIf(l,sl1,sl2)::c1 -> 
        if sl1=sl2 then eliminate_obvious_SubSIf c1
        else if l=LevInfty then (SubS(sl1,sl2)::(eliminate_obvious_SubSIf c1))
        else SubSIf(l,sl1,sl2)::(eliminate_obvious_SubSIf c1)
  | ac::c1 -> ac::(eliminate_obvious_SubSIf c1);;

(*** fromUtoPetri:
       usage * (subusage constraint) * (transition map) * (ids of atomic usage) 
      ----> (list of state vectors) * (transition map) * (ids of atomic usage)
     transition map consists of
        input transitions
        output transitions
        recursive usage expansions
 ***)
let unused_id idmap =
  match idmap with
    [] -> 0
  | (x,id)::_ -> id+1;;
let ext_idmap idmap x id =
  (x,id)::idmap;;
type int_or_infty = Fin of int | Unknown;;
type petri_state = PState of (int * int_or_infty) list;;
type state_vector = petri_state list;;
let add_intinf n1 n2 =
  match (n1,n2) with
    (Fin(x1), Fin(x2)) -> Fin(x1+x2)
  | _ -> Unknown;; (*** This loses information; note that 2+unknown >=2 ***)
let mul_unknown_intinf n =
  match n with
     Fin(0) -> Fin(0)
  | _ -> Unknown;;
(*** add two atomic state vectors. 
 *** assume that atomic state vectors are sorted 
 ***)
let rec par_atomic_state_vector asv1 asv2 =
  match (asv1, asv2) with
    ([], _) -> asv2
  | (_, []) -> asv1
  | ((id1,token1)::l1, (id2,token2)::l2) ->
        if id1<id2 then (id1,token1)::(par_atomic_state_vector l1 asv2)
        else if id1>id2 then 
             (id2,token2)::(par_atomic_state_vector asv1 l2)
        else (** id1=id2 **)
          (id1, add_intinf token1 token2)::(par_atomic_state_vector l1 l2);;
let par_state_vector sv1 sv2 =
  List.fold_right 
      (function PState(asv1) -> function sv -> 
             (List.map (function PState(asv2) -> 
                         PState(par_atomic_state_vector asv1 asv2)) sv2)@sv) sv1 [];;

let or_state_vector sv1 sv2 = sv1@sv2;;
let rep_atomic_state_vector =
  List.map (function (id, n) -> (id, mul_unknown_intinf n));;
let rep_state_vector =
  List.map (function PState(asv)-> PState(rep_atomic_state_vector asv));;
type transition_rules = 
   ((int * petri_state list) list) * ((int * petri_state list) list) * ((int * petri_state list) list);;
let empty_transition = ([],[],[]);;
let add_rec_transition (rec_tr, in_tr, out_tr) id sv:transition_rules =
  ((id, sv)::rec_tr, in_tr, out_tr);;
let add_in_transition (rec_tr, in_tr, out_tr) id sv:transition_rules =
  (rec_tr, (id, sv)::in_tr, out_tr);;
let add_out_transition (rec_tr, in_tr, out_tr) id sv:transition_rules =
  (rec_tr, in_tr, (id, sv)::out_tr);;
let add_transition (tr_kind:int) =
  if tr_kind =0 then add_rec_transition 
  else if tr_kind=1 then add_in_transition
  else add_out_transition;;

let rec fromUtoPetri u tr idmap =
  match u with
    UVar(uv) -> 
       (try 
         let id = List.assoc (UVar(uv)) idmap in
         let sv = [PState([(id, Fin(1))])] in
            (sv, tr, idmap)
       with Not_found -> 
         let id = unused_id idmap in
         let sv = [PState([(id, Fin(1))])] in
            (sv, tr, ext_idmap idmap (UVar(uv)) id))
  | ZeroU ->
       ([PState([])], tr, idmap)
  | InU(l1,l2,u1) ->
       (try
         let id = List.assoc u idmap in (*** transition rule for u has already been generated ***)
         let sv = [PState([(id, Fin(1))])] in
           (sv, tr, idmap)
       with Not_found ->
         let (sv1, tr1, idmap1) = fromUtoPetri u1 tr idmap in
         let id = unused_id idmap1 in
         let sv = [PState([(id, Fin(1))])] in
         let idmap2 = ext_idmap idmap1 u id in
         let tr2 = add_in_transition tr1 id (par_state_vector [PState([(id, Fin(-1))])] sv1) in
           (sv, tr2, idmap2))
  | OutU(l1,l2,u1) ->
       (try
         let id = List.assoc u idmap in (*** transition rule for u has already been generated ***)
         let sv = [PState([(id, Fin(1))])] in
           (sv, tr, idmap)
       with Not_found ->
         let (sv1, tr1, idmap1) = fromUtoPetri u1 tr idmap in
         let id = unused_id idmap1 in
         let sv = [PState([(id,Fin(1))])] in
         let idmap2 = ext_idmap idmap1 u id in
         let tr2 = add_out_transition tr1 id (par_state_vector [PState([(id, Fin(-1))])] sv1) in
           (sv, tr2, idmap2))
  | ParU(u1, u2) ->
       let (sv1,tr1,idmap1) = fromUtoPetri u1 tr idmap in
       let (sv2,tr2,idmap2) = fromUtoPetri u2 tr1 idmap1 in
       let sv = par_state_vector sv1 sv2 in
           (sv, tr2, idmap2)
  | OrU(u1, u2) ->
       let (sv1,tr1,idmap1) = fromUtoPetri u1 tr idmap in
       let (sv2,tr2,idmap2) = fromUtoPetri u2 tr1 idmap1 in
       let sv = or_state_vector sv1 sv2 in
           (sv, tr2, idmap2)
  | RepU(u1) -> (*** approximate *(U1&...&Un) with X and transition X-> !U1|...|!Un|X ***)
       let (sv1,tr1,idmap1) = fromUtoPetri u1 tr idmap in
       let sv2 = [PState (List.fold_left 
                           (fun x->function PState(y) -> par_atomic_state_vector x y) [] sv1)] in
       let sv3 = rep_state_vector sv2 in
       (try
         let id = List.assoc u idmap1 in (*** transition rule for u has already been generated ***)
         let sv = [PState([(id, Fin(1))])] in
           (sv, tr1, idmap1)
       with Not_found ->
         let id = unused_id idmap1 in
         let sv = [PState([(id,Fin(1))])] in
         let idmap2 = ext_idmap idmap1 u id in
         let tr2 = add_rec_transition tr1 id sv3 in
           (sv, tr2, idmap2))
  | _ -> raise (UnExpectedUsage u)
and fromUCtoPetri uc tr idmap =
  match uc with
    [] -> (tr, idmap)
  | SubU(uv,u)::uc' ->
      let (sv1,tr1,idmap1) = fromUtoPetri u tr idmap in
      let (id,idmap2) =
          (try ((List.assoc (UVar(uv)) idmap1), idmap1) with
            Not_found ->
             let id = unused_id idmap1 in
                 (id, ext_idmap idmap1 (UVar(uv)) id)) in
      let tr2 = add_rec_transition tr1 id (par_state_vector [PState([(id, Fin(-1))])] sv1) in
        fromUCtoPetri uc' tr2 idmap2
  | ac::uc' -> fromUCtoPetri uc' tr idmap;;


let set_union = merge_and_unify compare;;
let rec idset_in_asv asv =
   List.map fst asv;;
let rec idset_in_sv (sv:state_vector) =
   List.fold_right (function PState(asv) -> set_union (idset_in_asv asv)) sv [];;

let get_tr id (rec_tr,in_tr, out_tr) idmap =
  let (u, idmap') = list_assoc2 id idmap in
    match u with
      UVar(_) -> let (sv, rec_tr') = list_assoc2 id rec_tr in 
                        (0, sv, u, (rec_tr',in_tr,out_tr), idmap')
    | InU (_,_,_) -> let (sv, in_tr') = list_assoc2 id in_tr in 
                        (1, sv, u, (rec_tr,in_tr',out_tr), idmap')
    | OutU (_,_,_) -> let (sv, out_tr') = list_assoc2 id out_tr in 
                        (2, sv, u, (rec_tr,in_tr,out_tr'), idmap')
    | RepU (_) -> let (sv, rec_tr') = list_assoc2 id rec_tr in (0, sv, u, (rec_tr',in_tr,out_tr), idmap')
    | _ -> raise (UnExpectedUsage u);;
 
let rec closure_of_tr_rules idset tr idmap tr' idmap' =
  match idset with
    [] -> (tr', idmap')
  | id::idset' -> 
       try
         let (tr_kind, sv, u, tr1, idmap1) = get_tr id tr idmap in
           closure_of_tr_rules (set_union (idset_in_sv sv) idset') tr1 idmap1 
                      (add_transition tr_kind tr' id sv) (ext_idmap idmap' id u)
       with
         Not_found -> closure_of_tr_rules idset' tr idmap tr' idmap';;

(*** from tokens to bdd representation of the abstract state ***)
(***  0: 000,   1: 010,   2: 100,   >=3: 110
 ***  >=0: 001, >=1: 011,   >=2: 101,   infinity: 111
 ***)
(***  0: 00,   1: 01,   2: 10,   >=3: 11
 ***)
let mk_bdd_zero_token v =
  let b1 = 2*v in
    Obdd.bdd_and (Obdd.bdd_nvar b1) (Obdd.bdd_nvar (b1+1));;
(***
let mk_bdd_omega_token v =
  let b1 = 3*v in
    Obdd.bdd_and (Obdd.bdd_var b1) (Obdd.bdd_and (Obdd.bdd_var (b1+1)) (Obdd.bdd_var (b1+2)));;
***)
let mk_bdd_token v x =
 let b1 = 2*v in
  match x with
    Fin(n) ->
       if n>=3 then 
         Obdd.bdd_and (Obdd.bdd_var b1) (Obdd.bdd_var (b1+1))
       else if n=2 then
         Obdd.bdd_and (Obdd.bdd_var b1) (Obdd.bdd_nvar (b1+1))
       else if n=1 then
         Obdd.bdd_and (Obdd.bdd_nvar b1) (Obdd.bdd_var (b1+1))
       else mk_bdd_zero_token v
  | Unknown -> 
         Obdd.bdd_true

let rec asv_to_bdd asv id_renaming =
  match (asv, id_renaming) with
    ([], _) -> List.fold_right 
                (function (id,v)->(Obdd.bdd_and (mk_bdd_zero_token v))) id_renaming Obdd.bdd_true 
  | ((id1,n)::asv', (id2,v2)::id_renaming') ->
        if id1=id2 then Obdd.bdd_and (mk_bdd_token v2 n) (asv_to_bdd asv' id_renaming')
        else if id1>id2 then
           Obdd.bdd_and (mk_bdd_zero_token v2) (asv_to_bdd asv id_renaming')
        else raise (SomethingIsWrongIn "asv_to_bdd:1")
  | _ -> raise (SomethingIsWrongIn "asv_to_bdd:2");;
               
let rec sv_to_bdd sv id_renaming =
  match sv with
    [] -> Obdd.bdd_false
  | (PState(asv)::sv') ->
        let t1 = asv_to_bdd asv id_renaming in
        let t2 = sv_to_bdd sv' id_renaming in
            Obdd.bdd_or t1 t2;;

let mk_bdd_pos v1 v2  =
    Obdd.bdd_or (Obdd.bdd_var v1) (Obdd.bdd_var v2);;

(***
let mk_bdd_nonomega v1 v2 v3 =
    Obdd.bdd_orl [(Obdd.bdd_nvar v1); (Obdd.bdd_nvar v2); (Obdd.bdd_nvar v3)];;
let mk_bdd_spos v1 v2 v3 =
    Obdd.bdd_or (Obdd.bdd_var v1) (Obdd.bdd_var v2);;
***)
let mk_bdd_zero v1 v2 =  (*** 00 ***)
    Obdd.bdd_andl [(Obdd.bdd_nvar v1); (Obdd.bdd_nvar v2)];;

let mk_bdd_one v1 v2  =  (*** 01 ***)
    Obdd.bdd_andl [(Obdd.bdd_nvar v1); (Obdd.bdd_var v2)];;

let mk_bdd_two v1 v2 = (*** 10 ***)
    Obdd.bdd_andl [(Obdd.bdd_var v1); (Obdd.bdd_nvar v2)];;

let mk_bdd_gtthree v1 v2 = (*** 11 ***)
    Obdd.bdd_andl [(Obdd.bdd_var v1); (Obdd.bdd_var v2)];;

let mk_bdd_gtzero v1 v2  = 
    Obdd.bdd_true

let mk_bdd_gtone v1 v2  = (*** 011 ***)
    Obdd.bdd_orl [(Obdd.bdd_var v1);(Obdd.bdd_var v2)];;

let mk_bdd_gttwo v1 v2 = (*** 101 ***)
    (Obdd.bdd_var v1);;

(***
let mk_bdd_omega v1 v2 v3 =  (*** 111 ***)
    Obdd.bdd_andl [(Obdd.bdd_var v1); (Obdd.bdd_var v2); (Obdd.bdd_var v3)];;
***)

let mk_bddrel_eq v =
  let v1 = 4*v in
  let (v1',v2,v2') = (v1+1,v1+2,v1+3) in
    Obdd.bdd_andl [(Obdd.bdd_eq v1 v1'); (Obdd.bdd_eq v2 v2')];;

let mk_bddrel_eqs vl =
  List.fold_right (function v -> (Obdd.bdd_and (mk_bddrel_eq v))) vl Obdd.bdd_true;;

let bddrel_minusone =
      (Obdd.bdd_orl
         [(Obdd.bdd_and (mk_bdd_one 0 2) (mk_bdd_zero 1 3));
          (Obdd.bdd_and (mk_bdd_two 0 2) (mk_bdd_one 1 3));
          (Obdd.bdd_and (mk_bdd_gtthree 0 2) (mk_bdd_gttwo 1 3))]);;

let bddrel_plusone =
      (Obdd.bdd_orl
         [(Obdd.bdd_and (mk_bdd_zero 0 2) (mk_bdd_one 1 3));
          (Obdd.bdd_and (mk_bdd_one 0 2) (mk_bdd_two 1 3));
          (Obdd.bdd_and (mk_bdd_two 0 2) (mk_bdd_gtthree 1 3));
          (Obdd.bdd_and (mk_bdd_gtthree 0 2) (mk_bdd_gtthree 1 3))]);;

let bddrel_plustwo =
      (Obdd.bdd_orl
         [(Obdd.bdd_and (mk_bdd_zero 0 2) (mk_bdd_two 1 3));
          (Obdd.bdd_and (mk_bdd_one 0 2) (mk_bdd_gtthree 1 3));
          (Obdd.bdd_and (mk_bdd_two 0 2) (mk_bdd_gtthree 1 3));
          (Obdd.bdd_and (mk_bdd_gtthree 0 2) (mk_bdd_gtthree 1 3))]);;

let bddrel_plusthree =
      (mk_bdd_gtthree 1 3);;

(***
let bddrel_plusomega =
     mk_bdd_omega 1 3 5;;
***)
let bddrel_plusunknown =
      (Obdd.bdd_orl
         [(mk_bdd_zero 0 2);
          (Obdd.bdd_and (mk_bdd_one 0 2) (mk_bdd_gtone 1 3));
          (Obdd.bdd_and (mk_bdd_two 0 2) (mk_bdd_gttwo 1 3));
          (Obdd.bdd_and (mk_bdd_gtthree 0 2) (mk_bdd_gtthree 1 3))]);;

let mk_bddrel_pos x =
  let v1 = 4*x in  mk_bdd_pos v1 (v1+2);;
let mk_bddrel_minusone x =
    Obdd.rename (function n -> 4*x+n) (bddrel_minusone);;
let mk_bddrel_plusone x =
    Obdd.rename (function n -> 4*x+n) (bddrel_plusone);;
let mk_bddrel_plustwo x =
    Obdd.rename (function n -> 4*x+n) (bddrel_plustwo);;
let mk_bddrel_plusthree x =
    Obdd.rename (function n -> 4*x+n) (bddrel_plusthree);;
(***
let mk_bddrel_plusomega x =
    Obdd.rename (function n -> 4*x+n) (bddrel_plusomega);;
***)
let mk_bddrel_plusunknown x =
    Obdd.rename (function n -> 4*x+n) (bddrel_plusunknown);;

let mk_bddrel_vec x delta =
  match delta with
    Unknown -> mk_bddrel_plusunknown x
  | Fin(-1) -> mk_bddrel_minusone x
  | Fin(0) -> mk_bddrel_eq x
  | Fin(1) -> mk_bddrel_plusone x
  | Fin(2) -> mk_bddrel_plustwo x
  | _ -> mk_bddrel_plusthree x;;

let rec asv_to_bddrel id_renaming asv =
  match (id_renaming, asv) with
    (_, []) -> ([], Obdd.bdd_true)
  | ((id1,v1)::id_renaming', (id2, d1)::asv') ->
        if id1<id2 then asv_to_bddrel id_renaming' asv
        else if id1=id2 then
           (if not(d1=Fin(0)) then 
             let (vl, t) = asv_to_bddrel id_renaming' asv' in
               (v1::vl, Obdd.bdd_and (mk_bddrel_vec v1 d1) t)
            else
              asv_to_bddrel id_renaming' asv')
        else raise (SomethingIsWrongIn "asv_to_bddrel:1")
  | _ -> raise (SomethingIsWrongIn "asv_to_bddrel:2");;

let sv_to_bddrel id_renaming sv =
  List.map (function PState(asv) -> (asv_to_bddrel id_renaming asv)) sv;;

let one_rectr_to_bddl id_renaming (id, sv) =
  let t1 = (mk_bddrel_pos (List.assoc id id_renaming)) in
    List.map (function (vl,t2) -> (vl, Obdd.bdd_and t1 t2)) (sv_to_bddrel id_renaming sv);;

let rectr_to_bddlist rec_tr id_renaming =
  List.fold_right (function one_tr -> (@)(one_rectr_to_bddl id_renaming one_tr)) rec_tr [];;

let one_iotr_to_bddl id_renaming ((id1,sv1),(id2,sv2)) =
  let t1 = Obdd.bdd_and (mk_bddrel_pos (List.assoc id1 id_renaming)) 
                        (mk_bddrel_pos (List.assoc id2 id_renaming)) in
    List.map (function (vl,t2) -> (vl, Obdd.bdd_and t1 t2)) 
                 (sv_to_bddrel id_renaming (par_state_vector sv1 sv2));;

let inouttr_to_bddlist in_tr out_tr id_renaming =
  let iopairs = make_pairs in_tr out_tr in
    List.fold_right (function iopair -> (@)(one_iotr_to_bddl id_renaming iopair)) iopairs [];;

(*** tr_to_bddlist makes a list of 
 *** ([set of updated variables], [BDD representing a transition relation]) 
 ***)
let tr_to_bddlist (rec_tr, in_tr, out_tr) id_renaming =
  let bddl1 = rectr_to_bddlist rec_tr id_renaming in
  let bddl2 = inouttr_to_bddlist in_tr out_tr id_renaming in
     bddl1@bddl2;;

let tr_to_bddfunction tr id_renaming =
(***  let n = (List.length id_renaming)*3 in ***)
  let bddl = tr_to_bddlist tr id_renaming in
  let bddl' = List.map (function (vl,rel_bdd) -> 
		                 let bits = List.fold_right (function v-> function bl ->
                                        let b1 = 4*v in b1::(b1+2)::bl) vl [] in
                                 (bits, rel_bdd)) bddl in
    List.fold_right 
      (function (bits,rel_bdd) -> function st ->
       Obdd.bdd_or  (Obdd.rename (function v->v/2) 
                    (Obdd.relprod bits rel_bdd (Obdd.rename (function v->2*v) st)))
       st)
      bddl';;
 

let optimize_reptr trs idmap =
  let (rec_trs, in_trs, out_trs) = trs in
  let rec pos_places_in_asv asv =
     List.map fst (List.filter (function (p, Fin(n)) -> n>0 | _ -> true) asv) in
  let rec pos_places_in_sv sv =
    match sv with
      [] -> []
    | PState(asv)::sv' -> 
        let ps1 = pos_places_in_asv asv in
        let ps2 = pos_places_in_sv sv' in
          merge_and_unify compare ps1 ps2 in
  let rec tr_closure ps found  =
     match ps with
        [] -> found
      | p::ps' -> 
         if List.mem p found then tr_closure ps' found
         else 
            try 
              let sv = List.assoc p rec_trs in
              let ps1 = pos_places_in_sv sv in
                 tr_closure (ps1@ps') (p::found)
            with Not_found -> tr_closure ps' (p::found) in
               
  let optimize_onetr (id, sv) =
     match List.assoc id idmap with
        RepU(u) -> let ps = pos_places_in_sv sv in
                   let ps' = tr_closure ps [] in
                   let ps'' = List.sort compare ps' in
                   let sv2 = [PState(List.map (fun p -> (p, Unknown)) ps'')] in
                   let sv' = par_state_vector sv sv2 in     
                                (id, sv')
      | _ -> (id, sv) in

  let rec_trs' = List.map optimize_onetr rec_trs in
    (rec_trs', in_trs, out_trs)

let get_tr_closure u tr idmap =
  let (sv, tr1, idmap1) = fromUtoPetri u tr idmap in
  let (tr2, idmap2) = closure_of_tr_rules (idset_in_sv sv) 
                                          tr1 (List.map swap idmap1) empty_transition [] in
  let tr3 = optimize_reptr tr2 idmap2 in 
            (*** optimize transition rules so that the state space of RelU(u) converges quickly ***)
  let idmap2' = List.sort (function x->function y->compare (fst x) (fst y)) idmap2 in 
  let idset = List.map fst idmap2' in
(***  let n = List.length idset in  ***)
  let x = ref 0 in 
  let id_renaming = (List.map (function id -> (id, let new_id = !x in (x:=!x+1;new_id))) idset) in
  let u_bdd = sv_to_bdd sv id_renaming in
  let tr_bddfunction = tr_to_bddfunction tr3 id_renaming in
  let u_bdd_clos = Obdd.fixp tr_bddfunction u_bdd in
      (u_bdd_clos, idmap2, id_renaming);;

let rec get_RelU c =
  match c with
    [] -> ([], [])
  | RelU(u)::c1 -> 
        let (ul, c1') = get_RelU c1 in (u::ul, c1')
  | ac::c1 ->
        let (ul, c1') = get_RelU c1 in (ul, ac::c1');;

let add_to_levlist level id caplist =
  try
     (let entry = List.assoc level caplist in
      let _ =   entry := merge_and_unify compare [id] !entry in
         caplist)
  with Not_found ->
     (level, ref [id])::caplist;;

let rec get_in_caplist idmap id_renaming =
  match idmap with
    [] -> []
  | (InU(l1,l2,u), id)::idmap' -> 
          let incaplist = get_in_caplist idmap' id_renaming in
             add_to_levlist l2 (List.assoc id id_renaming) incaplist
  | (u,id)::idmap' -> get_in_caplist idmap' id_renaming;;

let rec get_out_caplist idmap id_renaming =
  match idmap with
    [] -> []
  | (OutU(l1,l2,u), id)::idmap' -> 
          let outcaplist = get_out_caplist idmap' id_renaming in
             add_to_levlist l2 (List.assoc id id_renaming) outcaplist
  | (u,id)::idmap' -> get_out_caplist idmap' id_renaming;;

let rec ob_of_usage u obtab =
  let (eI, eO) = List.split (List.map (fun (id, (oI, oO)) -> ((id,oI),(id,oO))) obtab) in
    get_levels u eI eO;;

let rec get_in_oblist idmap id_renaming obtab =
  match idmap with
    [] -> []
  | (InU(l1,l2,u),id)::idmap' ->
        let oblist = get_in_oblist idmap' id_renaming obtab in
           add_to_levlist l1 (List.assoc id id_renaming) oblist
  | (UVar(uv),id)::idmap' ->
        let oblist = get_in_oblist idmap' id_renaming obtab in
        let (obI,obO) = List.assoc uv obtab in
           add_to_levlist obI (List.assoc id id_renaming) oblist
  | (RepU(u), id)::idmap' ->
        let oblist = get_in_oblist idmap' id_renaming obtab in
        let (obI,obO) = ob_of_usage u obtab in        
           add_to_levlist obI (List.assoc id id_renaming) oblist
  | (u,id)::idmap' -> get_in_oblist idmap' id_renaming obtab;;

let rec get_out_oblist idmap id_renaming obtab =
  match idmap with
    [] -> []
  | (OutU(l1,l2,u),id)::idmap' ->
        let oblist = get_out_oblist idmap' id_renaming obtab in
           add_to_levlist l1 (List.assoc id id_renaming) oblist
  | (UVar(uv),id)::idmap' ->
        let oblist = get_out_oblist idmap' id_renaming obtab in
        let (obI,obO) = List.assoc uv obtab in
           add_to_levlist obO (List.assoc id id_renaming) oblist
  | (RepU(u), id)::idmap' ->
        let oblist = get_out_oblist idmap' id_renaming obtab in
        let (obI,obO) = ob_of_usage u obtab in        
           add_to_levlist obO (List.assoc id id_renaming) oblist
  | (u,id)::idmap' -> get_out_oblist idmap' id_renaming obtab;;


let mk_bdd_may_pos ids u_bdd =
  let bits = List.map (function id -> 2*id) ids in
  let pos_bdd = Obdd.bdd_orl (List.map (function b -> mk_bdd_pos b (b+1) ) bits) in
  let vars = List.fold_right (function b -> function vl -> b::(b+1)::vl) bits [] in
    Obdd.relprod vars u_bdd pos_bdd;;

let mk_bdd_must_pos b ids =
  let t1 = Obdd.bdd_var b in
  let bits = List.map (function id -> 2*id) ids in
  let t2 = Obdd.bdd_orl (List.map (function b -> mk_bdd_pos b (b+1)) bits) in
    Obdd.iff t1 t2;;

let obligation_of_bdd u_bdd oblist num_of_ids =
  let lookup_oblist i = fst(List.nth oblist i) in
  let bits_of_ids ids = List.fold_right (function id->function l->
                                          let b = 2*id in b::(b+1)::l) ids [] in
  let n = List.length oblist in (*** the number of groups having the same obligation level ***)
  let x = 2*num_of_ids in
  let l = fromto x (x+n) in
  let obl = List.combine l oblist in
  let bdd = List.fold_right (function (lid, (lev, r)) -> function t ->
                        Obdd.relprod (bits_of_ids !r) (mk_bdd_must_pos lid (!r)) t) obl u_bdd in
  let vl = fromto 0 x in
  let bdd' = Obdd.exists_vl vl bdd in
    Obdd.cata (function v -> function l1 -> function l2 ->
         lexp_max(lexp_min (lookup_oblist (v-x)) l1) l2) 
          (function b -> if b then LevInfty else LevConst 0) bdd'

let in_obligation_bdd u_bdd idmap id_renaming obtable =
  let inoblist = get_in_oblist idmap id_renaming obtable in
  let n = List.length id_renaming in 
    obligation_of_bdd u_bdd inoblist n;;

let out_obligation_bdd u_bdd idmap id_renaming obtable =
  let outoblist = get_out_oblist idmap id_renaming obtable in
  let n = List.length id_renaming in 
    obligation_of_bdd u_bdd outoblist n;;

let get_constr_on_in_cap u_bdd idmap id_renaming obtable (cap_level, ref_of_ids) =
  let ids = !ref_of_ids in
(***  let varnum = List.length id_renaming in ***)
  let u_bdd' = mk_bdd_may_pos ids u_bdd in
  let ob_level = out_obligation_bdd u_bdd' idmap id_renaming obtable in
    [SubL(ob_level, cap_level)];;

let get_constr_on_out_cap u_bdd idmap id_renaming obtable (cap_level, ref_of_ids) =
  let ids = !ref_of_ids in
(***  let varnum = List.length id_renaming in ***)
  let u_bdd' = mk_bdd_may_pos ids u_bdd in
  let ob_level = in_obligation_bdd u_bdd' idmap id_renaming obtable in
    [SubL(ob_level, cap_level)];;

let reduce_RelU tr idmap uc obtable u =
  let _ = if !Flag.debug then Printf.printf "Start Reducing RelU: %f\n" (Sys.time()) in
  let (u_clos, restr_idmap', id_renaming) = get_tr_closure u tr idmap in
  let _ = if !Flag.debug then Printf.printf "Closure obtained: %f\n" (Sys.time()) in
  let restr_idmap = List.map swap restr_idmap' in
  let incaplist = get_in_caplist restr_idmap id_renaming in
  let cl1 = List.map (get_constr_on_in_cap u_clos restr_idmap id_renaming obtable) incaplist in
  let outcaplist = get_out_caplist restr_idmap id_renaming in
  let cl2 = List.map (get_constr_on_out_cap u_clos restr_idmap id_renaming obtable) outcaplist in
  let _ = if !Flag.debug then Printf.printf "RelU reduced: %f\n" (Sys.time()) in
  let _ = Obdd.reset() in
  let _ = if !Flag.debug then Printf.printf "Hash table reset: %f\n" (Sys.time()) in
      list_flatten(cl1@cl2);;

let eliminate_RelU uc c obtable =
  let (tr, idmap) = fromUCtoPetri uc empty_transition [] in
  let (ul, c1) = get_RelU c in
  let newc = list_flatten (List.map (reduce_RelU tr idmap uc obtable) ul) in
     newc@c1;;

let rec get_SubL c =
  match c with
    [] -> ([], [])
  | SubL(l1,LevVar lv)::c' -> 
       let (c1,c2) = get_SubL c' in 
           ((lv, l1)::c1, c2)
  | ac::c' ->
       let (c1,c2) = get_SubL c' in (c1, ac::c2)

let rec levvars lexp =
  match lexp with
    LevVar(lv) -> [lv]
  | LevIfInfty(sl) -> levvars_in_sl(sl)
  | LevPlus(l,n) -> levvars(l)
  | LevMax(ll) -> List.fold_right (function l -> (merge_and_unify compare (levvars l))) ll []
  | LevMin(ll) -> List.fold_right (function l -> (merge_and_unify compare (levvars l))) ll []
  | _ -> []
and levvars_in_sl sl =
   match sl with
     HI -> []
   | LOW -> []
   | SVar(sv) -> []
   | Sglb(sl1, sl2) -> merge_and_unify compare (levvars_in_sl sl1) (levvars_in_sl sl2)
   | SIfInf(l,sl') -> levvars l;;

let delete_duplicated_levgeq =
   delete_duplication2 (function (lv1,_)->function (lv2,_)->compare lv1 lv2) 
                       (function (lv1,l1)->function (lv2,l2) -> (lv1, lexp_max l1 l2));;

let rec replace_lv_with_inf lv l =
  match l with
     LevRec _ -> l
   | LevConst(n) -> l
   | LevInfty -> LevInfty
   | LevVar(lv') -> if lv=lv' then LevInfty else l
   | LevIfInfty(sl) -> l  (*** We assume that sl does not contain level variables ***)
   | LevPlus(l1,n) -> levPlus n (replace_lv_with_inf lv l1)
   | LevMax(ll) -> let ll' = List.map (replace_lv_with_inf lv) ll in
                     if List.mem LevInfty ll' then LevInfty
                     else LevMax(ll')
   | LevMin(ll) -> let ll' =  List.map (replace_lv_with_inf lv) ll in
                     if List.mem (LevConst 0) ll' then (LevConst 0)
                     else LevMin(ll')
let rec solve_one_subl lv l =
  match l with
     LevRec _ -> l
   | LevConst(n) -> l
   | LevInfty -> LevInfty
   | LevVar(lv') -> if lv=lv' then LevConst(0) else l
   | LevIfInfty(sl) -> l  (*** We assume that sl does not contain level variables ***)
   | LevPlus(l1,n) -> if n=0 then solve_one_subl lv l1
                      else LevPlus(replace_lv_with_inf lv l1, n)
   | LevMax(ll) -> LevMax (List.map (solve_one_subl lv) ll)
   | LevMin(ll) -> LevMin (List.map (solve_one_subl lv) ll);;


let rec simplify_lexp_and_sl_in_l l =
  match l with
    LevVar _ -> l
  | LevConst _ -> l
  | LevInfty  -> l
  | LevRec _ -> l
  | LevIfInfty(sl) -> 
       let sl' = simplify_lexp_and_sl_in_sl(sl) in
         (match sl' with
            LOW -> LevInfty
	  | HI -> LevConst 0
          | _ -> LevIfInfty(sl'))
  | LevPlus(l1, n) ->
       let l1' = simplify_lexp_and_sl_in_l l1 in
         ( match l1' with
            LevInfty -> LevInfty
          | LevConst m -> LevConst (m+n)
          | _ -> LevPlus(l1',n))
  | LevMax(ll) ->
       let ll1 = List.map simplify_lexp_and_sl_in_l ll in
         if List.mem LevInfty ll1 then LevInfty
         else
           let m = List.fold_right (function LevConst(n)->max n | _ -> id) ll1 0 in
           let ll2 = 
               LevConst(m)::(List.filter (function LevConst(n)->false | _ -> true) ll1) in
           let ll3 = delete_duplication_unsorted ll2 in
           (match ll3 with
             [l] -> l
            | _ -> LevMax(ll3))
  | LevMin(ll) ->
       let ll1 = List.map simplify_lexp_and_sl_in_l ll in
       let m = 
           List.fold_right (function LevConst(n1)->
			      (function Some(n2) -> Some(min n1 n2) | _ -> Some(n1))
                           | _ -> id) ll1 None in
       let ll2 = match m with 
                    Some(n) -> 
                      LevConst(n)::
                       (List.filter (function LevConst(n1)->false | LevInfty -> false | _ -> true) ll1)
                  | None -> ll1 in
       let ll3 = delete_duplication_unsorted ll2 in
           (match ll3 with
             [l] -> l
            | _ -> LevMin(ll3))
and simplify_lexp_and_sl_in_sl sl =
  match sl with
     HI  -> HI
   | LOW -> LOW
   | SVar _ -> sl
   | Sglb (sl1,sl2) -> 
        let sl1' = simplify_lexp_and_sl_in_sl sl1 in
        let sl2' = simplify_lexp_and_sl_in_sl sl2 in
         sglb(sl1',sl2')
   | SIfInf(l,sl') ->
        let l' = simplify_lexp_and_sl_in_l l in
          (match l' with 
             LevInfty -> simplify_lexp_and_sl_in_sl sl'
           | LevConst _ -> HI
           | _ -> SIfInf(l',simplify_lexp_and_sl_in_sl sl'));;

let rec simplify_lexp_and_sl_in_u u =
  match u with
    UVar _ -> u
  | ZeroU -> u
  | InU(l1,l2,u1) -> InU(simplify_lexp_and_sl_in_l l1,
			 simplify_lexp_and_sl_in_l l2,
                         simplify_lexp_and_sl_in_u u1)
  | OutU(l1,l2,u1) -> OutU(simplify_lexp_and_sl_in_l l1,
                           simplify_lexp_and_sl_in_l l2,
                           simplify_lexp_and_sl_in_u u1)
  | ParU(u1,u2) -> ParU(simplify_lexp_and_sl_in_u u1,simplify_lexp_and_sl_in_u u2)
  | OrU(u1,u2) -> ParU(simplify_lexp_and_sl_in_u u1,simplify_lexp_and_sl_in_u u2)
  | RepU(u1) -> RepU(simplify_lexp_and_sl_in_u u1)
  | IncU(u1) -> IncU(simplify_lexp_and_sl_in_u u1)
  | UpU(u1,l1,l2) -> UpU(simplify_lexp_and_sl_in_u u1, 
                         simplify_lexp_and_sl_in_l l1,
                         simplify_lexp_and_sl_in_l l2)
  | URec(uv,u) -> URec(uv,simplify_lexp_and_sl_in_u u);;
let rec simplify_lexp_and_sl_in_ty t =
  match t with 
    TVar(_) -> t
  | UnitT -> t
  | BoolT sl -> BoolT(simplify_lexp_and_sl_in_sl sl)
  | IntT sl -> IntT(simplify_lexp_and_sl_in_sl sl)
  | ChanT(t1,sl,u1) -> ChanT(simplify_lexp_and_sl_in_ty t1, 
			     simplify_lexp_and_sl_in_sl sl, 
                             simplify_lexp_and_sl_in_u u1)
  | PairT(t1,t2) -> PairT(simplify_lexp_and_sl_in_ty t1,simplify_lexp_and_sl_in_ty t2)
  | FunT(t1,t2) -> FunT(simplify_lexp_and_sl_in_ty t1,simplify_lexp_and_sl_in_ty t2)
  | RecT(id, tl, sl) -> RecT(id, List.map simplify_lexp_and_sl_in_ty tl, simplify_lexp_and_sl_in_sl sl)
  | ParT(t1,t2) -> ParT(simplify_lexp_and_sl_in_ty t1,simplify_lexp_and_sl_in_ty t2)
  | IncT(t1) -> IncT(simplify_lexp_and_sl_in_ty t1)
  | RepT(t1) -> RepT(simplify_lexp_and_sl_in_ty t1)
  | UpT(t1,l) -> UpT(simplify_lexp_and_sl_in_ty t1, simplify_lexp_and_sl_in_l l);;
let simplify_lexp_and_sl_in_ac ac =
  match ac with
    EqT(t1,t2) -> EqT(simplify_lexp_and_sl_in_ty t1,simplify_lexp_and_sl_in_ty t2)
  | TopT(t1) -> TopT(simplify_lexp_and_sl_in_ty t1)
  | SubT(t1,t2) -> SubT(simplify_lexp_and_sl_in_ty t1,simplify_lexp_and_sl_in_ty t2)
  | WfT(sl,t1) -> WfT(sl, simplify_lexp_and_sl_in_ty t1)
  | LowT(t1) -> LowT(simplify_lexp_and_sl_in_ty t1)
  | NocapT(t1) -> NocapT(simplify_lexp_and_sl_in_ty t1)
  | SubU(uv,u) -> SubU(uv,simplify_lexp_and_sl_in_u u)
  | RelU(u) -> RelU(simplify_lexp_and_sl_in_u u)
  | SubL(l1,l2) -> SubL(simplify_lexp_and_sl_in_l l1, simplify_lexp_and_sl_in_l l2)
  | SubS(sl1,sl2) -> SubS(simplify_lexp_and_sl_in_sl sl1, simplify_lexp_and_sl_in_sl sl2)
  | SubSIf(l1,sl1,sl2) -> SubSIf(simplify_lexp_and_sl_in_l l1, 
                                 simplify_lexp_and_sl_in_sl sl1, 
                                 simplify_lexp_and_sl_in_sl sl2)
  | EqS(sl1,sl2) -> EqS(simplify_lexp_and_sl_in_sl sl1, simplify_lexp_and_sl_in_sl sl2);;

let simplify_lexp_and_sl_in_c c =
  List.map simplify_lexp_and_sl_in_ac c;;
let simplify_lexp_and_sl_in_te =
   List.map (function ((x:tvid),t) ->(x, simplify_lexp_and_sl_in_ty t));;
let simplify_lexp_and_sl_in_obtab: (pc*levexp) list -> (pc*levexp) list =
   List.map (function (pc,l) ->(pc, simplify_lexp_and_sl_in_l l));;
let simplify_lexp_and_sl_in_tytab =
   List.map (function (pc,t) ->(pc, simplify_lexp_and_sl_in_ty t));;
let simplify_lexp_and_sl_in_tsubst =
   List.map (function (x,t) -> (x,simplify_lexp_and_sl_in_ty t));;
let simplify_lexp_and_sl_in_slsubst =
   List.map (function (x,sl) -> (x,simplify_lexp_and_sl_in_sl sl));;
let simplify_lexp_and_sl_in_usubst =
   List.map (function (x,u) -> (x,simplify_lexp_and_sl_in_u u));;
let simplify_lexp_and_sl_in_lsubst =
   List.map (function (x,l) -> (x,simplify_lexp_and_sl_in_l l));;
let simplify_lexp_and_sl_in_subst (t_s,sl_s,u_s,l_s) =
  (simplify_lexp_and_sl_in_tsubst t_s,
   simplify_lexp_and_sl_in_slsubst sl_s,
   simplify_lexp_and_sl_in_usubst u_s,
   simplify_lexp_and_sl_in_lsubst l_s);;
(*** TODO: use cache to solve subl to avoid duplication of the same expression ***)
(*** to enable sharing, we should also cache the result of substitution? ***)
let rec solve_SubL levgeqs lsubst =
  match levgeqs with
     [] -> List.map (function (x,l)->(x, simplify_lexp_and_sl_in_l l)) lsubst
   | (lv,l)::levgeqs' ->
       let l' = simplify_lexp_and_sl_in_l(solve_one_subl lv l) in
       let news = lsubst2subst [(lv,l')] in
       let lsubst' = (lv,l')::(List.map (function (x,l)->(x,apply_subst_lev news l)) lsubst) in
       let new_levgeqs = List.map (function (x,l)->(x,apply_subst_lev news l)) levgeqs' in
          solve_SubL new_levgeqs lsubst';;



let eliminate_SubL c =
  let (levgeq, c1) = get_SubL c in
  let levvars = fromto 0 (!current_lvid) in
  let deflvars = 
          delete_duplication (List.sort compare (List.map (function (lv,l) -> lv) levgeq)) in
  let undef_lvars = List.filter (function lv ->not(List.mem lv deflvars)) levvars in
       (*** set of unconstrained level variables ***)
       (*** Note: No level variables have been bound so far. ***)
  let levgeq2 = levgeq@(List.map (function lv -> (lv, LevConst 0)) undef_lvars) in
  let levgeq3 = List.sort (function (lv1,_)->function (lv2,_)->compare lv1 lv2) levgeq2 in
  let norm_geq = delete_duplicated_levgeq levgeq3 in 
  let lsubst =  (solve_SubL norm_geq []) in
  let s = lsubst2subst(lsubst) in
     (apply_subst_c s c1, lsubst);;
  
let ignore_security_constraints levmap =
  let s = slsubst2subst (List.map (function sv -> (sv,HI)) (fromto 0 (!current_svid))) in
  let levmap' = 
        List.map (function (x,l)-> (x, simplify_lexp_and_sl_in_l (apply_subst_lev s l))) levmap in
    levmap';;


type iomode = Mode_I | Mode_O | Mode_NIL | Mode_IO

let io_union m1 m2 =
  match (m1, m2) with
    (Mode_NIL, m) -> m
  | (m, Mode_NIL) -> m
  | _ -> if m1=m2 then m1 else Mode_IO

let rec u2io u iomap =
  match u with
    UVar(uv) -> List.assoc uv iomap
  | ZeroU -> Mode_NIL
  | InU(_,_,u1) -> io_union Mode_I (u2io u1 iomap)
  | OutU(_,_,u1) -> io_union Mode_O (u2io u1 iomap)
  | ParU(u1,u2) -> io_union (u2io u1 iomap) (u2io u2 iomap)
  | OrU(u1,u2) -> io_union (u2io u1 iomap) (u2io u2 iomap)
  | RepU(u1) -> u2io u1 iomap
  | IncU(u1) -> u2io u1 iomap
  | UpU(u1,_,_) -> u2io u1 iomap
  | URec(uv, u1) -> u2io u1 ((uv,Mode_NIL)::iomap)

let get_io_modes umap =
  let rec get_io_modes_it umap iomap =
     let iomap' = List.map (fun (uv, io) -> (uv, io_union io (u2io (List.assoc uv umap) iomap))) iomap in
      if iomap' = iomap 
      then iomap
      else get_io_modes_it umap iomap'
  in
  let iomap = List.map (fun (uv,u) -> (uv, Mode_NIL)) umap in
     get_io_modes_it umap iomap
     

let merge_nocapob (nocaps1,noobs1) (nocaps2, noobs2) =
         (Utilities.merge_and_unify compare nocaps1 nocaps2,
          Utilities.merge_and_unify compare noobs1 noobs2)

let rec elim_nocap_type t iomap =
  match t with
    UnitT -> ([], [])
  | BoolT _ -> ([], [])
  | IntT _ ->  ([], [])
  | PairT(t1, t2) ->
         merge_nocapob (elim_nocap_type t1 iomap) (elim_nocap_type t2 iomap)
  | ChanT(t1, _, UVar uv) ->
      let m = List.assoc uv iomap in
      let (nocaps1, noobs1) =
        match m with
          Mode_NIL -> ([], [])
        | Mode_I -> elim_nocap_type t1 iomap
        | Mode_O -> elim_noob_type t1 iomap
        | Mode_IO ->
            merge_nocapob (elim_nocap_type t1 iomap) (elim_noob_type t1 iomap) 
      in
        merge_nocapob ([uv], []) (nocaps1, noobs1)
  | ChanT(t1, _, _) -> raise (SomethingIsWrongIn "elim_nocap_type")
  | RecT(_, tl, _) ->
      List.fold_left (fun (nocaps, noobs) -> fun t1 -> 
                        merge_nocapob (elim_nocap_type t1 iomap) (nocaps, noobs))
                      ([], [])
                      tl
  | _ -> raise (SomethingIsWrongIn "elim_nocap_type")
and elim_noob_type t iomap = 
  match t with
    UnitT -> ([], [])
  | BoolT _ -> ([], [])
  | IntT _ ->  ([], [])
  | PairT(t1, t2) ->
         merge_nocapob (elim_noob_type t1 iomap) (elim_noob_type t2 iomap)
  | ChanT(t1, _, UVar uv) ->
      let m = List.assoc uv iomap in
      let (nocaps1, noobs1) =
        match m with
          Mode_NIL -> ([], [])
        | Mode_I -> elim_noob_type t1 iomap
        | Mode_O -> elim_nocap_type t1 iomap
        | Mode_IO ->
            merge_nocapob (elim_nocap_type t1 iomap) (elim_noob_type t1 iomap) 
      in
        merge_nocapob ([], [uv]) (nocaps1, noobs1)
  | ChanT(t1, _, _) -> raise (UnExpectedType t)
  | RecT(_, tl, _) ->
      List.fold_left (fun (nocaps, noobs) -> fun t1 -> 
                        merge_nocapob (elim_noob_type t1 iomap) (nocaps, noobs))
                      ([], [])
                      tl
  | _ -> raise (UnExpectedType t)
                        

let rec elim_nocapT_in_c c iomap =
  match c with
     [] -> ([], [], [])
   | NocapT(t)::c1 -> 
        let (nocaps, noobs) = elim_nocap_type t iomap in
        let (nocaps', noobs', c') = elim_nocapT_in_c c1 iomap in
         (Utilities.merge_and_unify compare nocaps nocaps',
          Utilities.merge_and_unify compare noobs noobs', c')
   | ac::c1 -> 
        let (nocaps', noobs', c') = elim_nocapT_in_c c1 iomap in
           (nocaps', noobs', ac::c');;

let get_uv_closure uvs umap =
  let rec get_uv_it uvs umap uvs_checked =
    match uvs with
       [] -> uvs_checked
     | uv::uvs' ->
         if List.mem uv uvs_checked then get_uv_it uvs' umap uvs_checked
         else
           let uvs1 = uvars_in_usage (List.assoc uv umap) in
             get_uv_it (uvs1@uvs') umap (Utilities.merge_and_unify compare [uv] uvs_checked)
  in
    get_uv_it uvs umap [];;

let rec replace_ob_with_inf u =
  match u with
    UVar _ -> u
  | ZeroU -> u
  | InU(l1, l2, u1) -> InU(LevInfty, l2, replace_ob_with_inf u1)
  | OutU(l1, l2, u1) -> OutU(LevInfty, l2, replace_ob_with_inf u1)
  | ParU(u1, u2) -> ParU(replace_ob_with_inf u1,replace_ob_with_inf u2)
  | OrU(u1,u2) -> OrU(replace_ob_with_inf u1,replace_ob_with_inf u2)
  | RepU(u1) -> RepU(replace_ob_with_inf u1)
  | IncU(u1) -> replace_ob_with_inf u1
  | UpU(u1, l1, l2) -> replace_ob_with_inf u1
  | _ -> raise (SomethingIsWrongIn "replace_ob_with_inf")

let make_noob uv umap = 
  let (u, umap') = Utilities.list_assoc2 uv umap in
  let u' = replace_ob_with_inf u in
    (uv, u')::umap'

let rec mkC_infcap u =
  match u with
    UVar _ -> []
  | ZeroU -> []
  | InU(l1, l2, u1) -> 
         SubL(LevInfty, l2)::(mkC_infcap u1)
  | OutU(l1, l2, u1) -> 
         SubL(LevInfty, l2)::(mkC_infcap u1)
  | ParU(u1, u2) -> 
         unionClist [mkC_infcap u1; mkC_infcap u2]
  | OrU(u1,u2) -> 
         unionClist [mkC_infcap u1; mkC_infcap u2]
  | RepU(u1) -> mkC_infcap u1
  | IncU(u1) -> mkC_infcap u1
  | UpU(u1, l1, l2) -> mkC_infcap u1
  | _ -> raise (SomethingIsWrongIn "replace_ob_with_inf")

let make_nocap uv c = 
  let u = List.assoc uv c in
    mkC_infcap u

let eliminate_nocapT c =
   let uvs = uvars_in_c c in
   let (c1,c2) = get_subusage_constraint c in
   let c1' = normalize_subusage_constraint c1 uvs in
   let umap = List.map 
               (function SubU(uv,u) -> (uv, u) | _ -> raise (SomethingIsWrongIn "get_io_modes")) 
               c1' in
   let io_mode_map = get_io_modes umap in
   let (nocaps, noobs, c2') = elim_nocapT_in_c c2 io_mode_map in
   let nocap_uvs = get_uv_closure nocaps umap in
   let noob_uvs = get_uv_closure noobs umap in
   let c1'' = List.map
                 (fun (uv,u) -> SubU(uv,u))
                 (List.fold_left (fun umap -> fun uv -> make_noob uv umap) umap noob_uvs ) in
   let c3 = unionClist (List.map (fun uv -> make_nocap uv umap) nocap_uvs) in
     unionClist [c1'';c2';c3]


let lockfreedom_analysis_core p =
  let _ = reset_TEtable() in
  let (te1, sl1, c1, cm1, tytab1) = iflowPT p emptyVord in
  let (te2, sl2, c2, tc2, _,s) = 
      eliminate_type_constraints(te1,sl1,c1,[]) in
  let _ = map_TEtable (apply_subst_te s) in
  let c2 = if !Flag.wlock then eliminate_nocapT c2 else c2 in
  let (te3, uc3, c3, s3) = solve_SubU (te2, c2) in
  let _ = map_TEtable (apply_subst_te s3) in
  let (te4, uc4, c4, obtable,_) = eliminate_Up_and_Inc (te3,uc3,c3, empty_tytab) in
  let _ = if !Flag.debug then (print_string "eliminating RelU\n"; flush stdout) else () in
  let c5 = eliminate_RelU uc4 c4 obtable in
  let _ = if !Flag.debug then (print_string "RelU eliminated\n"; flush stdout) else () in
  let (c6,lsubst6) = eliminate_SubL(c5) in
  let lsubst7 = ignore_security_constraints(lsubst6) in
  let s7 = lsubst2subst lsubst7 in
  let uc7 = apply_subst_c s7 uc4 in
  let obtable' = List.map 
                  (fun (uv, (l1,l2)) -> 
                       (uv, (simplify_lexp_and_sl_in_l (apply_subst_lev s7 l1),
                             simplify_lexp_and_sl_in_l (apply_subst_lev s7 l2))))
                  obtable
                 in
  let _ = map_TEtable (apply_subst_te s7) in
  let cm7 = List.map (function (pc,l) -> 
			   (pc,simplify_lexp_and_sl_in_l
			         (apply_subst_lev (lsubst2subst lsubst7) l))) cm1 in
  let modtab = get_io_modes (List.map (function SubU(uv,u) -> (uv,u)
                                           | _ -> raise (SomethingIsWrongIn "modtab")) uc7) in
       (List.sort (function (pc1,_)->function (pc2,_)->compare pc1 pc2) cm7,
        !te_table,
        modtab,
        obtable');;

let lockfreedom_analysis p =
  let (p1, stenv) = Stype.elaborate p in
  let (_, ord) = Otype.elaborate(p1, stenv) in
  let _ = (name_order := ord) in
  let (result, _, _,_) = lockfreedom_analysis_core p
    in result;;

let eqclass_as_pairs vs = 
  let rec eqclass_as_pairs_aux vs =
     (match vs with
       [] -> []
     | [v] -> []
     | v1::v2::vs' -> merge_and_unify compare [(v1,v2)] (eqclass_as_pairs_aux (v2::vs'))) in
  match vs with
     [] -> []
   | [v] -> []
   | v1::v2::vs' -> merge_and_unify compare [(list_last(vs), v1)] (eqclass_as_pairs_aux vs);;

let mk_list_of_related_tyvars_in_ac ac =
  match ac with
    EqT(t1,t2) -> 
         let tvs1 = tyvars_in_t t1 in
         let tvs2 = tyvars_in_t t2 in
	 let tvs = merge_and_unify compare tvs1 tvs2 in
	    eqclass_as_pairs tvs
  | TopT(t1) ->
         let tvs = tyvars_in_t t1 in
	    eqclass_as_pairs tvs
  | SubT(t1,t2) ->    
         let tvs1 = tyvars_in_t t1 in
         let tvs2 = tyvars_in_t t2 in
	 let tvs = merge_and_unify compare tvs1 tvs2 in
	    eqclass_as_pairs tvs
  | WfT(sl1,t1) ->
         let tvs = tyvars_in_t t1 in
	    eqclass_as_pairs tvs
  | LowT(t1) ->
         let tvs = tyvars_in_t t1 in
	    eqclass_as_pairs tvs
  | NocapT(t1) ->
         let tvs = tyvars_in_t t1 in
	    eqclass_as_pairs tvs
  | _ -> raise (SomethingIsWrongIn "mk_list_of_related_tyvars_in_ac");;
    
let mk_list_of_related_tyvars_in_c tc =
  List.fold_left (function l -> function ac -> 
			   merge_and_unify compare l (mk_list_of_related_tyvars_in_ac ac))
                [] tc;;
(*** find_eqclass: 'a -> 'a list list -> 'a list ****)
(*** Given an element and an equivalence relation (represented as a list of
 *** equivalence classes), find the equivalence relation containing the element
 ***)
let rec find_eqclass v eqrel =
  match eqrel with
     [] -> [v]
   | c::eqrel' -> if List.mem v c then c else find_eqclass v eqrel';;

let rec closure_of_tyvars tvs tc =
  let r = mk_list_of_related_tyvars_in_c tc in
  let eqrel = Scc.compute_scc r in
    List.fold_left (function vs -> function v -> merge_and_unify compare vs (find_eqclass v eqrel))
                   [] tvs;;

let eliminate_unnecessary_tyvars te tc tytab =
  let tvs1 = tyvars_in_te te in
  let tvs2 = closure_of_tyvars tvs1 tc in
  let tvs3 = tyvars_in_tytab tytab in
  let tvs4 = diff_vars tvs3 tvs2 in
  let ts = List.map (function v -> (v, UnitT)) tvs4 in
  let s = tsubst2subst ts in
  let te' = apply_subst_te s te in
  let tytab' = apply_subst_tytab s tytab in
    (te', tytab', s);;

let is_tautology_ac ac =
  match ac with
    SubS(sl1,sl2) ->
      let sl1' = simplify_lexp_and_sl_in_sl sl1 in
      let sl2' = simplify_lexp_and_sl_in_sl sl2 in
       ( match (sl1',sl2') with
          (LOW, _) -> true
        | (_, HI) -> true
        | (HI, LOW) -> false
        | _ -> raise (SomethingIsWrongIn "is_tautology_ac:1"))
  | _ ->  raise (SomethingIsWrongIn "is_tautology_ac:2");;

let is_tautology c = List.fold_right (function ac -> (&&) (is_tautology_ac ac)) c true;;
let subsif_to_subs ac =
  match ac with
    SubSIf(l,sl1,sl2) -> SubS(sl1, SIfInf(l,sl2))
  | _ -> ac;;
let eliminate_SubSIf c = 
    List.map subsif_to_subs c;;

let compare_SubS (sv1,sl1) (sv2,sl2) =
  compare sv1 sv2;;

let rec take_SubS c =
  match c with
    [] -> ([], [])
  | SubS(SVar(sv1),sl2)::c' ->
       let (c_SubS, c_others) = take_SubS(c') in
         ((sv1,sl2)::c_SubS, c_others)
  | ac::c' ->
       let (c_SubS, c_others) = take_SubS(c') in
         (c_SubS, ac::c_others);;


let rec solve_rec_SubS_1step unchecked checked s =
  match unchecked with
    [] -> (checked, s)
  | (sv,sl)::c ->
       let sl' = simplify_lexp_and_sl_in_sl(apply_subst_sl s sl) in
         if sl'=HI then solve_rec_SubS_1step c ((sv,sl)::checked) s
         else if sl'=LOW then 
           let s' = comp_subst s (make_slsubst sv LOW) in
             solve_rec_SubS_1step c checked s'
         else raise (UnExpectedSL sl');;
let rec solve_rec_SubS c_SubS s = 
  let (c_SubS', s') = solve_rec_SubS_1step c_SubS [] s in
    if s==s' then s else solve_rec_SubS c_SubS' s';;

let solve_SubS c s = 
  let (c_SubS,c_others) = take_SubS c in
  let c_SubS' = List.sort compare_SubS c_SubS in
  let c_SubS'' =  delete_duplication2 compare_SubS
                   (function (sv1,sl1)->function (sv2,sl2) -> (sv1,sglb(sl1,sl2))) c_SubS' in
  let s1 = solve_rec_SubS c_SubS'' s in
    (s1, apply_subst_c s1 c_others);;

let add_subst_on_remaing_slvars s =
  let slvars = fromto 0 (!current_svid) in
  let s1 = slsubst2subst (List.map (function sv -> (sv,HI)) slvars) in
    comp_subst s1 s;;

let solve_EqS_and_SubS c = 
  let _ = if !Flag.debug then Printf.printf "solve_EqS_and_SubS phase 1: %f\n" (Sys.time()) in
(***  let _ = raise (Test c) in ***)
  let (s0, c0) = simplify_SubS_and_EqS c in
  let _ = if !Flag.debug then Printf.printf "solve_EqS_and_SubS phase 1.1: %f\n" (Sys.time()) in
  let (s1, c1) = eliminate_EqS c0 empty_subst [] in
  let _ = if !Flag.debug then Printf.printf "solve_EqS_and_SubS phase 2: %f\n" (Sys.time()) in
  let c2 = eliminate_SubSIf c1 in (*** SubSIf(l,s1,s2) --> SubS(s1,SIfInf(l,s2,H)) ***)
  let slvars = fromto 0 (!current_svid) in
  let s_init = slsubst2subst (List.map (function sv -> (sv,HI)) slvars) in
  let _ = if !Flag.debug then Printf.printf "solve_EqS_and_SubS phase 3: %f\n" (Sys.time()) in
  let (s3,c3) = solve_SubS c2 s_init in  (*** eliminate constraints of the form SubS(SVar(v1),sl) ***)
  let _ = if !Flag.debug then Printf.printf "solve_EqS_and_SubS phase 4: %f\n" (Sys.time()) in
  let c4 = simplify_lexp_and_sl_in_c c3 in
  let _ = if !Flag.debug then Printf.printf "solve_EqS_and_SubS phase 5: %f\n" (Sys.time()) in
  let s = (comp_subst s3 (comp_subst s1 s0)) in
    if is_tautology c4  (*** check whether the remaining constraints 
			 ***  (of the form SubS(sl1,sl2) where sl1 is a constant)
                         *** are tautology ***)
      then s
      else raise (SomethingIsWrongIn "solve_EqS_and_SubS");;

(*** (is_informative t) returns whether t contains low-security information ***)
let rec is_informative (t:vtype) =
  match t with
    TVar(_) -> true
  | UnitT -> false
  | BoolT(sl) -> not(sl=HI)
  | IntT(sl) -> not(sl=HI)
  | ChanT(t1,sl,u) -> not(sl=HI)
  | PairT(t1,t2) -> (is_informative t1)||(is_informative t2)
  | FunT(t1,t2) -> (is_informative t2)
  | RecT(id1, tl, sl) ->
      let al = attributes_of_RecT id1 in
      let l = List.combine al tl in
         not(sl=HI) 
        || (List.fold_right 
          (fun (a,t) -> fun b -> if is_strictly_positive(a) then b||(is_informative t) else b)
           l false)
  | ParT(t1,t2) -> (is_informative t1)||(is_informative t2)
  | IncT(t1) -> (is_informative t1)
  | RepT(t1) -> (is_informative t1)
  | UpT(t1,l) -> (is_informative t1);;

(*** (is_informative t) returns true if t contains no high-security information ***)
let rec is_free_of_garbage (t:vtype) =
  match t with
    TVar(_) -> true
  | UnitT -> true
  | BoolT(sl) -> not(sl=HI)
  | IntT(sl) -> not(sl=HI)
  | ChanT(t1,sl,u) -> not(sl=HI)
  | PairT(t1,t2) -> (is_free_of_garbage t1)&&(is_free_of_garbage t2)
  | FunT(t1,t2) -> (is_free_of_garbage t2)
  | ParT(t1,t2) -> (is_free_of_garbage t1)&&(is_free_of_garbage t2)
  | RecT(id1, tl, sl) ->
      let al = attributes_of_RecT id1 in
      let l = List.combine al tl in
         not(sl=HI)
         && (List.fold_right 
          (fun (a,t) -> fun b -> if is_strictly_positive(a) then b && (is_free_of_garbage t) else b)
           l false)
  | IncT(t1) -> (is_free_of_garbage t1)
  | RepT(t1) -> (is_free_of_garbage t1)
  | UpT(t1,l) -> (is_free_of_garbage t1);;

(*** delete_high_value replaces sub-expressions of high security level with ().
 ***)

let current_dummy_pc: pc ref = ref 0;;
let gen_dummy_pc () = 
     (current_dummy_pc := !current_dummy_pc-1;
      !current_dummy_pc);;
let reset_dummy_pc() =
     (current_dummy_pc := 0);;
let rec delete_high_value v t tytab = 
  if not(is_informative t) then
      Unit
  else
    match v with
      Var(x) -> 
        (match t with
           PairT(t1,t2) ->
             if (is_free_of_garbage t) then v
             else 
               let v1 = Var "first" in
               let v1' = delete_high_value v1 t1 tytab in
               let v2 = Var "second" in
               let v2' = delete_high_value v2 t2 tytab in
               let vs = comp_vsubst 
                         (make_vsubst "first" (App (Prim ("fst",None), v,gen_dummy_pc())))
	                 (make_vsubst "second" (App (Prim ("snd",None), v,gen_dummy_pc()))) in
                  apply_vsubst_v vs (Pair(v1',v2'))
         | _ -> v)
   | Unit -> Unit
   | SBool _ -> v
   | SInt _ -> v
   | Bool _ -> v
   | Int _ -> v
   | Prim (_,_) -> v
   | App(v1, v2, pc) -> 
       let t1 = List.assoc pc tytab in
       let v1' = delete_high_value v1 (FunT(t1,t)) tytab in
       let v2' = delete_high_value v2 t1 tytab in
         App(v1', v2', pc)
   | Pair(v1, v2) ->
        (match t with
          PairT(t1,t2) ->
            let v1' = delete_high_value v1 t1 tytab in
            let v2' = delete_high_value v2 t2 tytab in
              Pair(v1', v2')
         | _ -> raise (SomethingIsWrongIn "delete_high_value"));;
                    
let rec delete_high_process (p: unit Syntax.procexp) tytab = 
  match p with
    Zero -> Zero
  | Out(x,v,p1,pc) ->
      let p1' = delete_high_process p1 tytab in
      let t = List.assoc pc tytab in
        (match t with
          ChanT(t1,sl,u) ->
            if sl=HI then p1'
            else 
              let v' = delete_high_value v t1 tytab in Out(x,v',p1',pc)
         | _ -> raise (SomethingIsWrongIn "delete_high_process"))
  | In(x,y,t,p1,pc) ->
      let p1' = delete_high_process p1 tytab in
      let t1 = List.assoc pc tytab in
        (match t1 with
          ChanT(t1,sl,u) ->
            if sl=HI then Let(y,t,Unit,p1',pc)
            else 
              In(x,y,t,p1',pc)
         | _ -> raise (SomethingIsWrongIn "delete_high_process"))
  | Par(p1,p2) ->
      let p1' = delete_high_process p1 tytab in
      let p2' = delete_high_process p2 tytab in
         Par(p1',p2')
  | Rep(p1) ->
      let p1' = delete_high_process p1 tytab in
         Rep(p1')
  | Nu(x,t,p1,pc) ->
      let p1' = delete_high_process p1 tytab in
      let t1 = List.assoc pc tytab in
        (match t1 with
          ChanT(_,sl,u) ->
            if sl=HI then Let(x,t,Unit,p1',pc)
            else 
              Nu(x,t,p1',pc)
         | _ -> raise (SomethingIsWrongIn "delete_high_process"))
  | If(v,p1,p2,pc) ->
      let p1' = delete_high_process p1 tytab in
      let p2' = delete_high_process p2 tytab in
      let t = List.assoc pc tytab in
        (match t with
          BoolT(sl) ->
            if sl=HI then Zero
            else 
              let v' = delete_high_value v t tytab in
                  If(v',p1',p2',pc)
         | _ -> raise (SomethingIsWrongIn "delete_high_process"))
  | MatchL(x, ml, pc) ->
      let t = List.assoc pc tytab in
        (match t with
           RecT(id1,tl,sl) ->
             if sl=HI then Zero
             else
               let ml' = List.map (fun (con, p)->(con, delete_high_process p tytab)) ml in
                  MatchL(x, ml', pc)
         | _ -> raise (SomethingIsWrongIn "delete_high_process")
        )
  | Let(x,t,v,p1,pc) ->                  
      let p1' = delete_high_process p1 tytab in
      let t1 = List.assoc pc tytab in
      let v' = delete_high_value v t1 tytab in
        Let(x,t,v',p1',pc)                          
  | Special(p1) ->
      let p1' = delete_high_process p1 tytab in
         Special(p1')

let dependency_analysis p =
  let _ = if !Flag.debug then Printf.printf "Start Time: %f\n" (Sys.time()) in
  let (te1, sl1, c1, cm1, tytab1) = iflowPT p emptyVord in
  let _ = if !Flag.debug then Printf.printf "Phase 1 completed: %f\n" (Sys.time()) in
  let c2 = unionClist (c1::(List.map (function (x,t)->makeC_LowT(t)) te1)) in
  let (te3, sl3, c3, tc3, tytab3,s3) = eliminate_type_constraints(te1,sl1,c2,tytab1) in
  let _ = if !Flag.debug then Printf.printf "Phase 2 completed: %f\n" (Sys.time()) in
  let (te4, tytab4, s4) = eliminate_unnecessary_tyvars te3 tc3 tytab3 in
  let _ = if !Flag.debug then Printf.printf "Phase 3 completed: %f\n" (Sys.time()) in
  let (te5,uc5,c5,s5) = solve_SubU(te3, c3) in
  let _ = if !Flag.debug then Printf.printf "Phase 4 completed: %f\n" (Sys.time()) in
  let tytab5 = apply_subst_tytab s5 tytab4 in
  let (te6, uc6, c6, obtable, tytab6) = eliminate_Up_and_Inc (te5,uc5,c5, tytab5) in
  let _ = if !Flag.debug then Printf.printf "Phase 5 completed: %f\n" (Sys.time()) in
  let c7 = eliminate_RelU uc6 c6 obtable in
  let _ = if !Flag.debug then Printf.printf "Phase 6 completed: %f\n" (Sys.time()) in
  let (c8,lsubst8) = eliminate_SubL(c7) in
  let _ = if !Flag.debug then Printf.printf "Phase 7 completed: %f\n" (Sys.time()) in
  let s8 = lsubst2subst lsubst8 in
  let s9 = solve_EqS_and_SubS c8 in
  let _ = if !Flag.debug then Printf.printf "Phase 8 completed: %f\n" (Sys.time()) in
(***
  let te9 = simplify_lexp_and_sl_in_te(apply_subst_te s9 te8) in
***)
  let s10 = simplify_lexp_and_sl_in_subst (comp_subst s9 s8) in
  let tytab9 = simplify_lexp_and_sl_in_tytab(apply_subst_tytab s10 tytab6) in
  let _ = if !Flag.debug then Printf.printf "Phase 9 completed: %f\n" (Sys.time()) in
    tytab9;;
let useless_code_elimination (p:unit Syntax.procexp) =
  let tytab = dependency_analysis p in
  let _ = reset_dummy_pc() in
  let p' = reduce_proc(delete_high_process p tytab) in
     p';;

let rec annotate_v_with_sl v t tytab =
  match v with
     Var(x) -> v
   | Unit -> Unit
   | SBool(b,sl) -> 
       (match t with
         BoolT(sl') -> if sl'=HI then SBool(b,Hi) else SBool(b,Low)
       | _ -> raise (SomethingIsWrongIn "annotate_v_with_sl"))
   | SInt(n,sl) -> 
       (match t with
         IntT(sl') -> if sl'=HI then SInt(n,Hi) else SInt(n,Low)
       | _ -> raise (SomethingIsWrongIn "annotate_v_with_sl"))
   | Bool b -> 
       (match t with
         BoolT(sl') -> if sl'=HI then SBool(b,Hi) else SBool(b,Low)
       | _ -> raise (SomethingIsWrongIn "annotate_v_with_sl"))
   | Int n -> 
       (match t with
         IntT(sl') -> if sl'=HI then SInt(n,Hi) else SInt(n,Low)
       | _ -> raise (SomethingIsWrongIn "annotate_v_with_sl"))
   | Prim (_,_) -> v
   | App(v1, v2, pc) -> 
       let t1 = List.assoc pc tytab in
       let v1' = annotate_v_with_sl v1 (FunT(t1,t)) tytab in
       let v2' = annotate_v_with_sl v2 t1 tytab in
         App(v1', v2', pc)
   | Pair(v1, v2) ->
        (match t with
          PairT(t1,t2) ->
            let v1' = annotate_v_with_sl v1 t1 tytab in
            let v2' = annotate_v_with_sl v2 t2 tytab in
              Pair(v1', v2')
         | _ -> raise (SomethingIsWrongIn "annotate_v_with_sl"));;
    
let rec annotate_p_with_sl p tytab =
  match p with
    Zero -> Zero
  | Out(x,v,p1,pc) ->
      let p1' = annotate_p_with_sl p1 tytab in
      let t = List.assoc pc tytab in
        (match t with
          ChanT(t1,sl,u) ->
             let v' = annotate_v_with_sl v t1 tytab in Out(x,v',p1',pc)
         | _ -> raise (SomethingIsWrongIn "annotate_p_with_sl"))
  | In(x,y,t,p1,pc) ->
      let p1' = annotate_p_with_sl p1 tytab in
          In(x,y,t,p1',pc)
  | Par(p1,p2) ->
      let p1' = annotate_p_with_sl p1 tytab in
      let p2' = annotate_p_with_sl p2 tytab in
         Par(p1',p2')
  | Rep(p1) ->
      let p1' = annotate_p_with_sl p1 tytab in
         Rep(p1')
  | Nu(x,t,p1,pc) ->
      let p1' = annotate_p_with_sl p1 tytab in
         Nu(x,t,p1',pc)
  | If(v,p1,p2,pc) ->
      let p1' = annotate_p_with_sl p1 tytab in
      let p2' = annotate_p_with_sl p2 tytab in
      let t = List.assoc pc tytab in
      let v' = annotate_v_with_sl v t tytab in
         If(v',p1',p2',pc)
  | MatchL(x,ml,pc) ->
      let ml' = List.map (fun (con,p)->(con, annotate_p_with_sl p tytab)) ml in
         MatchL(x,ml',pc)
  | Let(x,t,v,p1,pc) ->                  
      let p1' = annotate_p_with_sl p1 tytab in
      let t1 = List.assoc pc tytab in
      let v' = annotate_v_with_sl v t1 tytab in
        Let(x,t,v',p1',pc)
  | Special(p1) ->
      let p1' = annotate_p_with_sl p1 tytab in
         Special(p1')

let iflow_analysis p =
  let tytab = dependency_analysis p in
    annotate_p_with_sl p tytab;;

(***
let test n = 
  let p = parse ("samples_u/test"^(string_of_int(n))^".pi") in
     useless_code_elimination p;;
     
let p = parse "samples/test1.pi";;
let (te,uc,tytab)=useless_code_elimination p;;
***)

let wlockfreedom_analysis p =
 (*** let p' = Syntax.alphaconv p in ***)
  let (p1, stenv) = Stype.elaborate p in
  let specials = Stype.get_specials p1 stenv in
  let tresult = List.map Termination.termination_check specials in
    if List.for_all (fun x->x) tresult
    then
      let _ = print_string "Robust termination analysis has succeeded\n" in
      let t = Sys.time() in
      let _ = (print_string ("Elapsed Time: "^(string_of_float (t -. !Flag.start_t))^"sec\n");
               print_string "Proceed to lock-freedom analysis...\n";
               flush stdout) in
      let (_, ord) = Otype.elaborate(p1, stenv) in
      let _ = (name_order := ord) in
      let (result, _, _,_) = lockfreedom_analysis_core p in
        result
    else
      let _ = print_string "Termination annotation is invalid\n" in
        raise FailWlockAnalysis

let binary_result res =
  List.map 
  (fun (pc, lev) -> 
      match lev with
        LevInfty -> (pc, LevInfty)
       | _ -> (pc, LevConst 0))
  res;;

let get_levels_from_obtab u obtab =
  let eI = List.map (fun (uv, (l1,l2)) -> (uv,l1)) obtab in
  let eO = List.map (fun (uv, (l1,l2)) -> (uv,l2)) obtab in
    (fun (l1,l2) -> (simplify_lexp_and_sl_in_l l1, simplify_lexp_and_sl_in_l l2))
              (get_levels u eI eO);;

(*** hasfinitecap t checks whether t has a finite capability
 *** For now, it investigates only outermost capabilities; It should be refined later
 ***)
let rec hasfinitecap_u u =
  match u with
    InU(l1,_,u) -> not(simplify_lexp_and_sl_in_l l1 = LevInfty)
  | OutU(l1,_,u) -> not(simplify_lexp_and_sl_in_l l1 = LevInfty)
  | ParU(u1,u2) -> (hasfinitecap_u u1)||(hasfinitecap_u u2)
  | OrU(u1,u2) -> (hasfinitecap_u u1)||(hasfinitecap_u u2)
  | RepU(u1) -> (hasfinitecap_u u1)
  | _ -> false

let rec hasfinitecap t =
  match t with
    ChanT(_,_,u) -> hasfinitecap_u u
  | PairT(t1,t2) -> (hasfinitecap t1)||(hasfinitecap t2)
  | _ -> false

let rec improvable_t modtab1 modtab2 obtab1 obtab2 (t1,t2) =
  match (t1,t2) with
    (UnitT,UnitT) -> false
  | (BoolT _, BoolT _) -> false
  | (IntT _, IntT _) -> false
  | (ChanT(t11,_,u1), ChanT(t21,_,u2)) ->
       let (l11,l12) = get_levels_from_obtab u1 obtab1 in
       let (l21,l22) = get_levels_from_obtab u2 obtab2 in
         if (l21=LevInfty && not(l11=LevInfty))||(l22=LevInfty && not(l12=LevInfty))
         then true
         else 
           let m1 = u2io u1 modtab1 in
             if m1=Mode_I then
                improvable_t modtab1 modtab2 obtab1 obtab2 (t11,t21)
             else
                false
  | (PairT(t11,t12),PairT(t21,t22)) ->
        ( (improvable_t modtab1 modtab2 obtab1 obtab2 (t11,t21))
          || (improvable_t modtab1 modtab2 obtab1 obtab2 (t12,t22)))
  | (FunT(_,_), FunT(_,_)) -> false
  | (RecT(_,tl1,_), RecT(_,tl2,_)) ->
        List.exists 
         (fun tpair ->(improvable_t modtab1 modtab2 obtab1 obtab2 tpair))
        (List.combine tl1 tl2)
  | _ -> raise (SomethingIsWrongIn "improvable_t");;

let improvable te1 te2 modtab1 modtab2 obtab1 obtab2 =
  let typairs = List.map (fun ((_,t1),(_,t2)) -> (t1,t2)) (List.combine te1 te2) in
  let (typairs1, typairs2) =
           List.partition (improvable_t modtab1 modtab2 obtab1 obtab2) typairs in
  let cost = List.length(List.filter hasfinitecap (List.map fst typairs2)) in
    (not(typairs1=[]), cost)

let wlockfreedom_auto_analysis p =
  let rec try_candidates candidates default =
    if candidates=[] then
       (print_string "No candidate for termination annotation was found\n\n";
        default)
    else
      let pc = List.hd candidates in
         let p' = Syntax.add_annotation p pc in
           try
             let result = wlockfreedom_analysis p' in
             let results = List.combine result default in
               List.map (fun ((pc1,l1),(pc2,l2)) -> (pc1, lexp_min l1 l2)) results
           with
             FailWlockAnalysis ->
               (print_string "Inferred termination annotation was wrong\n Trying the next candidate\n";
                try_candidates (List.tl candidates) default)
  in
  let (p1, stenv) = Stype.elaborate p in
  let (_, ord) = Otype.elaborate(p1, stenv) in
  let _ = (name_order := ord) in
  let _ = (Flag.livelock := false) in
  let _ = reset_names() in
  (*** deadlock-analysis ***)
  let (result1, tetable1, modtab1, obtable1) = lockfreedom_analysis_core p in 
  let _ = (Flag.livelock := true) in
  let _ = reset_names() in
  (*** lockfreedom analysis ***)
  let (result2, tetable2, modtab2, obtable2) = lockfreedom_analysis_core p in
  let diff = List.filter (fun ((pc1,l1),(pc2,l2)) -> not(l1=l2)) 
                 (List.combine (binary_result(result1)) (binary_result(result2))) in
  let diffpos = List.map (fun x -> fst(fst(x))) diff in
   if diffpos = []
   then result1
   else 
     let candidates = List.map (fun (pc,_) -> pc) tetable1 in
     let candidates1 = 
        List.filter (fun pc -> Syntax.rep_restricted (Syntax.get_subproc p pc)) candidates in
     let cand_tmp = List.map (fun pc -> 
                                    (pc, 
                                     improvable (List.assoc pc tetable1) (List.assoc pc tetable2)
                                                modtab1 modtab2 obtable1 obtable2))
                                    candidates1 in
     let cand_tmp2 = List.filter (fun (pc, (flag,cost)) -> flag) cand_tmp in
     let candidates_with_info = 
        List.map (fun (pc, (flag,cost)) -> 
                   let p1 = Syntax.get_subproc p pc in
                   let pcs = Syntax.pcs p1 in
                   let pcs' = List.filter (fun pc -> (List.mem pc diffpos)) pcs in
                   (pc, (cost, (List.length pcs', -(Syntax.sizeof p1)))))
                  cand_tmp2 in
     let candidates3 = 
             List.map fst (List.sort (fun p1 -> fun p2 -> compare (snd p1) (snd p2)) candidates_with_info)
     in
     let _ = (Flag.wlock := true) in
     let result = try_candidates candidates3 result2 in
        if not(binary_result(result)= binary_result(result2))
        then (*** If there is an improvement, print the result ***)
            result
        else (*** If there is no improvement, try global termination analysis ***)
          if Syntax.rep_restricted p 
          then
            (print_string "Modular analysis failed\n Will try global termination analysis\n";
             reset_names();
            try
             wlockfreedom_analysis (Syntax.Special p)
            with
              FailWlockAnalysis ->
               (print_string "Failed to prove global termination\n"; result)
             )
          else
            (print_string "Warning: Termination analysis failed to improve the result.\n";
             result)

            